c...Plane stress/strain elasticity "main engine"
c...for use with FELIPE version 3
c... - sample datafile is elcyl4.dat or eldemo.dat
c
c
c...If you want to re-compile this program, you must link it with
c...the file felsb.obj, compiled from felsb.for (provided in the
c...FELIPE folder). The file felsb.for contains additional subroutines 
c...used by all the 'main engines' except eladv.for. 
c...If you don't use linking, you should paste the contents of 
c...the file felsb.for to the end of this file, before compiling
c...and running it.
c
      program elast
c
      parameter(melem=200,mpoin=800,mnode=8,
     1msets=9,mcomp=6,mtotv=1600,mband=200)
c
c...these parameters fix the dimensions of the arrays:
c...melem = max. no. of elements
c...mpoin = max. no. of nodes
c...mnode = max. no. of nodes in a single element
c...msets = max. no. of mat. prop. sets (also used for loading sets)
c...mcomp = max. no. of mat. props (e.g. Young's mod, Poisson ratio)
c...mtotv = max no. of degrees of freedom (= 2*mpoin for elasticity)
c...mband = halfbandwidth+1 of global matrix gstif using band storage
c
c
      character char*1
      character*80 title,infile,outfil,prtfil
c
      dimension lnods(melem,mnode),lnodn(melem),lsidn(melem),
     1  coord(mpoin,2),nkode(mpoin),aslod(mtotv),ndof(mpoin),
     2        lpros(melem),props(msets,mcomp),asdis(mtotv),
     4      surft(msets,2),gstif(mtotv,mband),lmtyp(melem),
     5      nminvc(mtotv),astmp(mtotv),diagk(mtotv)
c...lnods(l,n) = global no. of n'th node of element l
c...lnodn(l) = no. of nodes in element l ( = 8 for 8-noded quad.)
c...lsidn = no. of sides in element l    ( = 4 for 8-noded quad.)
c...coord(np,1), coord(np,2) = x,y coords of node np
c...nkode(np) = fixity code of node np
c...aslod is global nodal load vector
c...ndof(np) = no. of dof.s at node np (not used in this program)
c...lpros(l) = mat. prop. set no. of element l
c...props(m,i) = value of i'th mat. property in mat. prop. set m
c...asdis is global nodal displacement vector
c...astmp is a scratch vector used in the equation solution substitutions
c...surft(m,1),surft(m,2) = normal,tangential surface tractions in set m
c...gstif holds assembled global stiffness matrix as a vector in skyline storage
c...      or as a matrix using band storage
c...nminvc(i)=j means that G(i,j) is the first nonzero entry in row i of gstif
c...lmtyp(l) = material type of elt l (not used in this program)
c...diagk stores the diagonal of the stiffness matrix (used to find reactions)
c
c...ifull = 0 for demo version: restricted to 50 elts or 50 nodes
      ifull = 1
c
      write(6,*)'Program ELAST: solves 2D elasticity problems.'
      write(6,*)'Handles 8-noded quadrilateral elements.'
      write(6,*)'Dimensioned for',melem,' elements, and',mpoin,' nodes.'
      if (ifull.eq.0)write(6,*)'Demonstration datafile is eldemo.dat'
      write(6,*)
c
c...read an existing .dat file
c
    1 continue
c
c...get the input file name in infile
      call askfil(infile,iflag)
      if (iflag.eq.1) then
        call yesno(char,'Do you want to exit?')
        if (char.eq.'y') stop
        goto 1
      endif
c
      write(6,*)'Opening file ',infile,'...'
      lenin = lenstr(infile)
      open(15,file=infile(1:lenin),status='old',err=999)
      write(6,*)
      lm4 = lenin - 4
      prtfil = infile(1:lm4)//'.prt'
      write(6,*)'Creating file ',prtfil(1:lenin),' for results...'
      write(6,*)
      open(16,file=prtfil(1:lenin),status='unknown',err=999)      
      outfil = infile(1:lm4)//'.out'
      write(6,*)'Creating file ',outfil(1:lenin),' for output...'
      write(6,*)
      open(17,file=outfil(1:lenin),status='unknown',err=999)      
c
c...read the input mesh data, and echo it to the output files
      call input(15,melem,mpoin,mnode,msets,mcomp,nelem,
     1 npoin,lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props
     2         ,ndofn,npros,title,ndof)
      call print(16,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1 lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2           npros,title,ndof)
      call output(17,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1  lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2            npros,title,ndof)
      call yesno(char,'Do you want the mesh details listed?')
      if (char.eq.'y') then
        call print(6,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1  lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2           npros,title,ndof)
      endif
c
      iflag = 0
      do 2 l=1,nelem
        nnode = lnodn(l)
        nside = lsidn(l)
        nodsid = 10*nnode + nside
        if (nodsid.ne.84) then
          iflag = 1
          write(6,*)'ERROR: Cannot handle element ',l
          write(6,*)'which has ',nside,' sides and ',nnode,' nodes'
        endif
    2 continue
      if (iflag.eq.1) then
        call prompt
        stop
      endif
c
      if (ifull.eq.0.and.(nelem.gt.51.or.npoin.gt.51)) then
        write(6,*)'Demo version will not process meshes with'
        write(6,*)'51 elements or more than 51 nodes.'
        call prompt
        stop
      endif
c
      write(6,*)
     1'Is this a plane stress, plane strain or axisymmetric analysis?'
    3 write(6,*)
     1'Type 1 for plane stress, 2 for plane strain, or 3 for axisymm...'
      read(5,'(i1)',err=3) mstyp
      write(16,*)
      if (mstyp.eq.1) then
        write(6,*) 'Plane stress analysis'
        write(16,*)'Plane stress analysis'
      else if (mstyp.eq.2) then
        write(6,*) 'Plane strain analysis'
        write(16,*)'Plane strain analysis'
      else if (mstyp.eq.3) then
        write(6,*) 'Axisymmetric analysis'
        write(16,*)'Axisymmetric analysis'
      else
        write(6,*)'Invalid input!'
        goto 3
      endif        
      write(16,*)
c
c...open a scratch file to write the element stiffness matrices on
      open(unit=18,status='scratch',form='unformatted')  
      write(6,*)'Scratch file opened ...'
      write(6,*)
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine STIFF...'
      call stiff(melem,mpoin,mnode,msets,mcomp,mstyp,
     1           nelem,npoin,lnods,lnodn,coord,lpros,props)
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine LOAD...'
      call load(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1          nelem,npoin,lnods,lnodn,coord,
     2          surft,aslod,lpros,props)
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine ASSMB...'
      call assmb(melem,mpoin,mnode,mtotv,mband,
     1           nelem,npoin,nhbw,lnods,lnodn,
     2           gstif,nminvc,diagk)
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine SOLVE...'
      call solve(mpoin,mtotv,mband,npoin,nkode,nhbw,
     1           aslod,asdis,astmp,gstif,nminvc,diagk)
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine RESULT...'
      call result(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1            nelem,npoin,lnods,lnodn,coord,nkode,lpros,props,
     2            aslod,asdis)
c
      write(6,*)
      write(6,*)'ELAST completed. Results are in file ',prtfil(1:lenin)
      write(6,*)
      write(6,*)
      call prompt
      close(15)
      close(16)
      close(17)
      stop
  999 write(6,*)
     1'Cannot open file with this name. Terminating program.'
      call prompt
      stop
      end
c
c
c      
      subroutine stiff(melem,mpoin,mnode,msets,mcomp,mstyp,
     1             nelem,npoin,lnods,lnodn,coord,lpros,props)
c      
c...calculates the element stiffness matrices, and writes them
c...to the scratch file.
c
      dimension props(msets,mcomp),lnods(melem,mnode),encod(2,8),
     1         estif(16,16),bmatx(4,16),dmatx(4,4),dbmat(4,16),
     2        gpwts(4),lpros(melem),lnodn(melem),coord(mpoin,2),
     3       shape(8),gploc(2,4),gpcod(2),deriv(2,8),cartd(2,8)
c...encod holds the nodal coords for the nodes around the current element
c...estif will be the stiffness matrix for the current element
c...bmatx, dmatx are the B and D matrices, and dbmat is D*B
c...gpwts,gploc are the Gauss point weights and locations (in xi,eta coords)
c...gpcod will be the x,y coords of the current Gauss point
c...shape holds the shape functions evaluated at the Gauss point
c...deriv holds the derivatives (w.r.t. xi, eta) of the shape functions
c...cartd holds the Cartesian derivatives (w.r.t x,y) of shape functions
c
      rewind 18
c
c...for axisymmetric analyses, D is a 4x4 matrix; otherwise it's 3x3
      ndmatx = 3
      if (mstyp.eq.3) ndmatx = 4 
c...and we need the value of pi, for integrating. Form it using inverse tangent
      pi = 4.*atan(1.0)
c...start loop over all elements
      do 80 l=1,nelem
c
c...extract Young's modulus & Poisson's ratio from props      
      lset = lpros(l)
      ym = props(lset,1)
      pr = props(lset,2)
c...also extract material thickness (default is 1.0) for plane stress/strain
      thick = props(lset,3)
      if (thick.eq.0.0) thick = 1.0
c
c...put nodal coords in encod
      nnode = lnodn(l)
      do 15 n=1,nnode
        nn = lnods(l,n)
        encod(1,n) = coord(nn,1)
        encod(2,n) = coord(nn,2)
   15 continue
c...zero stiffness matrix
      do 20 j=1,16
      do 20 i=1,16
        estif(i,j) = 0.0
   20 continue
c
c...get D matrix
      call dmat(mstyp,ym,pr,dmatx)
c
c...put local coords of 2x2 guass-pts in gploc
      call gauss(gploc,gpwts,ngaus)
c
c...start loop through gauss-pts for contribution to estif
      do 75 ij=1,ngaus
c...put local coords of gauss-pt in s,t
        s = gploc(1,ij)
        t = gploc(2,ij)
c...get shape fns (shape) & derivs (deriv) in local coords
        call sfr(s,t,shape,deriv)
c...get xy-coords of gauss-pt into gpcod, using shape fn interpolation
        do 25 m=1,2
          sum = 0.0
          do 24 n=1,8
   24     sum = sum + encod(m,n)*shape(n)
          gpcod(m) = sum
   25   continue
c...for axisymmetric analyses, we need the radius at the Gauss point
        if (mstyp.eq.3) radius = gpcod(1)
c...get xy-derivs (cartd) & derterminant of Jacobian (djacb)
        call jacob(encod,shape,deriv,cartd,djacb)
c...get B matrix 
        call bmat(cartd,mstyp,radius,shape,bmatx)
c...form matrix product D*B
        do 40 j=1,16
        do 40 i=1,ndmatx
          sum=0.0
          do 39 k=1,ndmatx
   39     sum = sum + dmatx(i,k)*bmatx(k,j)
          dbmat(i,j) = sum
   40   continue
c...integrate B^T *D*B over volume of element
        darea = gpwts(ij)*djacb
c...for axisymmetry, the 'thickness' is the circumference
        if (mstyp.eq.3) then
          dvolu = darea*2.*pi*radius
c...otherwise, get volume by multiplying by thickness
        else
          dvolu = darea*thick
        endif
c...B^T*D*B*dvolu = contribution to estif
        do 50 j=1,16
        do 50 i=1,16
          sum = 0.0
          do 49 k=1,ndmatx
   49     sum = sum + bmatx(k,i)*dbmat(k,j)
          estif(i,j) = estif(i,j) + sum*dvolu
   50   continue
   75 continue
c...end of gauss loop
c
c...write estif to file
      write (18) estif
      write(6,*)'Stiffness matrix for elt ',l,' written to file.'
   80 continue
c...end of element loop
c
      write(6,*)'Elt stiffness matrices written to file'
      endfile 18
c
      return
      end
c
c
      subroutine load(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1    nelem,npoin,lnods,lnodn,coord,surft,aslod,lpros,props)
c
c...compile global nodal load vector
c
      dimension aslod(mtotv),surft(msets,2),props(msets,mcomp),
     1     lnods(melem,mnode),press(2,3),npals(3),lpros(melem),
     2     coord(mpoin,2),gploc(2),shape(3),deriv(3),xjacm(2),      
     3     gpres(2),fcomp(2),encod(2,3),lnodn(melem)
c
      ntotv = npoin*2
c
      write(6,*)'Applied Loads'
      write(6,*)'-------------'
      write(6,*)
      write(16,*)'Applied Loads'
      write(16,*)'-------------'
      write(16,*)
c...initialize the global load vector
      do 10 i=1,mtotv
        aslod(i) = 0.0
   10 continue
c...we need the value of pi, for integrating in axisymmetry.
      if (mstyp.eq.3) pi = 4.*atan(1.0)
c
c...there are some dummy lines in the input data, which are read but ignored
      read(15,*,end=99)
c...npset>0 if there are surface tractions. nbset is not used.
      read (15,510) nbset,npset
  510 format(2i5)
c
c...nodal loads (ld=1, ignored)
      write(6,*)'Nodal loads:'
      write(6,*)'-----------'
      write(6,*)
      write(6,*)'  Node   x-component   y-component'
      write(16,*)'Nodal loads:'
      write(16,*)'-----------'
      write(16,*)
      write(16,*)'  Node   x-component   y-component'
      read(15,*)
   15 read(15,520) lcard,ld,np,fx,fy
  520 format(i1,i4,i5,2e10.3)
      if (np.ne.0) then
c...for axisymmetry, integrate load around element
        if (mstyp.eq.3) then
          radius = coord(np,1)
          circum = 2.*pi*radius
          fx = fx*circum
          fy = fy*circum
        endif
c...(Note that for a plane stress analysis where thickness is not 
c...equal to 1.0, the user must multiply line load magnitude by 
c...thickness, to get the point load value to apply.)
c...The i'th component of node np must be converted to global degree 
c...of freedom component
        nm = 2*(np - 1)
c...add nodal loads into aslod
        aslod(nm + 1) = aslod(nm + 1) + fx
        aslod(nm + 2) = aslod(nm + 2) + fy
        write(6,610) np,fx,fy
        write(16,610) np,fx,fy
  610   format(i5,2(4x,e10.3))
        if (lcard.eq.0) goto 15
      endif
      write(6,*)
      write(16,*)
c
c
c...surface tractions
      if (npset.gt.0) then
        read(15,*)
        write(6,*)'Surface Tractions'
        write(6,*)'-----------------'
        write(6,*)
        write(6,*)' set   normal cmpnt   tangential cmpnt' 
        write(16,*)'Surface Tractions'
        write(16,*)'-----------------'
        write(16,*)
        write(16,*)' set   normal cmpnt   tangential cmpnt' 
c
c...first read surface traction sets
   25   read(15,530) lcard,iset,i1,i2,pn,pt
  530   format(i1,i4,3x,2i2,8x,2e10.3)
        if (iset.gt.0) then 
          surft(iset,1) = pn
          surft(iset,2) = pt
          write(6,620)  iset,pn,pt
          write(16,620) iset,pn,pt
  620     format(i3,2(5x,e10.3))
        endif
        if (lcard.eq.0) goto 25
c
c...now apply these tractions to edges
        read(15,*)
        write(6,*)
        write(6,*)'   elt  end-nodes   set-numbers'
        write(16,*)
        write(16,*)'   elt   end-nodes   set-numbers'
   30   read(15,540) lcard,nnas,l,n1,iset1,n2,iset2
c...nnas is the number of nodes along the edge (=3 for 8-noded quad.)
  540   format(i1,i4,i5,2(i5,i2,1x))
        if (l.gt.0) then
          write(6,630) l,n1,n2,iset1,iset2
          write(16,630) l,n1,n2,iset1,iset2
  630     format(i5,2(2x,2i5))
c...put the tractions in press. Interpolate linearly for the
c...tractions at the midside node.
          do 35 i=1,2
            press(i,1) = surft(iset1,i)
            press(i,3) = surft(iset2,i)
            press(i,2) = (press(i,1) + press(i,3))/2.
   35     continue
c...store the node numbers along side in npals
          npals(1) = n1
          npals(3) = n2
c...we have to find the midside node no. from lnods
          nnode = lnodn(l)
          do 42 n=1,nnode
            if (lnods(l,n).eq.npals(1)) goto 45
   42     continue
c...if loop completed, npals(1) is not in element 1
          write(6,700) npals(1),l
  700     format('ERROR: node',i4,' is not in element ',i4,'.')
          call prompt
          stop
   45     k = n + 1
          npals(2) = lnods(l,k)
c
c...extract nodal coords and store in encod
          do 60 n=1,3
            np = npals(n)
            do 55 m=1,2
   55         encod(m,n) = coord(np,m)
   60     continue
c
c...extract material thickness, for integration of load over 
c...2D element edge (which is actually a face in 3D)
          lset = lpros(l)
          thick = props(lset,3)
          if (thick.eq.0.0) thick = 1.0
c...enter 1D Gauss loop along side (2-pt integration)
          r13 = 1.0/sqrt(3.0)
          gploc(1) = -r13
          gploc(2) = r13
          gpwts = 1.0
          do 90 ij=1,2
            s = gploc(ij)
            ss = s*s
            shape(1) = (ss-s)*0.5
            shape(2) = 1.0 - ss
            shape(3) = (ss + s)*0.5
            deriv(1) = s - 0.5
            deriv(2) = -2.0*s
            deriv(3) = s + 0.5
c...get jacobian xjacm
            do 75 j=1,2 
              sum = 0.0
              do 72 n=1,3
   72           sum = sum + deriv(n)*encod(j,n)
              xjacm(j) = sum
   75       continue
c...for axisymmetry, we need the radius, which is the x-coord of Gauss pt
            if (mstyp.eq.3) then
              sum = 0.0
              do 76 n=1,3
                sum = sum + encod(1,n)*shape(n)
   76         continue
              radius = sum
            endif
c...find gauss pt pressures, & integrate for components in local coords
            if (mstyp.eq.3) then
              darea = gpwts*2.*pi*radius
            else
              darea = gpwts*thick
            endif
            do 80 m=1,2
              sum = 0.0
              do 77 n=1,3
   77           sum = sum + press(m,n)*shape(n)
              gpres(m) = sum
   80       continue
            fcomp(1) = (xjacm(1)*gpres(2) - xjacm(2)*gpres(1))*darea
            fcomp(2) = (xjacm(1)*gpres(1) + xjacm(2)*gpres(2))*darea
c...accumulate equivalent nodal forces in aslod
            do 85 n=1,3
              np = npals(n)
              do 82 m=1,2
                nm = 2*(np - 1) + m
                aslod(nm) = aslod(nm) + fcomp(m)*shape(n)
   82         continue
   85       continue
   90     continue
c...end of Gauss loop
c
        endif
        if (lcard.eq.0) goto 30
c...go back for next loaded edge
      endif
c
c...report the non-zero loads assembled in aslod
      call prompt
      write(6,*) 
      write(6,*)'Non-zero entries in global load vector:'
      write(6,*)'  node  comp.     value'
      write(6,*)
      write(16,*) 
      write(16,*)'Non-zero entries in global load vector:'
      write(16,*)'  node  comp.     value'
      write(16,*)
      do 95 i=1,ntotv
        ai = aslod(i)
c...zero any tiny loads caused by rounding errors
        if (abs(ai).lt.1.e-5) then
          aslod(i) = 0.0
        else
          n = (i+1)/2
          m = i - 2*(n-1)
          write(6,650) n,m,ai
          write(16,650) n,m,ai
  650     format(i5,i5,5x,e10.3)
        endif
   95 continue
c
      write(6,*)
      write(6,*)'End of load data.'
      write(6,*)
      write(16,*)
      write(16,*)'End of load data.'
      write(16,*)
      return
   99 write(6,*)'ERROR: this input file has no loading data.'
      call prompt
      stop
      end
c      
c
      subroutine assmb(melem,mpoin,mnode,mtotv,mband,
     1      nelem,npoin,nhbw,lnods,lnodn,gstif,nminvc,diagk)
c
c...assembles the elt stiffness matrices, read from the scratch file,
c...into the global matrix gstif, stored compactly using its banded
c...structure. Also stores the diagonal of K in diagk
c
      dimension lnods(melem,mnode),estif(16,16),gstif(mtotv,mband),
     1         lnodn(melem),nminvc(mtotv),diagk(mtotv)
c
      ntotv = 2*npoin
c...first, a mock run-through of the assembly, to find the location of the
c...first non-zero element on each row i: store this in nminvc(i)
      do 2 i=1,ntotv
        nminvc(i) = i
    2 continue
c...loop through all elements
      do 25 l=1,nelem
        nnode = lnodn(l)
c...deal with elt nodes ni in rows, nj in columns
c... these will be global nodes nni, nnj
        ni=0
    5   ni = ni + 1
        i = 2*(ni - 1)
        nni = lnods(l,ni)
        ii = 2*(nni - 1)
        nj = 0
   10   nj = nj + 1
        j = 2*(nj - 1)
        nnj = lnods(l,nj)
        jj = 2*(nnj - 1)
c...now we could put the four entries from estif into gstif
        do 20 mi=1,2
          do 15 mj=1,2
            im = i + mi
            jm = j + mj
            iim = ii + mi
            jjm = jj + mj
            if (jjm.gt.iim) goto 15
c...ignore entries above diagonal in gstif
c...now we know that the (im,jm) entry of element l will
c...add into the (iim,jjm) position of K
            imin = nminvc(iim)
            if (jjm.lt.imin) imin = jjm
            nminvc(iim) = imin
   15     continue
   20   continue
        if (nj.lt.nnode) goto 10
        if (ni.lt.nnode) goto  5
   25 continue
c...end of element loop
c
c...now find halfbandwidth
      nhbw = 0
      do 30 i=1,ntotv
        idiff = i - nminvc(i) 
        if (idiff.gt.nhbw) nhbw = idiff
   30 continue
c
      write(6,*) ' Global matrix half-bandwidth = ', nhbw
      nband = nhbw + 1
      if (nband.gt.mband) then
        write(6,*)
     1   'ERROR: not enough storage. Increase parameter mband to '
     1            ,nband,' at start of program, and recompile.'
        call prompt
        stop
      endif
c
c...initialize gstif
      do 35 i=1,ntotv
      do 35 j=1,nband
        gstif(i,j) = 0.0
   35 continue
c
      rewind 18
c
c...loop through all elements
      do 80 l=1,nelem
        read (18) estif
c
        nnode = lnodn(l)
c...deal with elt nodes ni in rows, nj in columns
c... these will be global nodes nni, nnj
        ni=0
   55   ni = ni + 1
        i = 2*(ni - 1)
        nni = lnods(l,ni)
        ii = 2*(nni - 1)
        nj = 0
   60   nj = nj + 1
        j = 2*(nj - 1)
        nnj = lnods(l,nj)
        jj = 2*(nnj - 1)
c...put the four entries from estif into gstif
        do 70 mi=1,2
          do 65 mj=1,2
            im = i + mi
            jm = j + mj
            iim = ii + mi
            jjm = jj + mj
            if (jjm.gt.iim) goto 65
c...ignore entries above diagonal in gstif
c...translate column jjm to column of gstif with band storage
            jpos = jjm - iim + nhbw + 1
            gstif(iim,jpos) = gstif(iim,jpos) + estif(im,jm)
   65     continue
   70   continue
        if (nj.lt.nnode) goto 60
        if (ni.lt.nnode) goto 55
        write(6,*)'Stiffness matrix of element ',l,' assembled.'
   80 continue
c...end of element loop
c
      nhbw1 = nhbw + 1
      do 90 i=1,ntotv
        diagk(i) = gstif(i,nhbw1)
   90 continue
c
      return
      end
c
c
      subroutine solve(mpoin,mtotv,mband,npoin,nkode,
     1          nhbw,aslod,asdis,astmp,gstif,nminvc,diagk)
      dimension gstif(mtotv,mband),aslod(mtotv),asdis(mtotv),
     1    nkode(mpoin),nminvc(mtotv),astmp(mtotv),diagk(mtotv)
c
c...solve matrix system  gstif*asdis=aslod
c
      ntotv = 2*npoin
c...add 10e30 to diagonal elements for fixed dof 
      call fixdof(mtotv,mband,ntotv,nhbw,mpoin,npoin,gstif,
     1            nkode,diagk)
c...solution of equations by Choleski decomposition
      write(6,*)
      write(6,*)'Starting equation solution: decomposition...'
      call chodec(mtotv,mband,ntotv,nhbw,gstif)
      write(6,*)'Starting substitution...'
      call chosub(mtotv,mband,ntotv,nhbw,gstif,aslod,asdis,astmp)
c...calculate reactions at fixed dofs, put reactions into aslod
      write(6,*)'Starting calculation of reactions ...'
      call reactn(mtotv,mband,mpoin,ntotv,nhbw,npoin,
     1gstif,asdis,aslod,nkode,diagk)
      return
      end
c
c
      subroutine result(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1                  nelem,npoin,lnods,lnodn,coord,nkode,
     3                  lpros,props,aslod,asdis)
c
c...output the results to file
c
      dimension aslod(mtotv),props(msets,mcomp),
     1         lnods(melem,mnode),coord(mpoin,2),endis(16),encod(2,8),
     2        shape(8),deriv(2,8),cartd(2,8),dmatx(4,4),  
     3       bmatx(4,16),gploc(2,4),
     4      gpwts(4),gpsig(4),gpeps(4),gpcod(2),
     5      nkode(mpoin),asdis(mtotv),lnodn(melem),lpros(melem)
c...output displacements and reactions
c
c...in axisymmetry, there are 4 stress components, otherwise 3
      if (mstyp.eq.3) then
        nstres = 4
      else
        nstres = 3
      endif
c...output to screen (unit 6):
      write(6,*)
      write(6,*)'RESULTS'
      write(6,*)'======='
      write(6,*)
      write(6,610)
  610 format(//' nodal reactions and displacements'
     1/28('-')//3x,'node',
     212x,'x-reaction',5x,'y-reaction',9x,'x-displt',7x,'y-displt')
c...output to results file (unit 16):
      write(16,*)
      write(16,*)'RESULTS'
      write(16,*)'======='
      write(16,*)
      write(16,610)
c...and output to end of data file, for postprocessor:
      write(17,*)'  node  x-displt  y-displt '
      do 20 n=1,npoin  
        i = 2*(n - 1)
        dispx = asdis(i+1)
        dispy = asdis(i+2)
        reacx = aslod(i+1)
        reacy = aslod(i+2)
        write(17,612) n,dispx,dispy
  612   format(i5,2e10.3)
        if (mod(n,25).eq.0) call prompt
        write(6,615) n,reacx,reacy,dispx,dispy 
        write(16,615) n,reacx,reacy,dispx,dispy 
  615   format(3x,i3,11x,e10.3,5x,e10.3,10x,e10.3,5x,e10.3)    
   20 continue
c
      call prompt
c...element Gauss pt stressess
  620 format(//'Element Gauss pt. stresses:'/27('-')/)
      write(6,620)
      write(16,620)
      write(17,*)
     1' elt gp    x-coord   y-coord   sigma-x   sigma-y   tau-xy'
      do 100 l=1,nelem
        if (mod(l,3).eq.0) call prompt
        write(6,630) l
        write(16,630) l
  630   format(//' element',i5/1x,12('-')/)
        lset = lpros(l)
        nnode = lnodn(l)
        ym = props(lset,1)
        pr = props(lset,2)
c...store nodal coords, displts for elt in encod, endis
        do 25 n=1,nnode
          nn = lnods(l,n)
          encod(1,n) = coord(nn,1)
          encod(2,n) = coord(nn,2)
          i = 2*(n - 1)
          ii = 2*(nn - 1)
          endis(i+1) = asdis(ii + 1)
          endis(i+2) = asdis(ii + 2)
   25   continue
        call dmat(mstyp,ym,pr,dmatx)
c...find Gauss-pt stresses & strains
        write(6,640) (i,i=1,3)
        write(16,640) (i,i=1,3)
  640   format(/' gauss-pt stresses and strains'/6x,'x-coord',8x,
     1  'y-coord',10x,3('stress (',i1,')',5x))
        call gauss(gploc,gpwts,ngaus)
c...loop over each Gauss point
        do 80 ij=1,ngaus
          s = gploc(1,ij)
          t = gploc(2,ij)
          call sfr(s,t,shape,deriv)
          call jacob(encod,shape,deriv,cartd,djacb)
c...get Gauss point coords in gpcod
          do 27 m=1,2
            sum = 0.0
            do 26 n=1,8
   26       sum = sum + encod(m,n)*shape(n)
            gpcod(m) = sum
   27     continue
          if (mstyp.eq.3) radius = gpcod(1)
          call bmat(cartd,mstyp,radius,shape,bmatx)
c...find strains as bmatx*displts
          do 30 i=1,nstres
            sum = 0.0
            do 29 k=1,16
   29       sum = sum + bmatx(i,k)*endis(k)
            gpeps(i) = sum
   30     continue
c...stresses = dmatx*strains
          do 40 i=1,nstres
            sum = 0.0 
            do 39 k=1,nstres
   39       sum = sum + dmatx(i,k)*gpeps(k)
            gpsig(i) = sum
   40     continue
c...if a tensile strength has been input as mat. prop. 3, test:
          lset = lpros(l)
          sigt = props(lset,4)
          mkode = 0
          if (sigt.gt.0.0) then
c...find minor principal stress
            sx = gpsig(1)
            sy = gpsig(2)
            tau = gpsig(3)
            smean = 0.5*(sx + sy)
            sdev = 0.5*(sx - sy)
            sroot = sqrt(sdev*sdev + tau*tau)
            sig1 = smean + sroot
            sig3 = smean - sroot
            if (sig3.lt.-sigt) mkode = 1
          endif
c...only stresses printed; modify this to print strains too
          write(17,'(i5,i2,i3,2e10.3,4f10.5)') 
     1       l,ij,mkode,(gpcod(i),i=1,2),(gpsig(i),i=1,nstres)
          write(6,650) (gpcod(i),i=1,2),(gpsig(i),i=1,nstres)
          write(16,650) (gpcod(i),i=1,2),(gpsig(i),i=1,nstres)
  650     format(3x,2(e10.3,5x),3(5x,f10.5))
          if (mkode.ne.0) then
            write(6,*)'Gauss-pt yielded: minor stress =',sig3
            write(16,*)'Gauss-pt yielded: minor stress =',sig3
          endif
   80   continue
c...end of Gauss loop
  100 continue
c...end of element loop
c
      write(6,*)
      write(6,*)'END OF RESULTS'
      write(6,*)
      write(16,*)
      write(16,*)'END OF RESULTS'
      write(16,*)
      return
      end
c
c
c...subroutines for STIFF     
c
c
      subroutine gauss(gploc,gpwts,ngaus)
      dimension gploc(2,4),gpwts(4)
c...set up Gauss pts & weights for 2x2 integration
      ngaus = 4
      do 5 i=1,4
        gpwts(i) = 1.0
    5 continue
      r13 = 1.0/sqrt(3.0)
      gploc(1,1) = -r13
      gploc(1,2) = -r13
      gploc(1,3) = r13
      gploc(1,4) = r13
      gploc(2,1) = -r13
      gploc(2,2) = r13
      gploc(2,3) = -r13
      gploc(2,4) = r13
      return
      end
c
c
      subroutine dmat(mstyp,ym,pr,dmatx)
      dimension dmatx(4,4)
      do 5 i=1,4
        do 5 j=1,4
          dmatx(i,j) = 0.0
    5 continue
c...ym,pr are Young's mod., Poisson's ratio
c...g is the elastic shear modulus
      g = 0.5*ym/(1.0 + pr)
      if (mstyp.eq.1) then
c...form the D matrix for plane stress
        f = 0.5*ym/(1.0 - pr)
        dxx = f + g
        dxy = f - g
      else
c...form the D matrix for plane strain
        dxy = g*pr/(0.5 - pr)
        dxx = 2.*g + dxy
      endif
      dmatx(1,1) = dxx
      dmatx(1,2) = dxy
      dmatx(2,1) = dxy
      dmatx(2,2) = dxx
      dmatx(3,3) = g
      if (mstyp.eq.3) then
c...extra row and column, for axisymmetric D-matrix
        dmatx(1,4) = dxy
        dmatx(2,4) = dxy
        dmatx(4,1) = dxy
        dmatx(4,2) = dxy
        dmatx(4,4) = dxx
      endif
      return
      end
c
c
      subroutine bmat(cartd,mstyp,radius,shape,bmatx)
      dimension cartd(2,8),bmatx(4,16),shape(8)
c...form the B matrix
      do 5 j=1,16
        do 5 i=1,4
          bmatx(i,j) = 0.0
    5 continue
      j = 0
      do 10 n=1,8
        i = j + 1
        j = i + 1
        bmatx(1,i) = -cartd(1,n)
        bmatx(2,j) = -cartd(2,n)
        bmatx(3,i) = -cartd(2,n)
        bmatx(3,j) = -cartd(1,n)
        if (mstyp.eq.3) then
c...final row, for axisymmetry
          bmatx(4,i) = -shape(n)/radius
        endif
c...negative sign for 'compression +ve' sign convention
   10 continue
      return
      end
c
c
      subroutine sfr(s,t,shape,deriv)
      dimension shape(8),deriv(2,8)
c...evaluate the shape functions and 1st derivs at local coords s,t
      s2 = 2.0*s
      t2 = 2.0*t
      ss = s*s
      tt = t*t
      st = s*t
      sst = ss*t
      stt = s*tt
      st2 = 2.0*st
      shape(1) = (-1.0 + st + ss + tt - sst - stt)*0.25
      shape(2) = (1.0 - t - ss + sst)*0.5
      shape(3) = (-1.0 - st + ss + tt - sst + stt)*0.25
      shape(4) = (1.0 + s - tt - stt)*0.5
      shape(5) = (-1.0 + st + ss + tt + sst + stt)*0.25
      shape(6) = (1.0 + t - ss - sst)*0.5
      shape(7) = (-1.0 - st + ss + tt + sst - stt)*0.25
      shape(8) = (1.0 - s - tt + stt)*0.5
c     
      deriv(1,1) = (t + s2 - st2 - tt)*0.25
      deriv(1,2) = -s + st
      deriv(1,3) = (-t + s2 - st2 + tt)*0.25
      deriv(1,4) = (1.0 - tt)*0.5
      deriv(1,5) = (t + s2 + st2 + tt)*0.25
      deriv(1,6) = -s - st
      deriv(1,7) = (-t + s2 + st2 - tt)*0.25
      deriv(1,8) = (-1.0 + tt)*0.5
      deriv(2,1) = (s + t2 - ss - st2)*0.25
      deriv(2,2) = (-1.0 + ss)*0.5
      deriv(2,3) = (-s + t2 - ss + st2)*0.25
      deriv(2,4) = -t - st             
      deriv(2,5) = (s + t2 + ss + st2)*0.25
      deriv(2,6) = (1.0 - ss)*0.5
      deriv(2,7) = (-s + t2 + ss - st2)*0.25
      deriv(2,8) = -t + st
      return
      end
c
c     
      subroutine jacob(encod,shape,deriv,cartd,djacb)
      dimension shape(8),deriv(2,8),cartd(2,8),encod(2,8),xjacm(2,2),
     1         xjaci(2,2)
c...evaluate the Jacobian matrix, and its determinant
      do 5 j=1,2
      do 5 i=1,2
        sum = 0.0
        do 4 n=1,8
    4   sum = sum + deriv(i,n)*encod(j,n)
        xjacm(i,j) = sum
    5 continue
c...xjacm is the Jacobian matrix. find det & inverse by Cramer        
      djacb = xjacm(1,1)*xjacm(2,2) - xjacm(1,2)*xjacm(2,1)
      if (djacb.lt.1.0e-12) then
        write(6,*)'ERROR: determinant of Jacobian is -ve or zero'
        write(6,*)'Element is inside out, or nodes defined clockwise'
        call prompt
        stop
      endif
      xjaci(1,1) = xjacm(2,2)/djacb
      xjaci(2,2) = xjacm(1,1)/djacb
      xjaci(1,2) = -xjacm(1,2)/djacb
      xjaci(2,1) = -xjacm(2,1)/djacb
c...get Cartesian derivs of shape functions using J inverse
      do 50 i=1,2
      do 50 n=1,8
        sum = 0.0
        do 49 j=1,2
   49   sum = sum + xjaci(i,j)*deriv(j,n)
        cartd(i,n) = sum
   50 continue
      return
      end
c
c
c...subroutines for SOLVE
c
c
      subroutine fixdof(mtotv,mband,ntotv,nhbw,mpoin,npoin,gstif,
     1                  nkode,diagk)
      dimension gstif(mtotv,mband),kd(2),nkode(mpoin),diagk(mtotv)
c...add a large number to the diagonal coefft of fixed d.o.f.s
      do 20 n=1,npoin
        kode = nkode(n)
        kd(1) = kode/10
        kd(2) = kode - kd(1)*10
        do 10 m=1,2
          if (kd(m).eq.0) goto 10
          nposn = 2*(n - 1) + m
          gstif(nposn,nhbw+1) = gstif(nposn,nhbw+1) + 1.0e20
          diagk(nposn) = diagk(nposn) + 1.e20
   10   continue
   20 continue
      return
      end
c
c
      subroutine chodec(mtotv,mband,ntotv,nhbw,gstif)
      dimension gstif(mtotv,mband)
      double precision sum,gik,gjk,aij,clij,clik,clii,
     1cljj,aii,small,clii2
c...Choleski decomposition, using band storage 
c...K(i,j) is stored at gstif(i,j-i+nhbw+1) for j<=i
c
      small = 1.d-8
      nhbw1 = nhbw + 1
c...find the first diagonal entry
      i = 1
      aii = gstif(1,nhbw1)
      if (aii.lt.small) then
        write(6,*)'ERROR: zero or negative pivot in row ',i
        call prompt
        stop
      endif
      clii = sqrt(aii)
      gstif(1,nhbw1) = clii
c...decompose each row in turn
      do 20 i=2,ntotv
        jstart = max(i - nhbw,1)
c...work along the row
        do 10 j=jstart,i-1
          cljj = gstif(j,nhbw1)
          jipos = j - i + nhbw1
          aij = gstif(i,jipos)
          sum = 0.d0
          kstart=max(i-nhbw,j-nhbw,1)
          do 5 k=kstart,j-1
            kipos = k - i + nhbw1
            kjpos = k - j + nhbw1
            gik = gstif(i,kipos)
            gjk = gstif(j,kjpos)
c...next line done in double precision, for accuracy
            sum = sum + gik*gjk
    5     continue
          clij = (aij - sum)/cljj
          jipos = j - i + nhbw1
          gstif(i,jipos) = clij
   10   continue
c...finally find the diagonal entry on the row
        aii = gstif(i,nhbw1)
        sum = 0.d0
        kstart = max(i-nhbw,1)
        do 15 k=kstart,i-1
          kipos = k - i + nhbw1
          clik = gstif(i,kipos)
          sum = sum + clik*clik
   15   continue
        clii2 = aii - sum
        if (clii2.lt.small) then
          write(6,*)'ERROR: zero or negative pivot in row ',i
          call prompt
          stop
        endif
        clii = sqrt(clii2)
        gstif(i,nhbw1) = clii
   20 continue
      return
      end
c
c
      subroutine chosub(mtotv,mband,ntotv,nhbw,gstif,
     1                  aslod,asdis,astmp)
      dimension gstif(mtotv,mband),aslod(mtotv),asdis(mtotv)
     1         ,astmp(mtotv)
      double precision sum,bi,gik,ak,clii,yi,gki
c
c...forward substitution: solve Ly = b
c...store y in astmp
      astmp(1) = aslod(1)/gstif(1,nhbw+1)
      do 10 i=2,ntotv
        bi = aslod(i)
        sum = 0.d0
        kstart = max(i-nhbw,1)
        do 5 k=kstart,i-1
          kipos = k - i + nhbw + 1
          gik = gstif(i,kipos)
          ak = astmp(k)
          sum = sum + gik*ak
    5   continue
        clii = gstif(i,nhbw+1)
        astmp(i) = (bi - sum)/clii
   10 continue
c                                   T
c...back-substitution phase: solve L x = y
      asdis(ntotv) = astmp(ntotv)/gstif(ntotv,nhbw+1)
      do 20 i=ntotv-1,1,-1
        yi = astmp(i)
        sum = 0.d0
        kend = min(i+nhbw,ntotv)
        do 15 k=i+1,kend
c...next line only needed in skyline storage
c          if (k.gt.i+nhbw) goto 15
          ikpos = i - k + nhbw + 1
          gki = gstif(k,ikpos)
          ak = asdis(k)
          sum = sum + gki*ak
   15   continue
        clii = gstif(i,nhbw+1)
        asdis(i) = (yi - sum)/clii
   20 continue
      return
      end
c
c
      subroutine reactn(mtotv,mband,mpoin,ntotv,nhbw,npoin,
     1                  gstif,asdis,aslod,nkode,diagk)
c...calculate reactions, and store them in aslod.
c...zero the other entries in aslod.
      dimension gstif(mtotv,mband),asdis(mtotv),
     1    kd(2),nkode(mpoin),aslod(mtotv),diagk(mtotv)
c
      do 5 i=1,ntotv
        aslod(i) = 0.0
    5 continue
      do 20 nn=1,npoin
        kode = nkode(nn)
        kd(1) = kode/10
        kd(2) = kode - kd(1)*10
        do 10 m=1,2
          if(kd(m).eq.0) goto 10
          nposn = 2*(nn - 1) + m
          gnn = diagk(nposn)
c...gnn is the diagonal entry of K, which was modified in FIXDOF
          aslod(nposn) = -gnn*asdis(nposn)
          asdis(nposn) = 0.0
   10   continue
   20 continue
      return
      end
c
c
