c...poroelasticity "main engine"
c...for use with FELIPE version 3 - sample file is conslex2.dat
c
c
c...If you want to re-compile this program, you must link it with
c...the file felsb.obj, compiled from felsb.for (provided in the
c...FELIPE folder). The file felsb.for contains additional subroutines 
c...used by all the 'main engines' except eladv.for. 
c...If you don't use linking, you should paste the contents of 
c...the file felsb.for to the end of this file, before compiling
c...and running it.
c
      program consl
cp...program for poroelasticity (consolidation) problems.
cp...uses 8-noded quad elts where corner nodes have (u,v,p)
cp...degrees of freedom (p=pore pressure). Example of a 
cp...coupled problem. Theory: solve
cp    _                __               __ _
cp   (X + Y) u     =   th f  + f     - (th X  -  Y) u
cp            n+1          n    n+1                  n
cp __
cp th = (1.0-theta)/theta
cp
cp...Note: to make sign of pore pressures compatible with the
cp...compression-positive sign convention for u,f, -Q in
cp...standard theory is replaced by Q
cp...changes to elast.for are prefaced by cp... comment lines 
c
      parameter(melem=200,mpoin=400,mnode=8,
     1msets=8,mcomp=6,mtotv=1200,mstif=100000)
c...these parameters fix the dimensions of the arrays:
c...melem = max. no. of elements
c...mpoin = max. no. of nodes
c...mnode = max. no. of nodes in a single element
c...msets = max. no. of mat. prop. sets (also used for loading sets)
c...mcomp = max. no. of mat. props (e.g. Young's mod, Poisson ratio)
c...mtotv = max no. of degrees of freedom (corner nodes have 3 dofs)
c...mstif = max. length of vector gstif storing global matrix using
c                                               skyline storage
c
c
c
      character*1  char
      character*80 infile,outfil,prtfil,title
c
      dimension lnods(melem,mnode),lnodn(melem),lsidn(melem),
     1  coord(mpoin,2),nkode(mpoin),aslod(mtotv),asrhs(mtotv),
     2        lpros(melem),props(msets,mcomp),asdis(mtotv),
     4      surft(msets,2),gstif(mstif),lmtyp(melem)
     5  ,ndof(mpoin),nvar(mpoin),nnout(5)
     6      ,nminvc(mtotv),lpoint(mtotv),astmp(mtotv)
c...lnods(l,n) = global no. of n'th node of element l
c...lnodn(l) = no. of nodes in element l ( = 8 for 8-noded quad.)
c...lsidn = no. of sides in element l    ( = 4 for 8-noded quad.)
c...coord(np,1), coord(np,2) = x,y coords of node np
c...nkode(np) = fixity code of node np
c...aslod is global nodal load vector
c...asrhs is right-hand-side vector at each timestep
c...ndof(np) = no. of dof.s at node np 
cp..nvar(np) is a pointer array for global dof of node np
c...lpros(l) = mat. prop. set no. of element l
c...props(m,i) = value of i'th mat. property in mat. prop. set m
c...asdis is global nodal displacement/pressure vector in current timestep
c...surft(m,1),surft(m,2) = normal,tangential surface tractions in set m
c...gstif holds assembled global stiffness matrix in compact form
c...lmtyp(l) = material type of elt l (1 for structure, 2 for soil)
c...nminvc(i)=j means that G(i,j) is the first nonzero entry in row i of gstif
c...lpoint(i)=ii means that G(i,i) is stored at gstif(ii) in skyline storage
c
      write(6,*)'Program CONSL: soil consolidation problems.'
      write(6,*)'Handles 8-noded quadrilateral elements,'
      write(6,*)'with (u,v,t) d.o.f.s at corner nodes, and (u,v) at'
      write(6,*)'midside nodes. Time-dependent analysis. Writes'
      write(6,*)'output for undrained and drained solutions.'
      write(6,*)'Dimensioned for',melem,' elements, and',mpoin,' nodes.'
      write(6,*)  
c
c...read an existing .dat file
c
    1 continue
c
c...get the input file name in infile
      call askfil(infile,iflag)
      if (iflag.eq.1) then
        call yesno(char,'Do you want to exit?')
        if (char.eq.'y') stop
        goto 1
      endif
c
      write(6,*)'Opening file ',infile,'...'
      lenin = lenstr(infile)
      open(15,file=infile(1:lenin),status='old',err=999)
      write(6,*)
      lm4 = lenin - 4
      prtfil = infile(1:lm4)//'.prt'
      write(6,*)'Creating file ',prtfil(1:lenin),' for results...'
      write(6,*)
      open(16,file=prtfil(1:lenin),status='unknown',err=999)      
      outfil = infile(1:lm4)//'.out'
      write(6,*)'Creating file ',outfil(1:lenin),' for output...'
      write(6,*)
      open(17,file=outfil(1:lenin),status='unknown',err=999)      
c
c...read the input mesh data, and echo it to the output files
      call input(15,melem,mpoin,mnode,msets,mcomp,nelem,
     1 npoin,lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props
     2         ,ndofn,npros,title,ndof)
      call print(16,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1 lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2           npros,title,ndof)
      call output(17,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1  lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2            npros,title,ndof)
      call yesno(char,'Do you want the mesh details listed?')
      if (char.eq.'y') then
        call print(6,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1   lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2           npros,title,ndof)
      endif
c
      iflag = 0
      do 3 l=1,nelem
        nnode = lnodn(l)
        nside = lsidn(l)
        nodsid = 10*nnode + nside
        if (nodsid.ne.84) then
          iflag = 1
          write(6,*)'ERROR: Cannot handle element ',l
          write(6,*)'which has ',nside,' sides and ',nnode,' nodes'
        endif
    3 continue
      if (iflag.eq.1) then
        call prompt
        stop
      endif
c
cp...create the pointer array nvar. Needed because a node may
cp...have 2 or 3 d.o.f.s; this info stored in ndof
      nvar(1) = 0
      ntotv = 0
      do 5 np=1,npoin
        ndofnp = ndof(np)
        nvar(np) = ntotv
        ntotv = ntotv + ndofnp
    5 continue
cp...Now local dof i of node np is global dof nvar(np)+i
cp...and ntotv is total no. of degrees of freedom
c
c...analyse the "mask" of the global stiffness matrix, i.e. the range
c...of non-zero values in each row. This information is stored in
c...nminvc: row i starts in column nminvc(i). This is used to construct 
c...lpoint: g(i,i) will be located at gstif(lpoint(i))
c...this is a generalisation of the s/r in PLADV, to allow for variable 
c...d.o.f.s per node
c
      call mask(melem,mpoin,mnode,mtotv,mstif,ntotv,
     1       nelem,npoin,nhbw,lnods,lnodn,nminvc,lpoint,ndof,nvar)
c...doing this now, avoids repeating it within assmb at each iteration!
c
      write(6,*)'Type the numbers of 5 nodes '
      write(6,*)'for which you would like the displacements at each'
      write(6,*)'timestep to be output:'
      read(5,*) (nnout(i),i=1,5)
c
c...open a scratch file to write the element stiffness matrices on
      open(unit=18,status='scratch',form='unformatted')  
      write(6,*)'Scratch file opened ...'
      write(6,*)
c
cp...STIFF forms the X and Y matrices on file 18 for each element
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine STIFF...'
      call stiff(melem,mpoin,mnode,msets,mcomp,
     1        nelem,npoin,lnods,lnodn,lsidn,coord,lpros,props
     2           ,lmtyp,ndof,nvar)
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine LOAD...'
      call load(melem,mpoin,mnode,msets,mcomp,mtotv,
     1          nelem,npoin,lnods,lnodn,coord,
     2          surft,aslod,lmtyp,ndof,nvar)
c
c...read data for boundaries (should be blank)
      read(15,*)
      read(15,'(2i5)')nbrad,nbdir
c
      if (nbrad.gt.0) then
        write(6,*) 'ERROR: nbrad > 0.'
        write(6,*)'This program cannot handle radiating boundaries.'
        call prompt
        stop
      endif
c
      if (nbdir.gt.0) then
        write(6,*) 'ERROR: nbdir > 0.'
        write(6,*)'This program cannot handle Dirichlet boundaries.'
cp...Permeable boundaries (where the excess pore pressure p is fixed 
cp...at zero) are handled in the same way as fixed displacement dof's,
cp...via the fixity code only.
        call prompt 
        stop
      endif
c
c...read data defining load increments (not expecting any)
      read(15,*) 
      read(15,'(i5)') nincs
      if (nincs.gt.0) then
        write(6,*)
     1  'Error: cannot handle load increments - data ignored.'
      endif
c
cv...read the timestepping control parameters
      read(15,*)
      read(15,*) dtint,ftime,allow,theta,ttend,dtone
cv...theta is the time discretization parameter: =0 for explicit algorithm,
cv...                    between 0.5 and 1.0 for stable implicit algorithm.
cv...dtone is the small timestep used for the initial solution
cv...dtint is the initial timestep length in timestepping
cv...if no new g.pts. have yielded, next timestep is FTIME times previous one
cv...                    (DNEXT = FTIME*DTIME)
cv...max. timestep is ALLOW times initial one (ALLOW*DTINT)
cp...stop when time reaches TTEND
c
      tht = (1.0-theta)/theta
c
cv...initialize the timestepping
      ttime = 0.0
      dtime = dtone
c
c...form r.h.s. vector from initial load aslod. We take f_0 = 0 and 
c...f_1 = aslod. So r.h.s. is tht*f_0 + f_1 = f_1
c...There is no contibution from v_0, which is zero.
      do 20 i=1,ntotv
        asrhs(i) = aslod(i)
   20 continue
c
      write(6,*)'Undrained solution'
      write(6,*)'------------------'
      write(6,*)
      write(6,*)'Undrained timestep = ',dtone
      write(6,*)
      write(6,*)'Starting subroutine ASSMB...'
      call assmb(melem,mpoin,mnode,mtotv,mstif,
     1           nelem,npoin,lnods,lnodn,
     2           gstif,lmtyp,ndof,nvar,theta,dtime,nminvc,lpoint)
c...add 10e12 to diagonal elements for fixed dof (including
cp...pore pressure dof's on permeable boundaries).
      write(6,*)'Starting subroutine FIXDOF...'
      call fixdof(mtotv,mstif,ntotv,mpoin,npoin,gstif,nkode,coord,
     1         asrhs,ndof,nvar,lpoint)
c...                              T
c...solution of equations by L D L   decomposition
      write(6,*)
      write(6,*)'Starting matrix decomposition in LDLDEC...'
      call ldldec(mtotv,mstif,ntotv,gstif,lpoint)
c...solve 
      write(6,*)
      write(6,*)'Starting equation substitution in LDLSUB...'
      call ldlsub(mtotv,mstif,ntotv,gstif,asrhs,asdis,astmp,
     1            lpoint)
c
      ttime = ttime + dtime
      write(6,*)
      write(6,*)'Time = ',ttime
      write(6,*)
c
      call result(melem,mpoin,mnode,msets,mcomp,mtotv,
     1   nelem,npoin,lnods,lnodn,lsidn,coord,nkode,lpros,props,
     2            aslod,asdis,lmtyp,ndof,nvar)
c
      kount = 0
      dtime = dtint
      write(6,*)'Starting timestepping, with timestep = ',dtime
      write(6,*)
c
      write(6,*)'Forming new right-hand side vector...'
cv...now form new r.h.s. for next timestep in asrhs
c...now f_1 = aslod and f_2 = aslod, so rhs = tht*f_1 + f_2
      do 30 i=1,ntotv
        asrhs(i) = tht*aslod(i) + aslod(i)
   30 continue
      call newrhs(melem,mpoin,mnode,mtotv,
     1                 nelem,npoin,lnods,lnodn,nkode,
     2                 asrhs,asdis,lmtyp,ndof,nvar,theta,dtime)

c
cc..start of global timestepping 
cc..============================
  100   continue
        kount = kount + 1
c
        write(6,*)'Assembling...'
        call assmb(melem,mpoin,mnode,mtotv,mstif,
     1           nelem,npoin,lnods,lnodn,
     2           gstif,lmtyp,ndof,nvar,theta,dtime,nminvc,lpoint)
        write(6,*)'Applying fixities...' 
        call fixdof(mtotv,mstif,ntotv,mpoin,npoin,gstif,nkode,
     1   coord,asrhs,ndof,nvar,lpoint)
        write(6,*)'Decomposing matrix...'
        call ldldec(mtotv,mstif,ntotv,gstif,lpoint)
c
cp...perform 5 timesteps using the timestep dtime
        do 180 kk=1,5
c
cp...solve - no need to re-assemble or decompose, as if dtime is 
cp...unchanged then stiffness matrix will be unchanged 
          write(6,*)'Performing substitutions...'
          call ldlsub(mtotv,mstif,ntotv,gstif,asrhs,asdis,astmp,
     1            lpoint)
c
          ttime = ttime + dtime
          write(6,*)
          write(6,*)'Time = ',ttime
          write(6,*)
c
          do 170 i=1,5
cp...output results at selected nodes
            nn = nnout(i)
            ndofnn = ndof(nn)
            nvarnn = nvar(nn)
            write(6,*)'Displts/pressure at node ',nn
            dispx = asdis(nvarnn+1)
            dispy = asdis(nvarnn+2)
            if (ndofnn.eq.3) then
              pore = asdis(nvarnn+3)
              write(6,610) dispx,dispy,pore
            else
              write(6,610) dispx,dispy
            endif
  610     format(3(5x,e12.5))
  170     continue
          call prompt
c
c...form new rhs vector (if we will not be changing the timestep)
          if (kk.lt.5) then
            write(6,*)'Forming new right-hand side vector...'
c...both f_n and f_n+1 are = aslod in subsequent timesteps
            do 175 i=1,ntotv
              asrhs(i) = tht*aslod(i) + aslod(i)
  175       continue
            call newrhs(melem,mpoin,mnode,mtotv,
     1             nelem,npoin,lnods,lnodn,nkode,
     2             asrhs,asdis,lmtyp,ndof,nvar,theta,dtime)
          endif
  180   continue
c
        call yesno(char,'Do you want to stop?')
        if (char.eq.'y') goto 200
c
c...output results to file
      write(6,665)  ttime
      write(16,665) ttime
      write(17,665) ttime
  665 format('Consolidation progressing: time = ',e10.3)
c
      call result(melem,mpoin,mnode,msets,mcomp,mtotv,
     1   nelem,npoin,lnods,lnodn,lsidn,coord,nkode,lpros,props,
     2            aslod,asdis,lmtyp,ndof,nvar)
c
cv...set next timestep 
        dnext = ftime*dtime
        if (dnext.gt.allow*dtint) dnext = allow*dtint
        dtime = dnext
        write(6,*)
        write(6,*)'Changing timestep length to ',dtime
        write(6,*)
c
cv...now form new r.h.s. for next timestep in asrhs
        write(6,*)'Forming new right-hand side vector...'
        do 185 i=1,ntotv
          asrhs(i) = tht*aslod(i) + aslod(i)
  185   continue
        call newrhs(melem,mpoin,mnode,mtotv,
     1             nelem,npoin,lnods,lnodn,nkode,
     2             asrhs,asdis,lmtyp,ndof,nvar,theta,dtime)
c
        if (ttime.lt.ttend) goto 100
c
cc..end of timestepping
cc..===================
c
  200 write(6,*)'END OF TIMESTEPPING'
      write(6,*)
      write(6,*) kount,' global timesteps'
      write(6,*)
c
c
      write(6,666) ttime
      write(16,666) ttime
      write(17,666) ttime
  666 format('Drained solution. Time = ',e10.3)
c
      call result(melem,mpoin,mnode,msets,mcomp,mtotv,
     1   nelem,npoin,lnods,lnodn,lsidn,coord,nkode,lpros,props,
     2            aslod,asdis,lmtyp,ndof,nvar)
c
      write(6,*)
      write(6,*)'POREL completed. Results are in file ',prtfil(1:lenin)
      write(6,*)
      write(6,*)
      call prompt
      stop
  999 write(6,*)
     1'Cannot open file with this name. Terminating program.'
      call prompt
      stop
      end
c
c
c      
c
      subroutine mask(melem,mpoin,mnode,mtotv,mstif,ntotv,
     1       nelem,npoin,nhbw,lnods,lnodn,nminvc,lpoint,ndof,nvar)
c
c...determines the "mask" of gstif, i.e. the location of the nonzero 
c...section of each row. This is stored in nminvc and lpoint
c
      dimension lnods(melem,mnode),ndof(mpoin),nvar(mpoin),
     1         lnodn(melem),nminvc(mtotv),lpoint(mtotv)
c
      ntotv = nvar(npoin) + ndof(npoin)
c...first, a mock run-through of the assembly, to find the location of the
c...first non-zero element on each row i: store this in nminvc(i)
      do 2 i=1,ntotv
        nminvc(i) = i
    2 continue
c...loop through all elements
      do 25 l=1,nelem
        nnode = lnodn(l)
c...deal with elt nodes ni in rows, nj in columns
c... these will be global nodes nni, nnj
c...note this allows for variable d.o.f.s per node
        ni=0
        i  = 0
    5   ni = ni + 1 
        nni = lnods(l,ni)
        idof = ndof(nni)
        ii = nvar(nni)
        nj = 0
        j  = 0
   10   nj = nj + 1
        nnj = lnods(l,nj)
        jdof = ndof(nnj)
        jj = nvar(nnj)
c...now we could put the entries from estif into gstif
        do 20 mi=1,idof
          do 15 mj=1,jdof
            im = i + mi
            jm = j + mj
            iim = ii + mi
            jjm = jj + mj
            if (jjm.gt.iim) goto 15
c...ignore entries above diagonal in gstif
c...now we know that the (im,jm) entry of element l will
c...add into the (iim,jjm) position of K
            imin = nminvc(iim)
            if (jjm.lt.imin) imin = jjm
            nminvc(iim) = imin
   15     continue
   20   continue
        j = j + jdof
        if (nj.lt.nnode) goto 10
        i = i + idof
        if (ni.lt.nnode) goto  5
   25 continue
c...end of element loop
c
c...now find halfbandwidth, and fill in lpoint
      nhbw = 0
      idiag = 0
      do 30 i=1,ntotv
        idiff = i - nminvc(i) 
        if (idiff.gt.nhbw) nhbw = idiff
        idiag = idiag + idiff + 1
        lpoint(i) = idiag 
   30 continue
c...in skyline storage, G(i,i) is stored at gstif(lpoint(i))
c
      write(6,*) ' Global matrix half-bandwidth = ', nhbw
      nstif = idiag
      write(6,*)' Length of skyline storage vector gstif is',nstif
      if (nstif.gt.mstif) then
        write(6,*)
     1   'ERROR: not enough storage. Increase parameter mstif to '
     1            ,nstif,' at start of program, and recompile.'
        call prompt
        stop
      endif
c
      return
      end
c
c
c      
      subroutine stiff(melem,mpoin,mnode,msets,mcomp,
     1    nelem,npoin,lnods,lnodn,lsidn,coord,lpros,props
     2             ,lmtyp,ndof,nvar)
c      
c...calculates the element stiffness matrices, and writes them
c...to the scratch file.
c
      dimension props(msets,mcomp),lnods(melem,mnode),encod(2,8),
     1         estif(16,16),bmatx(3,16),dmatx(3,3),dbmat(3,16),
     2        gpwts(4),lpros(melem),lnodn(melem),coord(mpoin,2),
     3       shape(8),gploc(2,4),gpcod(2),deriv(2,8),cartd(2,8),
     4      lmtyp(melem),ndof(mpoin),nvar(mpoin),qmatx(4,16),
     5     smatx(4,4),xmatx(20,20),encodp(2,8),lsidn(melem),
     6    perm(2),ymatx(20,20),hmatx(4,4)
c
cp...for soil elements, stiffness matrices are
cp
cp       /         T  \  Form K in estif(16,16)    
cp       |  K    -Q   |  Form Q in qmatx(4,16)
cp  X =  |            |  Form H in hmatx(4,4)
cp       |  0    -H   |  Assemble whole matrix in xmatx(20,20)
cp       \            /  Write xmatx to file 
cp
cp
cp       /            \     
cp       |  0     0   |  Form Q in qmatx(4,16)
cp  Y =  |            |  Form S in smatx(4,4)
cp       | -Q    -S   |  Assemble whole matrix in ymatx(20,20)
cp       \            /  Write ymatx to file 
cp
cp...for structural elements (lmtyp=1), just form K in estif
cp...and write estif to file.
c
c...encod holds the nodal coords for the nodes around the current element
c...estif will be the stiffness matrix for the current element
c...bmatx, dmatx are the B and D matrices, and dbmat is D*B
c...gpwts,gploc are the Gauss point weights and locations (in xi,eta coords)
c...gpcod will be the x,y coords of the current Gauss point
c...shape holds the shape functions evaluated at the Gauss point
c...deriv holds the derivatives (w.r.t. xi, eta) of the shape functions
c...cartd holds the Cartesian derivatives (w.r.t x,y) of shape functions
c
      rewind 18
c
c...start loop over all elements
      do 80 l=1,nelem
c
c...extract Young's modulus & Poisson's ratio from props      
        lset = lpros(l)
        ym = props(lset,1)
        pr = props(lset,2)
cp...get elt material type: 1 for structure, 2 for soil
        lmtype = lmtyp(l)
        if (lmtype.eq.2) then
          perm(1) = props(lset,3)
          perm(2) = props(lset,4)
          poros = props(lset,5)
        endif
cp...poros is porosity: actually eta/K_f (K_f is bulk fluid modulus)
cp...perm are permeabilities: actually k/gammaw (water density)
c
c...put nodal coords in encod
        nnode = lnodn(l)
        nside = lsidn(l)
        do 15 n=1,nnode
          nn = lnods(l,n)
          encod(1,n) = coord(nn,1)
          encod(2,n) = coord(nn,2)
   15   continue
c...zero stiffness matrices
        do 16 j=1,4
          do 16 i=1,4
            hmatx(i,j) = 0.0
            smatx(i,j) = 0.0
   16   continue
        do 20 j=1,16
          do 17 i=1,4
            qmatx(i,j) = 0.0
   17     continue
          do 20 i=1,16
            estif(i,j) = 0.0
   20   continue
        do 21 i=1,20
          do 21 j=1,20
            xmatx(i,j) = 0.0
            ymatx(i,j) = 0.0
   21   continue  
c
cp...set up dimensions, to help generalisation to other elt types
        nevab = 2*nnode
c
cp...first form K in estif, whatever the elt material type
c
c...get D matrix
        call dmat(ym,pr,dmatx)
c
c...put local coords of 2x2 guass-pts in gploc
        call gauss(gploc,gpwts,ngaus)
c
c...start loop through gauss-pts for contribution to estif
        do 75 ij=1,ngaus
c...put local coords of gauss-pt in s,t
          s = gploc(1,ij)
          t = gploc(2,ij)
c...get shape fns (shape) & derivs (deriv) in local coords
          call sfr(nnode,nside,s,t,shape,deriv)
c...get xy-coords of gauss-pt into gpcod, using shape fn interpolation
          do 25 m=1,2
            sum = 0.0
            do 24 n=1,nnode
   24       sum = sum + encod(m,n)*shape(n)
            gpcod(m) = sum
   25     continue
c...get xy-derivs (cartd) & determinant of Jacobian (djacb)
          call jacob(nnode,encod,shape,deriv,cartd,djacb)
c...get B matrix 
          call bmat(cartd,bmatx)
c...form matrix product D*B
          do 40 j=1,16
            do 40 i=1,3
              sum=0.0
              do 39 k=1,3
   39         sum = sum + dmatx(i,k)*bmatx(k,j)
              dbmat(i,j) = sum
   40     continue
c...B transpose *d*b* djacb = contribution to estif
        dvolu = djacb*gpwts(ij)
        do 50 j=1,16
        do 50 i=1,16
          sum = 0.0
          do 49 k=1,3
   49     sum = sum + bmatx(k,i)*dbmat(k,j)
          estif(i,j) = estif(i,j) + sum*dvolu
   50   continue
        if (lmtype.eq.1) goto 75
c
cp...now form qmatx, hmatx and smatx, for soil elements
cp...get shape fns and derivs for a linear quad elt 
cp...based on the four corner nodes (nodes 1,3,5,7)
          nnodep = 4
          nsvab = nevab + nnodep
cp...nnodep is no. of nodes with porepressure dof
cp...nsvab is total no of variables (i.e. 16+4)
          do 52 i=1,2
            encodp(i,1) = encod(i,1) 
            encodp(i,2) = encod(i,3) 
            encodp(i,3) = encod(i,5) 
            encodp(i,4) = encod(i,7) 
   52     continue
          call sfr(nnodep,nside,s,t,shape,deriv)
          call jacob(nnodep,encodp,shape,deriv,cartd,djacb)
cp...note: B matrix and dvolu are unchanged 
cp
cp                       T  T
cp...form Q matrix as   N  M  B  
cp...where M selects only direct stresses (so k=1,2 below) 
          do 55 j=1,nnodep
          do 55 i=1,nevab
            sum=0.0
            do 54 k=1,2
   54         sum = sum + bmatx(k,i)*shape(j)
            qmatx(j,i) = qmatx(j,i) + sum*dvolu
   55     continue
cp                                T
cp...now form smatx as   poros * N  N
          do 60 j=1,nnodep
          do 60 i=1,nnodep
            sums = 0.0
            sumh = 0.0
            do 59 k=1,2
              sumh = sumh + cartd(k,i)*cartd(k,j)*perm(k)
              sums = sums + shape(i)*shape(j)*poros
   59       continue
            hmatx(i,j) = hmatx(i,j) + sumh*dvolu
            smatx(i,j) = smatx(i,j) + sums*dvolu
   60     continue
cp
   75 continue
c...end of gauss loop
c
      if (lmtype.eq.1) then
        do 77 i=1,16
          do 77 j=1,16
            xmatx(i,j) = estif(i,j)
   77   continue
      else
cp...for soil matrices, assemble xmatx from K, L and S
cp...arrange so that pore pressure dof of corner nodes
cp...comes after displacement dofs for that node
        jcol = 0
        np   = 0
        do 155 n=1,nnode
          nn = lnods(l,n)
          ndofnn = ndof(nn)
          jcol1 = jcol + 1
          jend = jcol + 2
c...top left submatrix
          do 140 j=jcol1,jend
            jmnp = j - np
            do 135 i=1,nevab
              xmatx(i,j) = estif(i,jmnp) 
  135       continue
c...bottom left submatrix
            nevab1 = nevab + 1
            do 140 i=nevab1,nsvab
              imnvb = i - nevab
              ymatx(i,j) = qmatx(imnvb,jmnp)
  140     continue
          jcol = jend
          if (ndofnn.eq.3) then
            np = np + 1
            jcol = jcol + 1
cp...top right submatrix
            do 145 i=1,nevab
              xmatx(i,jcol) = qmatx(np,i)
  145       continue
cp...bottom right submatrix
            do 150 i=nevab1,nsvab
              imnvb = i - nevab
              ymatx(i,jcol) =   smatx(imnvb,np)
              xmatx(i,jcol) =  -hmatx(imnvb,np)
  150       continue
          endif
  155   continue
cp...now swap rows around correspondingly
        np = 0
        irow = 0
        do 170 n=1,nnode
          nn = lnods(l,n)
          ndofnn = ndof(nn)
          irow = irow + ndofnn
          if (ndofnn.eq.3) then
            irow = irow - 1
            np = np + 1
            do 165 j=1,nsvab
              nvbnp = nevab + np
              tempx = xmatx(nvbnp,j)
              tempy = ymatx(nvbnp,j)
              i = nevab + np - 1
  160         continue
              xmatx(i+1,j) = xmatx(i,j)
              ymatx(i+1,j) = ymatx(i,j)
              i = i - 1
              if (i.ge.irow+1) goto 160
              xmatx(irow+1,j) = tempx
              ymatx(irow+1,j) = tempy
  165       continue
            irow = irow + 1
          endif
  170   continue
      endif
cp...end of formation of xmatx, ymatx
c
c...write stiffness matrices to file
      write(18) xmatx
      write(18) ymatx
c
      write(6,*)'Stiffness matrices for elt ',l,' written to file.'
c
   80 continue
c...end of element loop
c
      write(6,*)'Elt stiffness matrices written to file'
      endfile 18
c
      return
      end
c
c
      subroutine load(melem,mpoin,mnode,msets,mcomp,mtotv,
     1                nelem,npoin,lnods,lnodn,coord,surft,
     2                aslod,lmtyp,ndof,nvar)
c
c...compile global nodal load vector
c
      dimension aslod(mtotv),surft(msets,2),
     1     lnods(melem,mnode),press(2,3),npals(3),
     2     coord(mpoin,2),gploc(2),shape(3),deriv(3),xjacm(2),      
     3     gpres(2),fcomp(2),encod(2,3),lnodn(melem)
     4     ,lmtyp(melem),ndof(mpoin),nvar(mpoin)
c
      ntotv = npoin*2
c
      write(6,*)'Applied Loads'
      write(6,*)'-------------'
      write(6,*)
      write(16,*)'Applied Loads'
      write(16,*)'-------------'
      write(16,*)
c...initialize the global load vector
      do 10 i=1,mtotv
        aslod(i) = 0.0
   10 continue
c
c...there are some dummy lines in the input data, which are read but ignored
      read(15,*,end=99)
c...npset>0 if there are surface tractions. nbset is not used.
      read (15,510) nbset,npset
  510 format(2i5)
c
c...nodal loads (ld=1, ignored)
      write(6,*)'Nodal loads:'
      write(6,*)'-----------'
      write(6,*)
      write(6,*)'  Node   x-component   y-component'
      write(16,*)'Nodal loads:'
      write(16,*)'-----------'
      write(16,*)
      write(16,*)'  Node   x-component   y-component'
      read(15,*)
   15 read(15,520) lcard,ld,np,fx,fy
  520 format(i1,i4,i5,2e10.3)
      if (np.ne.0) then
c...the i'th component of node np must be converted to global degree 
c...of freedom component
cp...use nvar for this!
        nm = nvar(np)
c...add nodal loads into aslod
        aslod(nm + 1) = aslod(nm + 1) + fx
        aslod(nm + 2) = aslod(nm + 2) + fy
        write(6,610) np,fx,fy
        write(16,610) np,fx,fy
  610   format(i5,2(4x,e10.3))
        if (lcard.eq.0) goto 15
      endif
      write(6,*)
      write(16,*)
c
c
c...surface tractions
      if (npset.gt.0) then
        read(15,*)
        write(6,*)'Surface Tractions'
        write(6,*)'-----------------'
        write(6,*)
        write(6,*)' set   normal cmpnt   tangential cmpnt' 
        write(16,*)'Surface Tractions'
        write(16,*)'-----------------'
        write(16,*)
        write(16,*)' set   normal cmpnt   tangential cmpnt' 
c
c...first read surface traction sets
   25   read(15,530) lcard,iset,i1,i2,pn,pt
  530   format(i1,i4,3x,2i2,8x,2e10.3)
        if (iset.gt.0) then 
          surft(iset,1) = pn
          surft(iset,2) = pt
          write(6,620)  iset,pn,pt
          write(16,620) iset,pn,pt
  620     format(i3,2(5x,e10.3))
        endif
        if (lcard.eq.0) goto 25
c
c...now apply these tractions to edges
        read(15,*)
        write(6,*)
        write(6,*)'   elt  end-nodes   set-numbers'
        write(16,*)
        write(16,*)'   elt   end-nodes   set-numbers'
   30   read(15,540) lcard,nnas,l,n1,iset1,n2,iset2
c...nnas is the number of nodes along the edge (=3 for 8-noded quad.)
  540   format(i1,i4,i5,2(i5,i2,1x))
        if (l.gt.0) then
          write(6,630) l,n1,n2,iset1,iset2
          write(16,630) l,n1,n2,iset1,iset2
  630     format(i5,2(2x,2i5))
c...put the tractions in press. Interpolate linearly for the
c...tractions at the midside node.
          do 35 i=1,2
            press(i,1) = surft(iset1,i)
            press(i,3) = surft(iset2,i)
            press(i,2) = (press(i,1) + press(i,3))/2.
   35     continue
c...store the node numbers along side in npals
          npals(1) = n1
          npals(3) = n2
c...we have to find the midside node no. from lnods
          nnode = lnodn(l)
          do 42 n=1,nnode
            if (lnods(l,n).eq.npals(1)) goto 45
   42     continue
c...if loop completed, npals(1) is not in element 1
          write(6,700) npals(1),l
  700     format('ERROR: node',i4,' is not in element ',i4,'.')
          call prompt
          stop
   45     k = n + 1
          npals(2) = lnods(l,k)
c
c...extract nodal coords and store in encod
          do 60 n=1,3
            np = npals(n)
            do 55 m=1,2
   55         encod(m,n) = coord(np,m)
   60     continue
c
c...enter 1D Gauss loop along side (2-pt integration)
          r13 = 1.0/sqrt(3.0)
          gploc(1) = -r13
          gploc(2) = r13
          gpwts = 1.0
          do 90 ij=1,2
            s = gploc(ij)
            ss = s*s
            shape(1) = (ss-s)*0.5
            shape(2) = 1.0 - ss
            shape(3) = (ss + s)*0.5
            deriv(1) = s - 0.5
            deriv(2) = -2.0*s
            deriv(3) = s + 0.5
c...get jacobian xjacm
            do 75 j=1,2 
              sum = 0.0
              do 72 n=1,3
   72           sum = sum + deriv(n)*encod(j,n)
              xjacm(j) = sum
   75       continue
c...find gauss pt pressures, & convert to components in local coords
            do 80 m=1,2
              sum = 0.0
              do 77 n=1,3
   77           sum = sum + press(m,n)*shape(n)
              gpres(m) = sum
   80       continue
            fcomp(1) = (xjacm(1)*gpres(2) - xjacm(2)*gpres(1))*gpwts
            fcomp(2) = (xjacm(1)*gpres(1) + xjacm(2)*gpres(2))*gpwts
c...accumulate equivalent nodal forces in aslod
            do 85 n=1,3
              np = npals(n)
cp...use nvar to identify global dof's
              nvnp = nvar(np)
              do 82 m=1,2
                nm = nvnp + m
                aslod(nm) = aslod(nm) + fcomp(m)*shape(n)
   82         continue
   85       continue
   90     continue
c...end of Gauss loop
c
        endif
        if (lcard.eq.0) goto 30
c...go back for next loaded edge
c
c...report the non-zero loads assembled in aslod
        call prompt
        write(6,*) 
        write(6,*)'Non-zero entries in global load vector:'
        write(6,*)' d.o.f.     value'
        write(6,*)
        write(16,*) 
        write(16,*)'Non-zero entries in global load vector:'
        write(16,*)' d.o.f.     value'
        write(16,*)
        do 95 i=1,ntotv
          ai = aslod(i)
c...zero any tiny loads caused by rounding errors
          if (abs(ai).lt.1.e-5) then
            aslod(i) = 0.0
          else
            write(6,650) i,ai
            write(16,650) i,ai
  650       format(i5,5x,e10.3)
          endif
  95    continue
      endif
c
      write(6,*)
      write(6,*)'End of load data.'
      write(6,*)
      write(16,*)
      write(16,*)'End of load data.'
      write(16,*)
      return
   99 write(6,*)'ERROR: this input file has no loading data.'
      call prompt
      stop
      end
c      
c
      subroutine assmb(melem,mpoin,mnode,mtotv,mstif,
     1                 nelem,npoin,lnods,lnodn,
     2       gstif,lmtyp,ndof,nvar,theta,dtime,nminvc,lpoint)
c
c...assembles the elt stiffness matrices, read from the scratch file,
c...into the global matrix gstif, stored compactly using its banded
c...structure
c
      dimension lnods(melem,mnode),gstif(mstif),
     1         lnodn(melem),lmtyp(melem),ndof(mpoin),nvar(mpoin)
     2        ,xmatx(20,20),ymatx(20,20),nminvc(mtotv),lpoint(mtotv)
c
      ntotv = nvar(npoin) + ndof(npoin)
cp...ntotv is total no. of dof's
c
c...initialize gstif
      do 20 i=1,mstif
        gstif(i) = 0.0
   20 continue
c
      thedt = theta*dtime
      rewind 18
c
c...loop through all elements
      do 50 l=1,nelem
cp...get material type (1 for structural, 2 for soil)
        lmtype = lmtyp(l)
cp...read elt matrices
          read(18) xmatx
          read(18) ymatx
c     
        nnode = lnodn(l)
c...deal with elt nodes ni in rows, nj in columns
c... these will be global nodes nni, nnj
        ni=0
        i = 0
   25   ni = ni + 1
        nni = lnods(l,ni)
        idof = ndof(nni)
c...for structural elts we only use (u,v) d.o.f.s
        if (lmtype.eq.1) idof = 2
        ii = nvar(nni)
        nj = 0
        j = 0
   30   nj = nj + 1
        nnj = lnods(l,nj)
        jdof = ndof(nnj)
        if (lmtype.eq.1) jdof = 2
        jj = nvar(nnj)
c...put the entries from estif/xmatx into gstif
        do 40 mi=1,idof
          do 35 mj=1,jdof
            im = i + mi
            jm = j + mj
            xij = xmatx(im,jm)
            yij = ymatx(im,jm)
c...convert X to Xbar by multiplying H terms by (1-theta)*dt
            if (mi.eq.3.and.mj.eq.3) then
              xij = xij*thedt
            endif
            iim = ii + mi
            jjm = jj + mj
            if (jjm.lt.iim) goto 35
c...ignore entries below diagonal in gstif
c...translate (iim,jjm) to entry of array gstif with skyline storage
            ijpos = gpos(lpoint,ntotv,iim,jjm)            
            gstif(ijpos) = gstif(ijpos) + xij + yij
   35     continue
   40   continue
        j = j + jdof
        if (nj.lt.nnode) goto 30
        i = i + idof
        if (ni.lt.nnode) goto 25
   50 continue
c...end of element loop
c
      return
      end
c
c
      subroutine norm1(m,n,v,vnorm)
      dimension v(m)
c
c...finds the max. norm of vector v
      vnorm = 0.0
      do 5 i=1,n
        vi = abs(v(i))
        if (vi.gt.vnorm) vnorm = vi
    5 continue
c
      return
      end
c
c
      function gpos(lpoint,ntotv,i,j)
      dimension lpoint(*)
c...returns the position in the array gstif of a(i,j)
      if (i.lt.1.or.j.lt.1.or.i.gt.ntotv.or.j.gt.ntotv) then
        gpos = 0
        return
      endif
c...map entries in upper triangle, into lower triangle
      if (j.le.i) then
        ii = i
        jj = j
      else
        ii = j
        jj = i
      endif
      if (ii.eq.1) then
        gpos = 1
      else
        lpi = lpoint(ii)
        lpi1 = lpoint(ii-1)
        jmin = ii - lpi + lpi1 + 1
        if (jj.ge.jmin) then
          gpos = lpi - ii + jj
        else
          gpos = 0
        endif
      endif
      return
      end
c
c
      subroutine newrhs(melem,mpoin,mnode,mtotv,
     1                 nelem,npoin,lnods,lnodn,nkode,
     2                 asrhs,asdis,lmtyp,ndof,nvar,theta,dtime)
c
c...assembles the global rhs vector from asdis (previous displts)
c
      dimension lnods(melem,mnode),asrhs(mtotv),asdis(mtotv),
     1         lnodn(melem),lmtyp(melem),ndof(mpoin),nvar(mpoin)
     2        ,xmatx(20,20),ymatx(20,20),nkode(mpoin),kd(3)
c
      ntotv = nvar(npoin) + ndof(npoin)
cp...ntotv is total no. of dof's
c
      thedt = theta*dtime
      tht = (1.0-theta)/theta
      rewind 18
c
c...loop through all elements
      do 50 l=1,nelem
cp...get material type (1 for structural, 2 for soil)
        lmtype = lmtyp(l)
cp...read elt matrices
        read(18) xmatx
        read(18) ymatx
c     
        nnode = lnodn(l)
c...deal with elt nodes ni in rows, nj in columns
c... these will be global nodes nni, nnj
        ni=0
        i = 0
   25   ni = ni + 1
        nni = lnods(l,ni)
        idof = ndof(nni)
c...for structural elts we only use (u,v) d.o.f.s
        if (lmtype.eq.1) idof = 2
        ii = nvar(nni)
        nj = 0
        j = 0
   30   nj = nj + 1
        nnj = lnods(l,nj)
        jdof = ndof(nnj)
        if (lmtype.eq.1) jdof = 2
        jj = nvar(nnj)
c...multiply the entries from matrices X,Y by corresponding asdis
        do 40 mi=1,idof
          do 35 mj=1,jdof
            im = i + mi
            jm = j + mj
            xij = xmatx(im,jm)
            yij = ymatx(im,jm)
c...convert X to Xbar by multiplying H terms by (1-theta)*dt
            if (mi.eq.3.and.mj.eq.3) then
              xij = xij*thedt
            endif
            iim = ii + mi
            jjm = jj + mj
c...now form the matrix-vector product: asrhs = asrhs - (tht*Xbar - Y)*asdis
            cij = tht*xij - yij
            asrhs(iim) = asrhs(iim) - cij*asdis(jjm)
   35     continue
   40   continue
        j = j + jdof
        if (nj.lt.nnode) goto 30
        i = i + idof
        if (ni.lt.nnode) goto 25
   50 continue
c...end of element loop
c
c...now set the entries corresponding to fixed dof's, to zero
      do 120 nn=1,npoin
        kode = nkode(nn)
cp...have to use ndof(nn) rather than 2 for ndof
        ndofnn = ndof(nn)
        nvarnn = nvar(nn)
        if (ndofnn.eq.2) then
          nkd = 2
          kd(1) = kode/10
          kd(2) = kode - kd(1)*10
        else
          nkd = 3
          kd(1) = kode/100
          k2 = kode - kd(1)*100
          kd(2) = k2/10
          kd(3) = k2 - kd(2)*10
        endif
        do 110 m=1,nkd
          if(kd(m).eq.0) goto 110
          nposn = nvarnn + m
          asrhs(nposn) = 0.0
  110   continue
  120 continue
      return
      end
c
c
      subroutine result(melem,mpoin,mnode,msets,mcomp,mtotv,
     1            nelem,npoin,lnods,lnodn,lsidn,coord,nkode,
     3            lpros,props,aslod,asdis,lmtyp,ndof,nvar)
c
c...output the results to file
c
      dimension aslod(mtotv),props(msets,mcomp),
     1         lnods(melem,mnode),coord(mpoin,2),lsidn(melem),
     2         endis(16),encod(2,8),bmatx(3,16),gploc(2,4),
     3        shape(8),deriv(2,8),cartd(2,8),dmatx(3,3),  
     4      gpwts(4),gpsig(3),gpeps(3),gpcod(2),
     5      nkode(mpoin),asdis(mtotv),lnodn(melem),lpros(melem)
     6     ,lmtyp(melem),ndof(mpoin),nvar(mpoin)
c
c...output displacements. In this program, reactions are not 
c...calculated
c...output to screen (unit 6):
      write(6,*)
      write(6,*)'RESULTS'
      write(6,*)'======='
      write(6,*)
      write(6,610)
  610 format(//' nodal displacements, pore pressures'
     1/40('-')//3x,'node',
     212x,'x-displt',7x,'y-displt'
     3,5x,'pore pressure')
c...output to results file (unit 16):
      write(16,*)
      write(16,*)'RESULTS'
      write(16,*)'======='
      write(16,*)
      write(16,610)
c...and output to end of data file, for postprocessor:
      write(17,*)'  node  x-displt  y-displt  pore pressure'
      do 20 n=1,npoin  
        i = nvar(n)
        ndofnn = ndof(n)
        dispx = asdis(i+1)
        dispy = asdis(i+2)
        if (ndofnn.eq.3) then 
          press = asdis(i+3)
          write(17,'(i5,3e10.3)') n,dispx,dispy,press
        else        
          write(17,'(i5,2e10.3)') n,dispx,dispy
        endif
        if (mod(n,25).eq.0) call prompt
        if (ndofnn.eq.2) then
          write(6,615) n,dispx,dispy 
          write(16,615) n,dispx,dispy
        else
          write(6,615) n,dispx,dispy,press 
          write(16,615) n,dispx,dispy,press 
        endif 
  615   format
     1   (3x,i3,10x,e10.3,2(5x,e10.3))    
   20 continue
      call prompt
c
c...element Gauss pt stressess
  620 format(//'Element Gauss pt. stresses:'/27('-')/)
      write(6,620)
      write(16,620)
      write(17,*)
     1' elt gp    x-coord   y-coord   sigma-x   sigma-y   tau-xy'
      do 100 l=1,nelem
        if (mod(l,3).eq.0) call prompt
        write(6,630) l
        write(16,630) l
  630   format(//' element',i5/1x,12('-')/)
        lset = lpros(l)
        nnode = lnodn(l)
        nside = lsidn(l)
cp...only consider the 2 displacement dof's, so ...        
        nevab = 2*nnode
        ym = props(lset,1)
        pr = props(lset,2)
c...store nodal coords, displts for elt in encod, endis
        do 25 n=1,nnode
          nn = lnods(l,n)
          encod(1,n) = coord(nn,1)
          encod(2,n) = coord(nn,2)
cp...we only want to extract the 2 displt dof's
          i = 2*(n-1)
cp...have to use nvar for their location in global vector
          ii = nvar(nn)
          endis(i+1) = asdis(ii + 1)
          endis(i+2) = asdis(ii + 2)
   25   continue
        call dmat(ym,pr,dmatx)
c...find Gauss-pt stresses & strains
        write(6,640) (i,i=1,3)
        write(16,640) (i,i=1,3)
  640   format(/' gauss-pt stresses and strains'/6x,'x-coord',8x,
     1  'y-coord',10x,3('stress (',i1,')',5x))
        call gauss(gploc,gpwts,ngaus)
c...loop over each Gauss point
        do 80 ij=1,ngaus
          s = gploc(1,ij)
          t = gploc(2,ij)
          call sfr(nnode,nside,s,t,shape,deriv)
          call jacob(nnode,encod,shape,deriv,cartd,djacb)
          call bmat(cartd,bmatx)
c...get Gauss point coords in gpcod
          do 27 m=1,2
            sum = 0.0
            do 26 n=1,nnode
   26       sum = sum + encod(m,n)*shape(n)
            gpcod(m) = sum
   27     continue
c...find strains as bmatx*displts
          do 30 i=1,3
            sum = 0.0
            do 29 k=1,nevab
   29       sum = sum + bmatx(i,k)*endis(k)
            gpeps(i) = sum
   30     continue
c...stresses = dmatx*strains
          do 40 i=1,3
            sum = 0.0 
            do 39 k=1,3
   39       sum = sum + dmatx(i,k)*gpeps(k)
            gpsig(i) = sum
   40     continue
c...only stresses printed; modify this to print strains too
          write(17,'(i5,i2,3x,5e10.3)') 
     1       l,ij,(gpcod(i),i=1,2),(gpsig(i),i=1,3)
          write(6,650) (gpcod(i),i=1,2),(gpsig(i),i=1,3)
          write(16,650) (gpcod(i),i=1,2),(gpsig(i),i=1,3)
  650     format(3x,2(e10.3,5x),5x,3(e10.3,5x))
   80   continue
c...end of Gauss loop
  100 continue
      call prompt
c
c...end of element loop
c
      write(6,*)
      write(6,*)'END OF RESULTS'
      write(6,*)
      write(16,*)
      write(16,*)'END OF RESULTS'
      write(16,*)
      return
      end
c
c
c...subroutines for STIFF     
c
c
      subroutine gauss(gploc,gpwts,ngaus)
      dimension gploc(2,4),gpwts(4)
c...set up Gauss pts & weights for 2x2 integration
      ngaus = 4
      do 5 i=1,4
        gpwts(i) = 1.0
    5 continue
      r13 = 1.0/sqrt(3.0)
      gploc(1,1) = -r13
      gploc(1,2) = -r13
      gploc(1,3) = r13
      gploc(1,4) = r13
      gploc(2,1) = -r13
      gploc(2,2) = r13
      gploc(2,3) = -r13
      gploc(2,4) = r13
      return
      end
c
c
      subroutine dmat(ym,pr,dmatx)
      dimension dmatx(3,3)
c...ym,pr are Young's mod., Poisson's ratio
c...form the D matrix for plane strain
      g = 0.5*ym/(1.0 + pr)
      yl = g*pr/(0.5 - pr)
      g2yl = 2.*g + yl
      dmatx(1,1) = g2yl
      dmatx(1,2) = yl
      dmatx(1,3) = 0.0
      dmatx(2,1) = yl
      dmatx(2,2) = g2yl
      dmatx(2,3) = 0.0
      dmatx(3,1) = 0.0
      dmatx(3,2) = 0.0
      dmatx(3,3) = g
      return
      end
c
c
      subroutine bmat(cartd,bmatx)
      dimension cartd(2,8),bmatx(3,16)
c...form the B matrix
      do 5 j=1,16
      do 5 i=1,3
        bmatx(i,j) = 0.0
    5 continue
      j = 0
      do 10 n=1,8
        i = j + 1
        j = i + 1
        bmatx(1,i) = -cartd(1,n)
        bmatx(2,j) = -cartd(2,n)
        bmatx(3,i) = -cartd(2,n)
        bmatx(3,j) = -cartd(1,n)
c...negative sign for 'compression +ve' sign convention
   10 continue
      return
      end
c
c
      subroutine sfr(nnode,nside,s,t,shape,deriv)
      dimension shape(8),deriv(2,8)
c...evaluate the shape functions and 1st derivs at local coords s,t
cp...linear quad added, for porel
      nodsid = 10*nnode + nside
c
      if (nodsid.eq.84) then
      s2 = 2.0*s
      t2 = 2.0*t
      ss = s*s
      tt = t*t
      st = s*t
      sst = ss*t
      stt = s*tt
      st2 = 2.0*st
      shape(1) = (-1.0 + st + ss + tt - sst - stt)*0.25
      shape(2) = (1.0 - t - ss + sst)*0.5
      shape(3) = (-1.0 - st + ss + tt - sst + stt)*0.25
      shape(4) = (1.0 + s - tt - stt)*0.5
      shape(5) = (-1.0 + st + ss + tt + sst + stt)*0.25
      shape(6) = (1.0 + t - ss - sst)*0.5
      shape(7) = (-1.0 - st + ss + tt + sst - stt)*0.25
      shape(8) = (1.0 - s - tt + stt)*0.5
c     
      deriv(1,1) = (t + s2 - st2 - tt)*0.25
      deriv(1,2) = -s + st
      deriv(1,3) = (-t + s2 - st2 + tt)*0.25
      deriv(1,4) = (1.0 - tt)*0.5
      deriv(1,5) = (t + s2 + st2 + tt)*0.25
      deriv(1,6) = -s - st
      deriv(1,7) = (-t + s2 + st2 - tt)*0.25
      deriv(1,8) = (-1.0 + tt)*0.5
      deriv(2,1) = (s + t2 - ss - st2)*0.25
      deriv(2,2) = (-1.0 + ss)*0.5
      deriv(2,3) = (-s + t2 - ss + st2)*0.25
      deriv(2,4) = -t - st             
      deriv(2,5) = (s + t2 + ss + st2)*0.25
      deriv(2,6) = (1.0 - ss)*0.5
      deriv(2,7) = (-s + t2 + ss - st2)*0.25
      deriv(2,8) = -t + st
c
      else if (nodsid.eq.44) then
C
C...LINEAR QUADRILATERAL
        ST = S*T
        SHAPE(1) = (1.-S-T+ST)*.25
        SHAPE(2) = (1.+S-T-ST)*.25
        SHAPE(3) = (1.+S+T+ST)*.25
        SHAPE(4) = (1.-S+T-ST)*.25
        DERIV(1,1) = -(1.-T)*.25
        DERIV(1,2) =  (1.-T)*.25
        DERIV(1,3) =  (1.+T)*.25
        DERIV(1,4) = -(1.+T)*.25
        DERIV(2,1) = -(1.-S)*.25
        DERIV(2,2) = -(1.+S)*.25
        DERIV(2,3) =  (1.+S)*.25
        DERIV(2,4) =  (1.-S)*.25
      endif
      return
      end
c
c     
      subroutine jacob(nnode,encod,shape,deriv,cartd,djacb)
      dimension shape(8),deriv(2,8),cartd(2,8),encod(2,8),xjacm(2,2),
     1         xjaci(2,2)
c...evaluate the Jacobian matrix, and its determinant
      do 5 j=1,2
      do 5 i=1,2
        sum = 0.0
        do 4 n=1,nnode
    4   sum = sum + deriv(i,n)*encod(j,n)
        xjacm(i,j) = sum
    5 continue
c...xjacm is the Jacobian matrix. find det & inverse by Cramer        
      djacb = xjacm(1,1)*xjacm(2,2) - xjacm(1,2)*xjacm(2,1)
      if (djacb.lt.1.0e-12) then
        write(6,*)'ERROR: determinant of Jacobian is -ve or zero'
        write(6,*)'Element is inside out, or nodes defined clockwise'
        call prompt
        stop
      endif
      xjaci(1,1) = xjacm(2,2)/djacb
      xjaci(2,2) = xjacm(1,1)/djacb
      xjaci(1,2) = -xjacm(1,2)/djacb
      xjaci(2,1) = -xjacm(2,1)/djacb
c...get Cartesian derivs of shape functions using J inverse
      do 50 i=1,2
      do 50 n=1,nnode
        sum = 0.0
        do 49 j=1,2
   49   sum = sum + xjaci(i,j)*deriv(j,n)
        cartd(i,n) = sum
   50 continue
      return
      end
c
c
c...subroutines for SOLVE
c
c
      subroutine fixdof(mtotv,mstif,ntotv,mpoin,npoin,gstif,nkode,
     1    coord,asrhs,ndof,nvar,lpoint)
      dimension gstif(mstif),kd(3),nkode(mpoin),coord(mpoin,2),
     1  asrhs(mtotv)
     1,ndof(mpoin),nvar(mpoin),lpoint(mtotv)
c
cp...note that the r.h.s. vector is now asrhs, not aslod 
c
c...add a large number to the diagonal coefft of fixed d.o.f.s
      do 20 nn=1,npoin
        kode = nkode(nn)
cp...have to use ndof(nn) rather than 2 for ndof
        ndofnn = ndof(nn)
        nvarnn = nvar(nn)
        if (ndofnn.eq.2) then
          nkd = 2
          kd(1) = kode/10
          kd(2) = kode - kd(1)*10
        else
          nkd = 3
          kd(1) = kode/100
          k2 = kode - kd(1)*100
          kd(2) = k2/10
          kd(3) = k2 - kd(2)*10
        endif
c...do this for displt and pore pressure dof's
        do 10 m=1,nkd
          if (kd(m).eq.0) goto 10
          nposn = nvarnn + m
          nmpos = lpoint(nposn)
          gstif(nmpos) = gstif(nmpos) + 10.0e20
   10   continue
   20 continue
      return
      end
c
c
      subroutine ldldec(mtotv,mstif,ntotv,gstif,
     1                  lpoint)
      dimension gstif(mstif),lpoint(mtotv)
c        T
c...L D L   decomposition, using skyline storage 
cp...compare this with routine chodec in pladv.for,
cp                              T
cp...which uses the Choleski L L   decomposition
c
      small = 1.e-8
c...find the first diagonal entry, i.e. l(1,1)
      i = 1
      a11 = gstif(1)
      if (abs(a11).lt.small) then
        write(6,*)'ERROR: zero pivot in row ',i
        call prompt
        stop
      endif
      gstif(1) = a11
c...decompose each row in turn
      do 20 i=2,ntotv
        lpi = lpoint(i)
        lpi1 = lpoint(i-1)
        jstart = i - lpi + lpi1 + 1
c...work along the row
        do 10 j=jstart,i-1
          jjpos = lpoint(j)
          djj = gstif(jjpos)
          ijpos = gpos(lpoint,ntotv,i,j)
          aij = gstif(ijpos)
          sum = 0.0
          lpj = lpoint(j)
          if (j.eq.1) then
            lpj1 = 0
          else
            lpj1 = lpoint(j-1)
          endif
          kstart=max(i-lpi+lpi1+1,j-lpj+lpj1+1,1)
          do 5 k=kstart,j-1
            kkpos = lpoint(k)
            dkk = gstif(kkpos)
            ikpos = gpos(lpoint,ntotv,i,k)
            jkpos = gpos(lpoint,ntotv,j,k)
            if (ikpos.eq.0.or.jkpos.eq.0) goto 5
            sum = sum + dkk*gstif(ikpos)*gstif(jkpos)
    5     continue
          clij = (aij - sum)/djj
          gstif(ijpos) = clij
   10   continue
c...finally find the diagonal entry on the row
        iipos = lpoint(i)
        aii = gstif(iipos)
        sum = 0.0
        kstart = i - lpi +lpi1 + 1
        do 15 k=kstart,i-1
          kkpos = lpoint(k)
          dkk = gstif(kkpos)
          ikpos = gpos(lpoint,ntotv,i,k)
          clik = gstif(ikpos)
          sum = sum + dkk*clik*clik
   15   continue
        clii = aii - sum
        if (abs(clii).lt.small) then
          write(6,*)'ERROR: zero or negative pivot in row ',i
          call prompt
          stop
        endif
        gstif(iipos) = clii
   20 continue
      return
      end
c
c
      subroutine ldlsub(mtotv,mstif,ntotv,gstif,
     1                  asrhs,asdis,astmp,lpoint)
      dimension gstif(mstif),asdis(mtotv),asrhs(mtotv)
     1         ,lpoint(mtotv),astmp(mtotv)
c
cp...solution of gstif*asdis = asrhs, for the 
c        T
c...L D L   decomposition of gstif, using skyline storage 
c
c...forward substitution: solve Ly = b
c...store y in astmp
      astmp(1) = asrhs(1)
      do 10 i=2,ntotv
        bi = asrhs(i)
        sum = 0.0
        lpi = lpoint(i)
        lpi1 = lpoint(i-1)
        kstart = i - lpi + lpi1 + 1
        do 5 k=kstart,i-1
          ikpos = gpos(lpoint,ntotv,i,k)
          sum = sum + gstif(ikpos)*astmp(k)
    5   continue
        astmp(i) = bi - sum
   10 continue
c             
cp...diagonal bit: solution of Dz = y
      do 12 i=1,ntotv
        lpi = lpoint(i)
        dii = gstif(lpi)
        astmp(i) = astmp(i)/dii
   12 continue
c
c...back-substitution phase: solve L x = z
c...We can do this in a column-based manner!
      do 20 i=ntotv,1,-1 
        xi = astmp(i)
        asdis(i) = xi
c...once we have found x_i, transfer the term involving this, to
c...the r.h.s. in all earlier equations
        if (i.eq.1) goto 20
        lpi = lpoint(i)
        lpi1 = lpoint(i-1)
        ni = lpi - lpi1 - 1
        do 15 k=1,ni
          clij = gstif(lpi-k)
          jpos = i - k
          astmp(jpos) = astmp(jpos) - clij*xi
   15   continue
   20 continue
c
      return
      end
c
c
      subroutine reactn(mtotv,mstif,mpoin,npoin,
     1    gstif,asdis,aslod,nkode,ndof,nvar,lpoint)
c...calculate reactions, and store them in aslod.
c...zero the other entries in aslod.
      dimension gstif(mstif),asdis(mtotv),
     1          kd(3),nkode(mpoin),aslod(mtotv)
     1          ,ndof(mpoin),nvar(mpoin),lpoint(mtotv)
c
      do 5 i=1,mtotv
        aslod(i) = 0.0
    5 continue
      do 20 nn=1,npoin
        kode = nkode(nn)
cp...have to use ndof(nn) rather than 2 for ndof
        ndofnn = ndof(nn)
        nvarnn = nvar(nn)
        if (ndofnn.eq.2) then
          nkd = 2
          kd(1) = kode/10
          kd(2) = kode - kd(1)*10
        else
          nkd = 3
          kd(1) = kode/100
          k2 = kode - kd(1)*100
          kd(2) = k2/10
          kd(3) = k2 - kd(2)*10
        endif
        do 10 m=1,nkd
          if(kd(m).eq.0) goto 10
          nposn = nvarnn + m
          nmpos = lpoint(nposn)
          gnn = gstif(nmpos)
          aslod(nposn) = -gnn*asdis(nposn)
          asdis(nposn) = 0.0
   10   continue
   20 continue
      return
      end
c
c