c...Plane stress/strain advanced elasticity "main engine"
c...for use with FELIPE version 3 - sample datafiles are:
c...   eladbm1.dat (3-noded beam in elastic block)
c...   eladft3.dat (footing on semi-infinite soil mass)
c...   eladcv2.dat (cavern excavated in infinite stressed rock mass)
c...These illustrate use of different element and loading types
c
      program eladv
cb..."big" version of elast.for 2D elasticity program. 
cb...Includes coding for full range of element types in PREFEL,
cb...and for loading by spec. displts, body forces and 
cb...excavation. Lines in CAPITALS are added in eladv, as are
cb...comments starting cb... 
c
      parameter(melem=900,mpoin=2000,mnode=8,
     1msets=9,mcomp=6,mtotv=4000,mfixv=250,
     2mfron=250,mstif=31375,mbufa=100)
c...these parameters fix the dimensions of the arrays:
c...melem = max. no. of elements
c...mpoin = max. no. of nodes
c...mnode = max. no. of nodes in a single element
c...msets = max. no. of mat. prop. sets (also used for loading sets)
c...mcomp = max. no. of mat. props (e.g. Young's mod, Poisson ratio)
c...mtotv = max no. of degrees of freedom (= 2*mpoin for elasticity)
cb...mfixv = max no. of fixed degrees of freedom
cb...mfron = max frontwidth in solver FRONT
cb...mstif = (mfron+1)*mfron/2
cb...mbufa = size of buffer used in FRONT
c
c
      character char*1
      character*80 title,infile,outfil,prtfil
c
      dimension lnods(melem,mnode),lnodn(melem),lsidn(melem),
     1  coord(mpoin,2),nkode(mpoin),aslod(mtotv),ndof(mpoin),
     2        lpros(melem),props(msets,mcomp),asdis(mtotv),
     3      surft(msets,2),lmtyp(melem)
     4    ,nvar(mpoin),nfvar(mfixv),spdis(mfixv),TLOAD(MTOTV)
     5   ,strsg(4,4,melem)
c...lnods(l,n) = global no. of n'th node of element l
c...lnodn(l) = no. of nodes in element l ( = 8 for 8-noded quad.)
c...lsidn(l) = no. of sides in element l    ( = 4 for 8-noded quad.)
c...coord(np,1), coord(np,2) = x,y coords of node np
c...nkode(np) = fixity code of node np
c...aslod is global nodal load vector
c...ndof(np) = no. of dof.s at node np 
cp..nvar(np) is a pointer array for global dof of node np
cb...nfvar(i) is the global dof of the i'th fixed deg of freedom
cb...spdis(i) is the specified value of the i'th specified dof
cb...tload(i) is total load so far (used with excavations)
cb...strsg(i,ij,l) is stress component i at gausspt ij of element l
c...lpros(l) = mat. prop. set no. of element l
c...props(m,i) = value of i'th mat. property in mat. prop. set m
c...asdis is global nodal displacement vector
c...
c...surft(m,1),surft(m,2) = normal,tangential surface tractions in set m
c...gstif holds assembled global stiffness matrix in compact form
c...lmtyp(l) = material type of elt l
c 
      real*8 pival,eqrhs(mbufa)
     1      ,vecrv(mfron),gstif(mstif),gload(mfron),equat(mfron,mbufa)
      DIMENSION NPIVO(MBUFA),NAMEV(MBUFA),NACVA(MFRON)
cb...arrays above are used in the FRONT frontal solver
c
      write(6,*)'Program ELADV: advanced elasticity.'
      write(6,*)'Handles all 2D finite and infinite elements,'
      write(6,*)'as well as 2-noded and 3-noded 1D beam elements.'
      write(6,*)'Dimensioned for',melem,' elements, and',mpoin,' nodes.'
      write(6,*)
c
c...read an existing .dat file
c
    1 continue
c
c...get the input file name in infile
      call askfil(infile,iflag)
      if (iflag.eq.1) then
        call yesno(char,'Do you want to exit?')
        if (char.eq.'y') stop
        goto 1
      endif
c
      write(6,*)'Opening file ',infile,'...'
      lenin = lenstr(infile)
      open(15,file=infile(1:lenin),status='old',err=999)
      write(6,*)
      lm4 = lenin - 4
      prtfil = infile(1:lm4)//'.prt'
      write(6,*)'Creating file ',prtfil(1:lenin),' for results...'
      write(6,*)
      open(16,file=prtfil(1:lenin),status='unknown',err=999)      
      outfil = infile(1:lm4)//'.out'
      write(6,*)'Creating file ',outfil(1:lenin),' for output...'
      write(6,*)
      open(17,file=outfil(1:lenin),status='unknown',err=999)      
c
c...read the input mesh data, and echo it to the output files
      call input(15,melem,mpoin,mnode,msets,mcomp,nelem,
     1 npoin,lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props
     2   ,ndofn,npros,title,ndof)
      call print(16,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1 lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2           npros,title,ndof)
      call output(17,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1  lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2            npros,title,ndof)
      call yesno(char,'Do you want the mesh details listed?')
      if (char.eq.'y') then
        call print(6,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1 lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2           npros,title,ndof)
      endif
c
      write(6,*)
     1'Is this a plane stress, plane strain or axisymmetric analysis?'
    3 write(6,*)
     1'Type 1 for plane stress, 2 for plane strain, or 3 for axisymm...'
      read(5,'(i1)',err=3) mstyp
      write(16,*)
      if (mstyp.eq.1) then
        write(6,*) 'Plane stress analysis'
        write(16,*)'Plane stress analysis'
      else if (mstyp.eq.2) then
        write(6,*) 'Plane strain analysis'
        write(16,*)'Plane strain analysis'
      else if (mstyp.eq.3) then
        write(6,*) 'Axisymmetric analysis'
        write(16,*)'Axisymmetric analysis'
      else
        write(6,*)'Invalid input!'
        goto 3
      endif        
      write(16,*)    
c
cp...create the pointer array nvar. Needed because a node may
cp...have 2 or 3 d.o.f.s; this info stored in ndof
cp...also store info on fixed d.o.f.s in nfvar and spdis,
cp...rather than in nkode
      ntotv = 0
      nfixv = 0
      do 5 np=1,npoin
        ndofnp = ndof(np)
        nvar(np) = ntotv
        kode = nkode(np)
C...SORT OUT KODE AND STORE FIXED VARIABLE NO. IN NFVAR()
        KOD = KODE
        DO 4 M=1,NDOFNP
          KD = KOD/10
          K = KOD - KD*10
          KOD = KD
          IF(K.ne.0) then
            NFIXV=NFIXV+1
            IF (NFIXV.GT.MFIXV) THEN
              WRITE(6,*)
     1        'No. of fixed dof.s exceeds parameter MFIXV'
              write(6,*)'Increase MFIXV and recompile.'
              call prompt
              stop
            endif
            M1 = NDOFNP + 1 - M
            NFVAR(NFIXV) = ntotv + M1
            spdis(nfixv) = 0.0
          endif
    4   continue
        ntotv = ntotv + ndofnp
    5 continue
      nsdis = nfixv
cp...s/r LOAd will add spec displts, total no will be held in nsdis
cp...Now local dof i of node np is global dof nvar(np)+i
cp...and nfvar(i) is global dof of i'th fixed dof
cp...and ntotv is total no. of degrees of freedom
c
c
c...open a scratch file to write the element stiffness matrices on
      open(unit=18,status='scratch',form='unformatted')  
      write(6,*)'Scratch file opened ...'
      write(6,*)
c
cb...open further scratch files for use in FRONT
      OPEN(19,STATUS='SCRATCH',FORM='UNFORMATTED')
      OPEN(20,STATUS='SCRATCH',FORM='UNFORMATTED')
      OPEN(21,STATUS='SCRATCH',FORM='UNFORMATTED')
      PIVAL = 1.d-8
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine STIFF...'
      call stiff(melem,mpoin,mnode,msets,mcomp,mstyp,
     1    nelem,npoin,lnods,lnodn,lsidn,coord,lpros,props
     2           ,lmtyp,ndof,nvar)
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine LOAD...'
      call load(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1          nelem,npoin,lnods,lnodn,lsidn,coord,
     2 surft,aslod,lmtyp,ndof,nvar,tload,strsg,
     3 lpros,props,mfixv,nfixv,nsdis,nfvar,spdis)
c
cb...the next subroutine replaces ASSMB and SOLVE in ELAST.FOR
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine FRONT...'
      irsol = 0
cb...irsol=0 for first solution, irsol=1 for re-solution
      CALL FRONT(MTOTV,MPOIN,MELEM,MFIXV,MNODE,MFRON,MSTIF,
     1NPOIN,NELEM,NSDIS,LNODS,LNODN,NDOF,NVAR,NFVAR,ASLOD,ASDIS,SPDIS,
     2VECRV,GSTIF,GLOAD,EQUAT,NACVA,lsidn,
     3PIVAL,IRSOL,EQRHS,NPIVO,NAMEV,MBUFA)
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine RESULT...'
      call result(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1     nelem,npoin,lnods,lnodn,lsidn,coord,nkode,lpros,props,
     2            aslod,asdis,lmtyp,ndof,nvar,strsg,tload)
c
      write(6,*)
      write(6,*)'ELADV completed. Results are in file ',prtfil(1:lenin)
      write(6,*)
      write(6,*)
      call prompt
      stop
  999 write(6,*)
     1'Cannot open file with this name. Terminating program.'
      call prompt
      stop
      end
c
c
      subroutine stiff(melem,mpoin,mnode,msets,mcomp,mstyp,
     1  nelem,npoin,lnods,lnodn,lsidn,coord,lpros,props
     2           ,lmtyp,ndof,nvar)
c      
c...calculates the element stiffness matrices, and writes them
c...to the scratch file.
c
      dimension props(msets,mcomp),lnods(melem,mnode),encod(2,8),
     1         estif(16,16),bmatx(4,16),dmatx(4,4),dbmat(4,16),
     2        gpwts(4),lpros(melem),lnodn(melem),coord(mpoin,2),
     3       shape(8),gploc(2,4),gpcod(2),deriv(2,8),cartd(2,8),
     4      lsidn(melem),lmtyp(melem),ndof(mpoin),nvar(mpoin),
     5     smapf(8),dmapf(2,8)
c
      REAL*8 SUM
C
c...encod holds the nodal coords for the nodes around the current element
c...estif will be the stiffness matrix for the current element
c...bmatx, dmatx are the B and D matrices, and dbmat is D*B
c...gpwts,gploc are the Gauss point weights and locations (in xi,eta coords)
c...gpcod will be the x,y coords of the current Gauss point
c...shape holds the shape functions evaluated at the Gauss point
c...deriv holds the derivatives (w.r.t. xi, eta) of the shape functions
c...cartd holds the Cartesian derivatives (w.r.t x,y) of shape functions
cb...smapf,dmapf are the shape and deriv mapping functions, for infinite elts
c
      rewind 18
c...for axisymmetric analyses, D is a 4x4 matrix; otherwise it's 3x3
      ndmatx = 3
      if (mstyp.eq.3) ndmatx = 4 
c...and we need the value of pi, for integrating. Form it using inverse tangent
      pi = 4.*atan(1.0)
c
c...start loop over all elements
      do 80 l=1,nelem
c
c...extract Young's modulus & Poisson's ratio from props      
        nnode = lnodn(l)
        nside = lsidn(l)
        lset = lpros(l)
        nodsid = nnode*10 + nside
c          
c...put nodal coords in encod
CB...COUNT NO. OF DOF'S IN THE ELEMENT: STORE IN NEVAB
        NEVAB = 0
        do 15 n=1,nnode
          np = lnods(l,n)
          NDOFNp = NDOF(Np)
cb...ignore 3rd d.o.f. unless a 1D beam element
          if (ndofnp.gt.2.and.nside.gt.1) ndofnp = 2
          NEVAB = NEVAB + NDOFNp
          encod(1,n) = coord(np,1)
          encod(2,n) = coord(np,2)
   15   continue
cb...nevab now has no. of d.o.f.s in the element
c
c...zero stiffness matrix
        do 20 j=1,16
        do 20 i=1,16
          estif(i,j) = 0.0
   20   continue
c
        if (nside.eq.1) then
c
c...1D beam elements (2- or 3-noded)
c   ----------------
c...extract Young's modulus be and moment of inertia bi from props      
          lset = lpros(l)
          be = props(lset,1)
          bi = props(lset,2)
c...also extract cross-sectional area ba of beam
          ba = props(lset,3)
          if (ba.eq.0.0) ba = 1.0
c
          call estifb(nnode,ba,be,bi,encod,estif)
c
        else
c
c...2D elasticity element: code as in elast.for
c   ---------------------
c
c...extract material properties
           ym = props(lset,1)
           pr = props(lset,2)
c...also extract material thickness (default is 1.0)
           thick = props(lset,3)
           if (thick.eq.0.0) thick = 1.0
c
c...get D matrix
           call dmat(mstyp,ym,pr,dmatx)
c
c...put local coords of Gauss-pts in gploc, weights in gpwts
cb...ngaus is the number of Gauss pts
           call gauss(nnode,nside,gploc,gpwts,ngaus)
c
c...start loop through gauss-pts for contribution to estif
           do 75 ij=1,ngaus
c...put local coords of gauss-pt in s,t
             s = gploc(1,ij)
             t = gploc(2,ij)
c...get shape fns (shape) & derivs (deriv) in local coords
             call sfr(nnode,nside,s,t,shape,deriv)
           IF    (NODSID.EQ.53.OR.NODSID.EQ.32.
     1         OR.NODSID.EQ.43) THEN
C...MAPPING FUNCTIONS FOR INFINITE ELEMENT
	    CALL MAPFUN(nnode,nside,s,t,smapf,dmapf)
           ELSE
c...for standard f.e., mapping fn is same as shape fn
	      DO 23 N=1,NNODE
	        SMAPF(N) = SHAPE(N)
	        DO 23 M=1,2
	          DMAPF(M,N) = DERIV(M,N)
   23         CONTINUE
           ENDIF
C
c...get xy-coords of gauss-pt into gpcod, using shape fn interpolation
             do 25 m=1,2
               sum = 0.D0
               do 24 n=1,nnode
   24     sum = sum + DBLE(encod(m,n))*DBLE(smapf(n))
               gpcod(m) = sum
   25        continue
C
c...for axisymmetric analyses, we need the radius at the Gauss point
        if (mstyp.eq.3) radius = gpcod(1)
c...get xy-derivs (cartd) & derterminant of Jacobian (djacb)
        call jacob(nnode,nside,encod,shape,deriv,cartd,djacb,
     1             smapf,dmapf)
c...get B matrix 
        call bmat(nnode,cartd,mstyp,radius,shape,bmatx)
C
c...form matrix product D*B
             do 40 j=1,NEVAB
             do 40 i=1,ndmatx
               sum=0.D0
               do 39 k=1,ndmatx
   39          sum = sum + DBLE(dmatx(i,k))*DBLE(bmatx(k,j))
               dbmat(i,j) = sum
   40        continue
C
c...integrate B^T *D*B over volume of element
             darea = gpwts(ij)*djacb
c...for axisymmetry, the 'thickness' is the circumference
             if (mstyp.eq.3) then
               dvolu = darea*2.*pi*radius
c...otherwise, get volume by multiplying by thickness
             else
               dvolu = darea*thick
             endif
c...B transpose *d*b* dvolu = contribution to estif
             do 50 j=1,NEVAB
             do 50 i=1,NEVAB
               sum = 0.D0
               do 49 k=1,ndmatx
   49     sum = sum + DBLE(bmatx(k,i))*DBLE(dbmat(k,j))
               estif(i,j) = estif(i,j) + sum*dvolu
   50        continue
   75      continue
c...end of gauss loop, and of estif calculation for 2D elts
        endif
c...write estif to file
   77   write (18) estif
   80 continue
c...end of element loop
c
      write(6,*)'Elt stiffness matrices written to file'
      endfile 18
      return
      end
c
c
      subroutine estifb(nnode,ba,be,bi,encod,estif)
      dimension encod(2,8),estif(16,16)
c
c...stiffness matrices for 2-noded and 3-noded beam elements
c...find length, and angle of inclination of beam
           dx = encod(1,nnode) - encod(1,1)
           dy = encod(2,nnode) - encod(2,1)
           bl = sqrt(dx*dx + dy*dy)
           alpha = atan2(dy,dx)
           c = cos(alpha)
           s = sin(alpha)
           c2 = c*c
           s2 = s*s
           cs = c*s
           bl2 = bl*bl
           bil = bi/bl
           bil2 = bi/bl2
c
      if (nnode.eq.2) then
c...2-noded beam element: stiffness matrix code from frame.for
c   --------------------
           nevab = 6
           estif(1,1) =   ba*c2 + 12.*bil2*s2 
           estif(1,2) =  (ba    - 12.*bil2)*cs
           estif(1,3) =         -  6.*bil*s
           estif(1,4) = -(ba*c2 + 12.*bil2*s2)
           estif(1,5) = -(ba    - 12.*bil2)*cs
           estif(1,6) =         -  6.*bil*s
           estif(2,2) =   ba*s2 + 12.*bil2*c2
           estif(2,3) =            6.*bil*c
           estif(2,4) = -(ba    - 12.*bil2)*cs
           estif(2,5) = -(ba*s2 + 12.*bil2*c2)
           estif(2,6) =            6.*bil*c
           estif(3,3) =            4.*bi
           estif(3,4) =            6.*bil*s
           estif(3,5) =         -  6.*bil*c
           estif(3,6) =            2.*bi
           estif(4,4) =   ba*c2 + 12.*bil2*s2
           estif(4,5) =  (ba    - 12.*bil2)*cs
           estif(4,6) =            6.*bil*s
           estif(5,5) =   ba*s2 + 12.*bil2*c2
           estif(5,6) =         -  6.*bil*c
           estif(6,6) =            4.*bi
c
      else
c...3-noded Chao beam element - see manual for theory
c...there are (u,v,th) dof's at end-nodes, and (u,v) at mid-node
           nevab = 8
           ba7  = 7.*ba/3.
           ba8  = 8.*ba/3.
           ba16 = 16.*ba/3.
           ba1  = ba/3.
           estif(1,1) =   ba7*c2 +  63.2*bil2*s2 
           estif(1,2) =  (ba7    -  63.2*bil2)*cs
           estif(1,3) =          -  37.6*bil*s
           estif(1,4) = -ba8*c2  - 102.4*bil2*s2
           estif(1,5) = (-ba8    + 102.4*bil2)*cs
           estif(1,6) =   ba1*c2 +  39.2*bil2*s2
           estif(1,7) = ( ba1    -  39.2*bil2)*cs
           estif(1,8) =             13.6*bil*s
           estif(2,2) =   ba7*s2 +  63.2*bil2*c2
           estif(2,3) =             37.6*bil*c
           estif(2,4) = (-ba8    + 102.4*bil2)*cs
           estif(2,5) = -ba8*s2  - 102.4*bil2*c2
           estif(2,6) = (ba1     -  39.2*bil2)*cs
           estif(2,7) = ba1*s2   +  39.2*bil2*c2
           estif(2,8) =          -  13.6*bil*c
           estif(3,3) =             28.8*bi
           estif(3,4) =             51.2*bil*s
           estif(3,5) =          -  51.2*bil*c
           estif(3,6) =          -  13.6*bil*s
           estif(3,7) =             13.6*bil*c
           estif(3,8) =          -   4.8*bi
           estif(4,4) =  ba16*c2 + 204.8*bil2*s2
           estif(4,5) =  (ba16   - 204.8*bil2)*cs
           estif(4,6) = -ba8*c2  - 102.4*bil2*s2
           estif(4,7) = (-ba8    + 102.4*bil2)*cs
           estif(4,8) =          -  51.2*bil*s
           estif(5,5) =  ba16*s2 + 204.8*bil2*c2
           estif(5,6) = (-ba8    + 102.4*bil2)*cs
           estif(5,7) =  -ba8*s2 - 102.4*bil2*c2
           estif(5,8) =             51.2*bil*c
           estif(6,6) =  ba7*c2  +  63.2*bil2*s2
           estif(6,7) = (ba7     -  63.2*bil2)*cs
           estif(6,8) =             37.6*bil*s
           estif(7,7) =  ba7*s2  +  63.2*bil2*c2
           estif(7,8) =          -  37.6*bil*c
           estif(8,8) =             28.8*bi 
c
      endif
c
c...all terms are multiplied by E/l
      bel = be/bl
      do 35 i=1,nevab
        do 35 j=i,nevab
          estif(i,j) = bel*estif(i,j)
   35 continue
c
c...now fill in lower triangle of estif
      do 36 i=2,nevab
        do 36 j=1,i-1
          estif(i,j) = estif(j,i)
   36 continue
c
      return
      end
c
c
c
      subroutine load(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1           nelem,npoin,lnods,lnodn,lsidn,coord,surft,
     2  aslod,lmtyp,ndof,nvar,TLOAD,strsg,lpros,props,
     3  mfixv,nfixv,nsdis,nfvar,spdis)
c
c...compile global nodal load vector
c
      dimension aslod(mtotv),surft(msets,2),LSIDN(MELEM),
     1    lnods(melem,mnode),press(2,3),npals(8),KARD(16),
     2     coord(mpoin,2),shape(8),deriv(2,8),xjacm(2),      
     3     gpres(4),fcomp(2),encod(2,8),lnodn(melem)
     4 ,lmtyp(melem),ndof(mpoin),nvar(mpoin),TLOAD(MTOTV)
     5 ,gploc(2,4),gpwts(4),gpcod(2),smapf(8),dmapf(2,8)
     6 ,strsg(4,4,melem),enbod(2,8),lpros(melem)
     7 ,props(msets,mcomp),nfvar(mfixv),spdis(mfixv)
     8 ,bmatx(4,16),cartd(2,8)
      real*8 sum
c
      ntotv = nvar(npoin) + ndof(npoin)
      zero = 0.0
c...we need the value of pi, for integrating in axisymmetry.
      if (mstyp.eq.3) pi = 4.*atan(1.0)
c
      write(6,*)'Applied Loads'
      write(6,*)'-------------'
      write(6,*)
      write(16,*)'Applied Loads'
      write(16,*)'-------------'
      write(16,*)
c...initialize the global load vector
      do 10 i=1,mtotv
        aslod(i) = 0.0
        tload(i) = 0.0
   10 continue
c...and the stress field
      do 12 i=1,4
        do 12 ij=1,4
          do 12 l=1,melem
            strsg(i,ij,l) = 0.0
   12 continue
c
c...there are some dummy lines in the input data, which are read but ignored
      read(15,*,end=99)
c...npset>0 if there are surface tractions.
cb...nbset>0 if there are body forces.
      read (15,510) nbset,npset
  510 format(2i5)
c
c...nodal loads/displts (ld=1 for load, ld=2 for specified displt)
c...-------------------
c
      write(6,*)'Nodal loads, specified displacements:'
      write(6,*)'-------------------------------------'
      write(6,*)
      write(6,*)'Nonzero entries below indicate loads/displts'
      write(6,*)
      write(6,*)'  Node   x-component   y-component'
      write(16,*)'Nodal loads:'
      write(16,*)'-----------'
      write(16,*)
      write(16,*)'Nonzero entries below indicate loads/displts'
      write(16,*)
      write(16,*)'  Node   x-component   y-component'
      read(15,*)
   15 read(15,520) lcard,ld,np,fx,fy
  520 format(i1,i4,i5,2e10.3)
      if (np.ne.0) then
c...for axisymmetry, integrate load around element
        if (mstyp.eq.3) then
          radius = coord(np,1)
          circum = 2.*pi*radius
          fx = fx*circum
          fy = fy*circum
        endif
c...the i'th component of node np must be converted to global degree 
c...of freedom component
cp...use nvar for this!
        nm = nvar(np)
c...add nodal loads into aslod
        if (ld.eq.1) then
          aslod(nm + 1) = aslod(nm + 1) + fx
          aslod(nm + 2) = aslod(nm + 2) + fy
          write(6,610) np,fx,fy,'point load'
          write(16,610) np,fx,fy,'point load'
        else
Cb...New coding for entering specified displacements
cb...only deals with first two d.o.f.s of node
cb...nsdis starts out as being equal to nfixv 
          iyesx = 0
          iyesy = 0 
          DO 17 K=1,NFIXV
            IF (NFVAR(K).EQ.NM+1) iyesx = 1
            IF (NFVAR(K).EQ.NM+2) iyesy = 1
   17     CONTINUE
          IF (FX.NE.0.0.and.iyesx.eq.0) THEN
            NSDIS = NSDIS + 1
            IF(NSDIS.GT.MFIXV) THEN
              WRITE(6,612)
              call prompt
              stop
            ENDIF
            SPDIS(NSDIS) = FX 
            NFVAR(NSDIS) = NM + 1
            write(6,610) np,fx,zero,'spec dsplt'
          ENDIF
          IF (FY.NE.0.0.and.iyesy.eq.0) THEN
            NSDIS = NSDIS + 1
            IF(NSDIS.GT.MFIXV) THEN
              WRITE(6,612)
              call prompt
              stop
            ENDIF
            SPDIS(NSDIS) = FY 
            NFVAR(NSDIS) = NM + 2
            write(16,610) np,zero,fy,'spec dsplt'
          ENDIF
        endif
  610   format(i5,2(4x,e10.3),3x,a10)
  612   format(/'Adding spec displts exceeds MFIXV.'/
     1  'Increase MFIXV parameter and recompile.')
      endif
      if (lcard.eq.0) goto 15
      write(6,*)
      write(16,*)
c
c
cb...new coding in eladv for body force sets
C....                        ---------------
      IF (nbset.gt.0) then
        read(15,*)
        write(6,*)'Body forces'
        write(6,*)'-----------'
        write(6,*)
        write(6,*)' set      x-cmpnt      y-cmpnt' 
        write(16,*)'Body forces'
        write(16,*)'-----------'
        write(16,*)
        write(16,*)' set      x-cmpnt      y-cmpnt' 
cb...we use surft to store body forces, to avoid defining a new array
        do 105 i=1,msets
          surft(i,1) = 0.0
          surft(i,2) = 0.0
  105   continue
c...first read body force sets
  108   read(15,532) lcard,iset,i12,bxy
  532   format(i1,i4,3x,i2,10x,e10.3)
        if (iset.gt.msets) then
          write(6,*)'Error: body force set no. ',iset
          write(6,*)'exceeds MSETS. Increase MSETS and recompile.'
          call prompt
          stop
        endif 
        if (iset.gt.0) then 
          if (i12.eq.1) then
            surft(iset,1) = bxy
          else if (i12.eq.2) then
            surft(iset,2) = bxy
          endif
          write(6,620)  iset,surft(iset,1),surft(iset,2)
          write(16,620) iset,surft(iset,1),surft(iset,2)
        endif
        if (lcard.eq.0) goto 108
C
c...now apply these BODY FORCES TO ELEMENTS
        read(15,*)
        write(6,*)
        write(6,*)'   elt  set-number'
        write(16,*)
        write(16,*)'   elt   set-number'
  110   read(15,542) lcard,iset,(KARD(I),I=1,14)
  542   format(i1,i4,14i5)
        if (ISET.gt.0) then
          IL = 0
  115     IL = IL + 1
          L = KARD(IL)
          if (L.gt.0) then
            write(6,632) l,iset
            write(16,632) l,iset
  632       format(i5,2x,i5)
            BX = SURFT(ISET,1)
            BY = SURFT(ISET,2)
CB...CONVERT BODY FORCE ON ELT L TO NODAL FORCES
            NNODE = LNODN(L)
            NSIDE = LSIDN(L)
            nodsid = 10*nnode + nside
c...ignore 1D elements
            if (nside.eq.1) goto 137
c...shouldn't put body forces on infinite elts, either!
            DO 118 N=1,NNODE
              NP = LNODS(L,N)
              ENCOD(1,N) = COORD(NP,1)
              ENCOD(2,N) = COORD(NP,2)
  118       CONTINUE
            NODSID = NNODE*10 + NSIDE
c...material thickness is used in integrating over volume of elt
            lset = lpros(l)
            thick = props(lset,3)
            if (thick.eq.0.0) thick = 1.0
C
            call gauss(nnode,nside,gploc,gpwts,ngaus)
c
c...start loop through gauss-pts 
            do 135 ij=1,ngaus
c...put local coords of gauss-pt in s,t
              s = gploc(1,ij)
              t = gploc(2,ij)
c...get shape fns (shape) & derivs (deriv) in local coords
              call sfr(nnode,nside,s,t,shape,deriv)
              IF    (NODSID.EQ.53.OR.NODSID.EQ.32.
     1        OR.NODSID.EQ.43) THEN
C...MAPPING FUNCTIONS FOR INFINITE ELEMENT
	          CALL MAPFUN(nnode,nside,s,t,smapf,dmapf)
              ELSE
       	    DO 123 N=1,NNODE
	            SMAPF(N) = SHAPE(N)
	            DO 123 M=1,2
	              DMAPF(M,N) = DERIV(M,N)
  123           CONTINUE
              ENDIF
c...for axisymmetry, we need the radius, which is the x-coord of Gauss pt
              if (mstyp.eq.3) then
                sum = 0.0
                do 124 n=1,nnode
                  sum = sum + encod(1,n)*shape(n)
  124           continue
                radius = sum
              endif
c...get xy-derivs (cartd) & determinant of Jacobian (djacb)
              call jacob(nnode,nside,encod,shape,deriv,cartd,djacb,
     1             smapf,dmapf)
C...ACCUMULATE EQUIV. NODAL FORCES IN ASLOD
              if (mstyp.eq.3) then
                dvolu = djacb*gpwts(ij)*2.*pi*radius
              else
                dvolu = djacb*gpwts(ij)*thick
              endif
              DO 130 N=1,NNODE
                NP = LNODS(L,N)
                NVNP = NVAR(NP)
                IF (BX.NE.0.0) THEN
                  NM = NVNP + 1
                  ASLOD(NM) = ASLOD(NM) + BX*SHAPE(N)*Dvolu
                ENDIF
                IF (BY.NE.0.0) THEN
                  NM = NVNP + 2
                  ASLOD(NM) = ASLOD(NM) + BY*SHAPE(N)*Dvolu
                ENDIF
  130         CONTINUE
  135       CONTINUE
CB...END OF GAUSS LOOP
          ENDIF
  137     IF (IL.LT.14) GOTO 115
        ENDIF
        IF (LCARD.EQ.0) GOTO 110
      ENDIF
CB...END OF BODY FORCE SECTION
C
C
c...surface tractions
C...-----------------
C
      if (npset.gt.0) then
        read(15,*)
        write(6,*)'Surface Tractions'
        write(6,*)'-----------------'
        write(6,*)
        write(6,*)' set   normal cmpnt   tangential cmpnt' 
        write(16,*)'Surface Tractions'
        write(16,*)'-----------------'
        write(16,*)
        write(16,*)' set   normal cmpnt   tangential cmpnt' 
c
        do 22 i=1,msets
          surft(i,1) = 0.0
          surft(i,2) = 0.0
   22   continue
c...first read surface traction sets
   25   read(15,530) lcard,iset,i1,i2,pn,pt
  530   format(i1,i4,3x,2i2,8x,2e10.3)
        if (iset.gt.msets) then
          write(6,*)'Error: surface traction set no. ',iset
      write(6,*)'exceeds MSETS. Increase MSETS and recompile.'
          call prompt
          stop
        endif 
        if (iset.gt.0) then 
          surft(iset,1) = pn
          surft(iset,2) = pt
          write(6,620)  iset,pn,pt
          write(16,620) iset,pn,pt
  620     format(i3,2(5x,e10.3))
        endif
        if (lcard.eq.0) goto 25
c
c...now apply these tractions to edges
        read(15,*)
        write(6,*)
        write(6,*)'   elt  end-nodes   set-numbers'
        write(16,*)
        write(16,*)'   elt   end-nodes   set-numbers'
   30   read(15,540) lcard,nnas,l,n1,iset1,n2,iset2
c...nnas is the number of nodes along the edge (=3 for 8-noded quad.)
  540   format(i1,i4,i5,2(i5,i2,1x))
        if (l.gt.0) then
c...put the tractions in press.
cb...coding for nnas=2 (linear side) added in eladv.for
          if (nnas.eq.2) then
            do 33 i=1,2
              press(i,1) = surft(iset1,i)
              press(i,2) = surft(iset2,i)
   33       continue
            npals(1) = n1
            npals(2) = n2
          else 
c...Interpolate linearly for the
c...tractions at the midside node.
            do 35 i=1,2
              press(i,1) = surft(iset1,i)
              press(i,3) = surft(iset2,i)
              press(i,2) = (press(i,1) + press(i,3))/2.
   35       continue
c...store the node numbers along side in npals
            npals(1) = n1
            npals(3) = n2
c...we have to find the midside node no. from lnods
            nnode = lnodn(l)
            do 42 n=1,nnode
              if (lnods(l,n).eq.npals(1)) goto 45
   42       continue
c...if loop completed, npals(1) is not in element 1
            write(6,700) npals(1),l
  700     format('ERROR: node',i4,' is not in element ',i4,'.')
            call prompt
            stop
   45       k = n + 1
            npals(2) = lnods(l,k) 
          endif
          write(6,630) l,n1,n2,iset1,iset2
          write(16,630) l,n1,n2,iset1,iset2
  630     format(i5,2(2x,2i5))
c
c...extract nodal coords and store in encod
          do 60 n=1,nnas
            np = npals(n)
            do 55 m=1,2
   55         encod(m,n) = coord(np,m)
   60     continue
c
          nnode = lnodn(l)
          nside = lsidn(l)
          nodsid = 10*nnode + nside
c
          if (nodsid.eq.21) then
c
c...2-noded beam element: coding from frame.for
c...find the angle of inclination of the beam
            dx = encod(1,2) - encod(1,1)
            dy = encod(2,2) - encod(2,1)
            bl = sqrt(dx*dx + dy*dy)
            alpha = atan2(dy,dx)
            cosa = cos(alpha)
            sina = sin(alpha)
c...equiv nodal forces in normal and tangential directions at node 1:
c...note: normal forces use the cubic shape fns, while tangential 
c...      forces use the linear shape fns, hence the different coeffts
c...distributed loading also creates moments at nodes
            fn = (7.0*press(1,1) + 3.0*press(1,2))/20.
            ft = (2.0*press(2,1) +     press(2,2))/6.
            bm = (3.0*press(1,1) + 2.0*press(1,2))/60. 
c...now resolve into x and y direction forces
            fx = -fn*sina + ft*cosa
            fy =  fn*cosa + ft*sina
            np1 = npals(1)
            nm = nvar(np1)
c...add nodal loads into aslod
            aslod(nm + 1) = aslod(nm + 1) + fx*bl
            aslod(nm + 2) = aslod(nm + 2) + fy*bl
            aslod(nm + 3) = aslod(nm + 3) + bm*bl*bl
c...now deal with node 2
            fn =  (3.0*press(1,1) + 7.0*press(1,2))/20.
            ft =  (    press(2,1) + 2.0*press(2,2))/6.
            bm = -(2.0*press(1,1) + 3.0*press(1,2))/60. 
            fx = -fn*sina + ft*cosa
            fy =  fn*cosa + ft*sina
            np2 = npals(2)
            nm = nvar(np2)
            aslod(nm + 1) = aslod(nm + 1) + fx*bl
            aslod(nm + 2) = aslod(nm + 2) + fy*bl
            aslod(nm + 3) = aslod(nm + 3) + bm*bl*bl
c
          else if (nodsid.eq.31) then
c
c...3-noded beam element
c
            dx = encod(1,3) - encod(1,1)
            dy = encod(2,3) - encod(2,1)
            bl = sqrt(dx*dx + dy*dy)
            alpha = atan2(dy,dx)
            cosa = cos(alpha)
            sina = sin(alpha)
c...equiv nodal forces in normal and tangential directions at node 1:
c...note: normal forces use the quartic shape fns, while tangential 
c...      forces use the linear shape fns, hence the different coeffts
c...distributed loading also creates moments at nodes
            fn = (13.0*press(1,1) +    press(1,3))/60.
            ft =       press(2,1)/6.
            bm =       press(1,1)/30. 
c...now resolve into x and y direction forces
            fx = -fn*sina + ft*cosa
            fy =  fn*cosa + ft*sina
            np1 = npals(1)
            nm = nvar(np1)
c...add nodal loads into aslod
            aslod(nm + 1) = aslod(nm + 1) + fx*bl
            aslod(nm + 2) = aslod(nm + 2) + fy*bl
            aslod(nm + 3) = aslod(nm + 3) + bm*bl*bl
c...now deal with node 2
            fn =  4.0*(press(1,1) + press(1,3))/15.
            ft =  (    press(2,1) + press(2,3))/3.
            fx = -fn*sina + ft*cosa
            fy =  fn*cosa + ft*sina
            np2 = npals(2)
            nm = nvar(np2)
            aslod(nm + 1) = aslod(nm + 1) + fx*bl
            aslod(nm + 2) = aslod(nm + 2) + fy*bl
c...finally, node 3
            fn = (     press(1,1) + 13.0*press(1,3))/60.
            ft =       press(2,3)/6.
            bm =     - press(1,3)/30. 
            fx = -fn*sina + ft*cosa
            fy =  fn*cosa + ft*sina
            np3 = npals(3)
            nm = nvar(np3)
            aslod(nm + 1) = aslod(nm + 1) + fx*bl
            aslod(nm + 2) = aslod(nm + 2) + fy*bl
            aslod(nm + 3) = aslod(nm + 3) + bm*bl*bl
c
          else
c
c...2D element
c   ----------
c
c...extract material thickness, for integration of load over 
c...2D element edge (which is actually a face in 3D)
            lset = lpros(l)
            thick = props(lset,3)
            if (thick.eq.0.0) thick = 1.0
c...enter 1D Gauss loop along side (2-pt integration if nnas=3)
            if (nnas.eq.2) then
              ngaus = 1
              gploc(1,1) = 0.0
              gpwts(1) = 2.0
            else
              ngaus = 2
              r13 = 1.0/sqrt(3.0)
              gploc(1,1) = -r13
              gploc(1,2) = r13
              gpwts(1) = 1.0
              gpwts(2) = 1.0
            endif
            do 90 ij=1,ngaus
              s = gploc(1,ij)
              if (nnas.eq.2) then
                shape(1) = 0.5*(1.-s)
                shape(2) = 0.5*(1.+s)
                deriv(1,1) = -0.5
                deriv(1,2) = 0.5
              else
                ss = s*s
                 shape(1) = (ss-s)*0.5
                shape(2) = 1.0 - ss
                shape(3) = (ss + s)*0.5
                deriv(1,1) = s - 0.5
                deriv(1,2) = -2.0*s
                deriv(1,3) = s + 0.5
              endif
c...get jacobian xjacm
              do 75 j=1,2 
                sum = 0.d0
                do 72 n=1,nnas
   72           sum = sum + dble(deriv(1,n))*dble(encod(j,n))
                xjacm(j) = sum
   75         continue
c...for axisymmetry, we need the radius, which is the x-coord of Gauss pt
              if (mstyp.eq.3) then
                sum = 0.0
                do 76 n=1,3
                  sum = sum + dble(encod(1,n))*dble(shape(n))
   76           continue
                radius = sum
              endif
c...find gauss pt pressures, & integrate for components in local coords
              gpwt = gpwts(ij)
              if (mstyp.eq.3) then
                darea = gpwt*2.*pi*radius
              else
                darea = gpwt*thick
              endif
c...find gauss pt pressures, & convert to components in local coords
              do 80 m=1,2
                sum = 0.d0
                do 77 n=1,nnas
   77           sum = sum + dble(press(m,n))*dble(shape(n))
                gpres(m) = sum
   80         continue
              fcomp(1) = (xjacm(1)*gpres(2) - xjacm(2)*gpres(1))*darea
              fcomp(2) = (xjacm(1)*gpres(1) + xjacm(2)*gpres(2))*darea
c...accumulate equivalent nodal forces in aslod
              do 85 n=1,nnas
                np = npals(n)
cp...use nvar to identify global dof's
                nvarnp = nvar(np)
                do 82 m=1,2
                  nm = nvarnp + m
                  aslod(nm) = aslod(nm) + fcomp(m)*shape(n)
   82           continue
   85         continue
   90       continue
c...end of Gauss loop
c
          endif
        endif
        if (lcard.eq.0) goto 30
      endif
c...go back for next loaded edge
c...end of surface tractions
c
c
c...in situ stress field
c...--------------------
c
      read(15,*)
      read(15,'(i5)') nbdry
      if (nbdry.gt.0) then
        read(15,*)
        write(6,*)'In situ stresses'
        write(6,*)'------------------'
        write(6,*)
        write(16,*)'In situ stresses'
        write(16,*)'------------------'
        write(16,*)
        write(6,*)
        write(6,*)'  node  set-number'
        write(16,*)
        write(16,*)'  node   set-number'
  210   read(15,544) lcard,istyp,iset,real1,real2,(KARD(I),I=1,10)
  544   format(i1,i2,i2,2g10.3,10i5)
cb...iset not used in eladv.for - assume only one boundary set
        istyp1 = istyp
        if (istyp1.gt.0) then
cb...the two reals define the insitu stress state, their meaning depending on istyp
          if (istyp.eq.1) then
            sx = real1
            sy = real2
          else if (istyp.eq.2) then
            specwt = real1
            hvrat = real2
          endif
c...note boundary nodes by changing ndof to -ve; nbd counts no. of bdry nodes
          nbd = 0
          i = 0
  212     i = i + 1
          np = kard(i)
          if (np.gt.0) then
            nbd = nbd + 1
            ndofnp = ndof(np)
            ndof(np) = -ndofnp
            write(6,'(i5,5x,i5)')np,iset
            write(16,'(i5,5x,i5)')np,iset
          endif
          if (i.lt.10) goto 212
c...now read further data lines
          if (lcard.eq.0) then
            read(15,544) lcard,idummy,iset,real1,real2,(KARD(I),I=1,10)
            i = 0
            goto 212
          endif
          write(6,*)
          write(16,*)
          if (istyp.eq.1) then
            write(6,'(a,f12.5)')'Horizontal stress = ',sx
            write(6,'(a,f12.5)')'Vertical stress   = ',sy
            write(16,'(a,f12.5)')'Horizontal stress = ',sx
            write(16,'(a,f12.5)')'Vertical stress   = ',sy
          else if (istyp.eq.2) then
            write(6,'(a,f12.5)')'Specific weight of rock = ',specwt
            write(6,'(a,f12.5)')'Lateral stress ratio    = ',hvrat
            write(16,'(a,f12.5)')'Specific weight of rock = ',specwt
            write(16,'(a,f12.5)')'Lateral stress ratio    = ',hvrat
          endif
          write(6,*)
          write(16,*)
          write(6,*)
          write(16,*)
c
c...now ready to calculate Gauss pt stresses (in strsg) and 
c...equivalent nodal loads (in tload or aslod)
c
c...start loop over all elements - this is just like in STIFF
          do 480 l=1,nelem
C
c...put nodal coords in encod
CB...COUNT NO. OF DOF'S IN THE ELEMENT: STORE IN NEVAB
            nnode = lnodn(l)
            nside = lsidn(l)
            ltype = lmtyp(l)
            nodsid = nnode*10 + nside
C...SKIP REST OF LOOP FOR STRUCTURAL or 1D ELT 
            IF (ltype.EQ.1) GOTO 480
            if (nside.eq.1) goto 480
            NEVAB = 0
            do 415 n=1,nnode
              np = lnods(l,n)
              NDOFNP = abs(NDOF(np))
              NEVAB = NEVAB + NDOFNP
              encod(1,n) = coord(np,1)
              encod(2,n) = coord(np,2)
  415       continue
c...zero body force matrix
            do 420 j=1,8
            do 420 i=1,2
              enbod(i,j) = 0.0
  420       continue
c
            lset = lpros(l)
            thick = props(lset,3)
            if (thick.eq.0.0) thick = 1.0
c
            call gauss(nnode,nside,gploc,gpwts,ngaus)
c
c...start loop through gauss-pts for contribution to enbod
            do 475 ij=1,ngaus
c...put local coords of gauss-pt in s,t
              s = gploc(1,ij)
              t = gploc(2,ij)
c...get shape fns (shape) & derivs (deriv) in local coords
              call sfr(nnode,nside,s,t,shape,deriv)
              IF (NODSID.EQ.53.OR.NODSID.EQ.32.
     1          OR.NODSID.EQ.43) THEN
C...MAPPING FUNCTIONS FOR INFINITE ELEMENT
	          CALL MAPFUN(nnode,nside,s,t,smapf,dmapf)
              ELSE
	          DO 423 N=1,NNODe
	            SMAPF(N) = SHAPE(N)
	            DO 423 M=1,2
	              DMAPF(M,N) = DERIV(M,N)
  423           CONTINUE
              ENDIF
C
c...get xy-coords of gauss-pt into gpcod, using shape fn interpolation
              do 425 m=1,2
                sum = 0.D0
                do 424 n=1,nnode
  424           sum = sum + DBLE(encod(m,n))*DBLE(smapf(n))
                gpcod(m) = sum
  425         continue
              if (mstyp.eq.3) radius = gpcod(1)
c...if overburden loading, ignore gauss pts above y=0 surface
              if (istyp.eq.2.and.gpcod(2).gt.0.0) goto 475
c...find horizontal and vertical stresses
              if (istyp.eq.1) then
                gpres(1) = sx
                gpres(2) = sy
              else
                sygp = -specwt*gpcod(2)
                gpres(1) = sygp
                gpres(2) = hvrat*sygp
              endif
              gpres(3) = 0.0
              if (mstyp.eq.3) gpres(4) = gpres(1)
c...get xy-derivs (cartd) & determinant of Jacobian (djacb)
              call jacob(nnode,nside,encod,shape,deriv,cartd,djacb,
     1             smapf,dmapf)
              gpwt = gpwts(ij)
              if (mstyp.eq.3) then
                dvolu = gpwt*djacb*2.*pi*radius
                nbrow = 4
              else
                dvolu = gpwt*djacb*thick
                nbrow = 2
              endif
C... PUT STRESSES INTO STRSG 
   	      STRSG(1,IJ,L) = gpres(1)
  	      STRSG(2,IJ,L) = gpres(2)
              if (mstyp.eq.3) strsg(4,ij,l) = gpres(4)
c...get B matrix
              call bmat(nnode,cartd,mstyp,radius,shape,bmatx)
C
              I = 0
              DO 445 N=1,NNODE
                DO 440 M=1,2
                  I = I + 1
                  SUM = 0.0d0
                  DO 435 J=1,nbrow
                    SUM = SUM + dble(BMATX(J,I))*dble(gpres(J))
  435             CONTINUE
                  ENBOD(M,N) = ENBOD(M,N) + SUM*DVOLU
  440           CONTINUE
  445         CONTINUE
  475       continue
C...END OF GAUSS LOOP.
c
C     ADD EQUIVT NODAL LOADS TO TLOAD
C-
            DO 455 N=1,NNODE
              NP = LNODS(L,N)
              NIK = NVAR(NP)
              DO 452 M=1,2
                NM = NIK + M
                TLOAD(NM) = TLOAD(NM) + ENBOD(M,N)
  452         continue
  455       CONTINUE

            IF (nbd.EQ.0) goto 480
c...for edges which lie on unloading boundaries, put reactions normal
c...to the edge into aslod.
c...only consider the finite-length edge of infinite elements
            nedge = nside
            if (nodsid.eq.53.or.nodsid.eq.43) nedge = 1
            if (nodsid.eq.32) goto 480
c...NNAS is the number of nodes along an edge
            nnas = 3
            if (nnode.le.4) nnas = 2
c
c...now loop over each edge
            n1 = 1
            do 477 iedge = 1,nedge
c...put the local node nos of nodes along this edge, into NPALS
              do 462 i=1,nnas
                ni = n1 + i - 1
                if (ni.gt.nnode) ni = ni - nnode
                npals(i) = ni
  462         continue
c
c...see if this edge is to be unloaded
c...check the first node of our edge to see if it lies on boundary
              np1 = LNODS(l,NPALS(1))
              if (ndof(np1).gt.0) goto 472
c...check if all nodes of the edge are there
              do 463 i=2,nnas
                npi = LNODS(l,NPALS(I))
                if (ndof(npi).gt.0) goto 472
  463         continue
c...if we have reached here, there is unloading of this edge to be done
c
c...coding to calculate normal reaction at nodes along edge
  465         continue
c...nnas is nnode for the 1D element type: linear or quadratic bar
c...S is the local coordinate value at each edge in turn
              s = -1.0
c...take each node along edge in turn
              do 468 i=1,nnas
c...evaluate derivatives of shape functions at this node
                if (nnas.eq.2) then
                  deriv(1,1) = -0.5
                  deriv(1,2) = 0.5
                else
                  deriv(1,1) = s - 0.5
                  deriv(1,2) = -2.0*s
                  deriv(1,3) = s + 0.5
                endif
                ni = npals(i)
c...calculate direction of edge at this point
                ax = 0.0
                ay = 0.0
                do 466 j=1,nnas
                  nj = npals(j)
                  ax = ax + deriv(1,j)*encod(1,nj)
                  ay = ay + deriv(1,j)*encod(2,nj)
  466           continue
                djacb = sqrt(ax*ax + ay*ay)
                cosb = ax/djacb
                sinb = ay/djacb
c...now take the nodal forces, and find reaction normal to surface
                fx = enbod(1,ni)
                fy = enbod(2,ni)
                react = fx*sinb - fy*cosb
c...find components in x,y directions of this reaction, to add to ASLOD
                rx = -react*sinb
                ry =  react*cosb
c...now add to the corresponding d.o.f.s of global load vector
                npi = LNODS(l,NI)
                nik = nvar(npi)
                aslod(nik+1) = aslod(nik+1) + rx
                aslod(nik+2) = aslod(nik+2) + ry
c...local coordinate S for next node:
                s = s + 1.0
                if (nnas.eq.2) s = s + 1.0
  468         continue
c...move on to starting node of next edge
  472         n1 = n1 + nnas - 1
c...for 6-noded quad, nnas alternates between 2 and 3
              if (nodsid.eq.64) nnas = 5 - nnas
c...end of loop over edges
  477       continue
c
  480     continue
c...end of element loop
c
cb...don't forget to put ndof back in original state
          do 485 np=1,npoin
            ndofnp = ndof(np)
            ndof(np) = abs(ndofnp)
  485     continue
        endif
      endif
c
c
c...report the non-zero loads assembled in aslod
        call prompt
        write(6,*)
        write(6,*)'Non-zero entries in global load vector:'
        write(6,*)' d.o.f.     value'
        write(6,*)
        write(16,*) 
        write(16,*)'Non-zero entries in global load vector:'
        write(16,*)' d.o.f.     value'
        write(16,*)
        do 95 i=1,ntotv
          ai = aslod(i)
c...zero any tiny loads caused by rounding errors
          if (abs(ai).lt.1.e-5) then
            aslod(i) = 0.0
          else
            write(6,650) i,ai
            write(16,650) i,ai
  650       format(i5,5x,e10.3) 
          endif
  95    continue
c
      write(6,*)
      write(6,*)'End of load data.'
      write(6,*)
      write(16,*)
      write(16,*)'End of load data.'
      write(16,*)
      return
   99 write(6,*)'ERROR: this input file has no loading data.'
      call prompt
      stop
      end
c      
c
c
      SUBROUTINE FRONT(MTOTV,MPOIN,MELEM,MFIXV,MNODE,MFRON,MSTIF,
     1NPOIN,NELEM,NSDIS,LNODS,LNODN,NDOF,NVAR,NFVAR,ASLOD,ASDIS,SPDIS,
     2VECRV,GSTIF,GLOAD,EQUAT,NACVA,lsidn,
     3PIVAL,IRSOL,EQRHS,NPIVO,NAMEV,MBUFA)
C************************************************************************
C...solves (assembled ESTIF)*ASDIS = ASLOD  by frontal algorithm. For
C...symmetric matrices. Code based on code from FINEPACK package, by
c...kind permission of David Naylor. 
C************************************************************************
      real*8 eqr,spd,eqj,pivot,pivmin,pivmax,pival,eqpiv,SMALL
     1      ,vecrv(mfron),gstif(mstif),gload(mfron),equat(mfron,mbufa)
     2      ,eqrhs(mbufa),eqjp
      DIMENSION LNODS(MELEM,MNODE),LNODN(MELEM),NDOF(MPOIN)
     1  ,NFVAR(MFIXV),ASLOD(MTOTV),ASDIS(MTOTV),SPDIS(MFIXV)
     3  ,ESTIF(16,16),NPIVO(MBUFA),NAMEV(MBUFA),nldof(16)
     4  ,NACVA(MFRON),LOCEL(24),NDEST(24),NVAR(MPOIN),lsidn(melem)
C...VECRV...NDEST INTERNAL USE ONLY
C
C...FUNCTION STATEMENT
      NFUNC(I,J) = (J*J - J)/2 + I
c...this relates the (i,j)th matrix entry to the vector storage
      IF(NFUNC(MFRON,MFRON).GT.MSTIF) GO TO 500
C
      SMALL = 1.d-6
C...CHANGE SIGN OF LAST APPEARANCE OF EACH NODE
C...NEEDED FOR INITIAL SOLUTION (IRSOL=0) ONLY
C
      IF (IRSOL.EQ.0) THEN
        DO 40 NP=1,NPOIN
          KLAST = 0
          DO 30 L=1,NELEM
            NNODE = LNODN(L)
            DO 20 N=1,NNODE
              IF(LNODS(L,N).EQ.NP) THEN
                KLAST = L
                NLAST = N
              ENDIF
   20       CONTINUE
   30     CONTINUE
          IF(KLAST.NE.0) LNODS(KLAST,NLAST) = -NP
   40   CONTINUE
      ENDIF
C
C...ZERO ARRAYS
      DO 50 I=1,MSTIF
   50 GSTIF(I) = 0.0D0
      DO 60 I=1,MFRON
        GLOAD(I) = 0.0D0
        VECRV(I) = 0.0D0
   60 NACVA(I) = 0
      DO 70 I=1,MBUFA
        EQRHS(I) = 0.0D0
        DO 70 J=1,MFRON
          EQUAT(J,I) = 0.0D0
   70 CONTINUE
      PIVMAX = 0.0D0
      PIVMIN = 1.0D9
      NFMAX = 0
      NBUFA = 0
C
      REWIND (18)
      REWIND (19)
      REWIND (20)
C
C...NEXT SECTION SOLVES USING NEW STIFFNESS MATRIX
      IF (IRSOL.GT.1) GOTO 522
C...ENTER MAIN ASSEMBLY - REDUCTION LOOP
      NFRON = 1
      KELVA = 0
      DO 380 L=1,NELEM
        nrdof = 0
        NNODE = LNODN(L)
        nside = lsidn(l)
C...NEXT BIT ONLY NEEDED FOR INITIAL SOLUTION
        IF (IRSOL.EQ.0) THEN
C...SET UP LOCEL TO RELATE ELEMENT VARIABLE POSITIONS TO OVERALL POSNS
          NEVAB = 0
          nlvab = 0
          DO 170 N=1,NNODE
            NP = LNODS(L,N)
            IS = 1
            IF(NP.GT.0) GO TO 160
            IS = -1
            NP = -NP
  160       NDOFNP = NDOF(NP)
            nldfnp = ndofnp
c...allow for 2D elements not contributing to the
c...rotation d.o.f.s of beam-elt nodes contained in them
            if (nside.gt.1.and.ndofnp.gt.2) nldfnp = 2
            do 164 i=1,nldfnp
              nlvab = nlvab + 1
              nldof(nlvab) = nevab + i
  164       continue 
            NIK = NVAR(NP)
            DO 165 I=1,NDOFNP
              NEVAB = NEVAB + 1             
  165       LOCEL(NEVAB) = (NIK+I)*IS
            LNODS(L,N) = NP
  170     CONTINUE
C...THE ROLE OF LNODS FOR LOCATING LAST NODE APPEARANCE HAS BEEN TRANS-
C...FERRED TO LOCEL. HENCEFORTH WE FORGET D.OF F. AT NODE AND CONSIDER
C...ONLY THE VARIABLE POSITION (IE EQN NO.)
C
C...LOOK FOR EXISTING DESTINATIONS
          KEVAB = 0
          DO 210 IE=1,NEVAB
            NIKNO = IABS(LOCEL(IE))
            KEXIS = 0
            DO 180 I=1,NFRON
              IF(NIKNO.NE.NACVA(I)) GO TO 180
              KEVAB = KEVAB+1
              KEXIS = 1
              NDEST(KEVAB) = I
  180       CONTINUE
            IF(KEXIS.EQ.1) GO TO 210
C...ALLOCATE VARIABLE A SPACE IN NACVA
            DO 190 I=1,MFRON
              IF(NACVA(I).NE.0) GO TO 190
              NACVA(I) = NIKNO
              KEVAB = KEVAB+1
              NDEST(KEVAB) = I
              GO TO 200
  190       CONTINUE
            GO TO 502
C...EXPAND FRONT WIDTH AS REQUIRED
  200       NF = NDEST(KEVAB)
            IF(NF.GT.NFRON) NFRON=NF
            IF(NFRON.GT.NFMAX) NFMAX = NFRON
  210     CONTINUE
C
C...CHECK MSTIF DIMENSION SUFFICIENT
          NSTIF = NFUNC(NFRON,NFRON)
          IF(NSTIF.GT.MSTIF) GO TO 503
C...WRITE ARRAYS TO FILE FOR FUTURE USE WITH MATRICES OF SAME STRUCTURE
          WRITE (20) LOCEL,NDEST,NACVA,NFRON,NEVAB
C
        ELSE
C...READ ARRAYS IF STRUCTURE ALREADY ESTABLISHED
          READ (20) LOCEL,NDEST,NACVA,NFRON,NEVAB
          IF (NFRON.GT.NFMAX) NFMAX = NFRON
        ENDIF
C
C
        READ (18) ESTIF
C
C...COMMENCE ASSEMBLY - FIRST LOADS
        do 215 ie=1,nevab
          I = NDEST(IE)
          NIK = LOCEL(IE)
          NIKLE = -NIK
          IF (NIK.LT.0) GLOAD(I) = GLOAD(I) + ASLOD(NIKLE)
  215   continue
C...AND THEN ELEMENT STIFFNESSES - SKIP IF RESOLUTION
        DO 240 IEl=1,NlVAB
          ie = nldof(iel)
          I = NDEST(IE)
          DO 220 JEl=1,IEl
            je = nldof(jel)
            J = NDEST(JE)
            NASH = NFUNC(I,J)
            NISH = NFUNC(J,I)
            STIF = ESTIF(IEl,JEl)
            IF (J.GE.I) THEN
              GSTIF(NASH) = GSTIF(NASH) + STIF
            ELSE
              GSTIF(NISH) = GSTIF(NISH) + STIF
            ENDIF
  220     CONTINUE
  240   CONTINUE
C
C...FIND OUT WHICH ELEMENT NODES CAN BE ELIMINATED
        DO 370 IE=1,NEVAB
          NIKNO = -LOCEL(IE)
          IF(NIKNO.LE.0) GO TO 370
C...VARIABLES TO BE ELIMINATED NOW IDENTIFIED - FIND THEIR POSITIONS
          DO 350 I=1,NFRON
            IF(NACVA(I).NE.NIKNO) GO TO 350
C
            NBUFA = NBUFA + 1
            IF(NBUFA.GT.MBUFA) THEN
              NBUFA = 1
              WRITE(19) EQUAT,EQRHS,NPIVO,NAMEV
            ENDIF
C
C...EXTRACT COEFS OF NEW EQN FOR ELIMINATION (FIRST SOLUTION ONLY)
            DO 250 J=1,NFMAX
              IF (I.LT.J) THEN
                NL = NFUNC(I,J)
              ELSE
                NL = NFUNC(J,I)
              ENDIF
              EQUAT(J,NBUFA) = GSTIF(NL)
              GSTIF(NL) = 0.0d0
  250       CONTINUE
C...AND THE CORRESPONDING R.H.S.
            EQRHS(NBUFA) = GLOAD(I)
            GLOAD(I) = 0.0d0
            KELVA = KELVA+1
C
            NAMEV(NBUFA) = NIKNO
            NPIVO(NBUFA) = I
            II = I
C
C...TAKE DIAGONAL COEF. OF VARIABLE TO BE ELIM. AS PIVOT. CHECK FOR SIZE
            PIVOT = EQUAT(II,NBUFA)
            IF (PIVOT.GT.PIVMAX) PIVMAX = PIVOT
            IF (PIVOT.LT.PIVMIN) PIVMIN = PIVOT
            IF(pivot.LT.100.d0*PIVAL) WRITE(6,650) real(pivot),NIKNO
            IF(ABS(PIVOT).LT.PIVAL) GO TO 501
            EQUAT(II,NBUFA) = 0.0d0
C
C...ENQUIRE IF VARIABLE NIKNO FREE OR PRESCRIBED
            DO 285 N=1,NSDIS
              IF(NFVAR(N).EQ.NIKNO) GO TO 287
  285       CONTINUE
            GO TO 300
C...NIKNO IS PRESCRIBED
  287       SPD = SPDIS(N)
            DO 290 J=1,NFRON
  290       GLOAD(J) = GLOAD(J) - SPD*EQUAT(J,NBUFA)
            GO TO 340
C
C...ELIMINATE FREE VARIABLE - FIRST DEAL WITH RHS THEN LHS COEFS.
  300       EQPIV = EQRHS(NBUFA)/PIVOT
            DO 330 J=1,NFRON
              EQJ = EQUAT(J,NBUFA)
              GLOAD(J) = GLOAD(J) - EQJ*EQPIV
              if (eqj.eq.0.0d0) goto 330
              NL = NFUNC(0,J)
              EQJP = EQJ/PIVOT
              DO 310 K=1,J
                KN = K + NL
  310         GSTIF(KN) = GSTIF(KN) - EQJP*EQUAT(K,NBUFA)
  330       CONTINUE
C
  340       EQUAT(II,NBUFA) = PIVOT
            NACVA(II) = 0
            GO TO 360
  350     CONTINUE
C...ELIMINATION COMPLETE
C
C...REDUCE FRONT WIDTH IF POSSIBLE
  360     IF(NACVA(NFRON).NE.0) GO TO 370
          NFRON = NFRON-1
          IF(NFRON.GT.0) GO TO 360
  370   CONTINUE
  380 CONTINUE
C...ELEMENT LOOP COMPLETED IN FORWARD ELIMINATION
C
      WRITE (19) EQUAT,EQRHS,NPIVO,NAMEV
      BACKSPACE (19)
C
C
C...START BACK SUBSTITUTION.
      DO 420 IE=1,KELVA
        IF(NBUFA.EQ.0) THEN
          BACKSPACE (19)
          READ (19) EQUAT,EQRHS,NPIVO,NAMEV
          BACKSPACE (19)
          NBUFA = MBUFA
        ENDIF
        I = NPIVO(NBUFA)
        NIKNO = NAMEV(NBUFA)
C
C...CHECK IF DISP. PRESCRIBED
        DO 400 N=1,NSDIS
          IF(NFVAR(N).EQ.NIKNO) GO TO 410
  400   CONTINUE
C...DISP. NOT PRESCRIBED - COMPUTE IT
        PIVOT = EQUAT(I,NBUFA)
        EQUAT(I,NBUFA) = 0.0d0
        EQR = EQRHS(NBUFA)
        DO 405 J=1,NFMAX
  405   EQR = EQR - VECRV(J)*EQUAT(J,NBUFA)
        VECRV(I) = EQR/PIVOT
        GO TO 415
C...DISP. IS PRESCRIBED - PUT REACTIONS (WHICH INCL APP. LOADS) IN ASLOD
  410   VECRV(I) = SPDIS(N)
        EQR = EQRHS(NBUFA)
        DO 412 J=1,NFMAX
  412   EQR = EQR - VECRV(J)*EQUAT(J,NBUFA)
        ASLOD(NIKNO) = -EQR
  415   ASDIS(NIKNO) = VECRV(I)
        NBUFA = NBUFA - 1
  420 CONTINUE
C
C...ADVISE WHAT MAX FRONT WIDTH WAS
      WRITE(6,600) NFMAX
      WRITE (6,610) PIVMIN,PIVMAX
      RETURN
C
C
C...THIS SECTION FOR RE-SOLUTIONS WITH PREVIOUSLY-DECOMPOSED MATRIX
  522 CONTINUE
C
C
      REWIND (21)
C
C...ENTER MAIN ASSEMBLY - REDUCTION LOOP
      NFRON = 1
      KELVA = 0
      nbufa = mbufa
      DO 580 L=1,NELEM
        READ (20) LOCEL,NDEST,NACVA,NFRON,NEVAB
        IF (NFRON.GT.NFMAX) NFMAX=NFRON
C
C...COMMENCE ASSEMBLY - FIRST LOADS
        DO 525 IE=1,NEVAB
          I = NDEST(IE)
          NIK = LOCEL(IE)
          NIKLE = -NIK
          IF (NIK.LT.0) GLOAD(I) = GLOAD(I) + ASLOD(NIKLE)
  525   CONTINUE
C
C...FIND OUT WHICH ELEMENT NODES CAN BE ELIMINATED
        DO 570 IE=1,NEVAB
          NIKNO = -LOCEL(IE)
          IF(NIKNO.LE.0) GO TO 570
C...VARIABLES TO BE ELIMINATED NOW IDENTIFIED - FIND THEIR POSITIONS
          DO 550 I=1,NFRON
            IF(NACVA(I).NE.NIKNO) GO TO 550
C
            NBUFA = NBUFA + 1
            IF(NBUFA.GT.MBUFA) THEN
              NBUFA = 1
              WRITE(21) EQRHS
              READ(19) EQUAT,EQRHS,NPIVO,NAMEV
            ENDIF
C
C...AND THE CORRESPONDING R.H.S.
            EQRHS(NBUFA) = GLOAD(I)
            GLOAD(I) = 0.0d0
            KELVA = KELVA+1
C
            NAMEV(NBUFA) = NIKNO
            NPIVO(NBUFA) = I
            II = I
C
C...TAKE DIAGONAL COEF. OF VARIABLE TO BE ELIM. AS PIVOT. CHECK FOR SIZE
            PIVOT = EQUAT(II,NBUFA)
            IF (PIVOT.GT.PIVMAX) PIVMAX = PIVOT
            IF (PIVOT.LT.PIVMIN) PIVMIN = PIVOT
            IF(pivot.LT.100.d0*PIVAL) WRITE(6,650) real(pivot),NIKNO
            IF(ABS(PIVOT).LT.PIVAL) GO TO 501
            EQUAT(II,NBUFA) = 0.0d0
C
C...ENQUIRE IF VARIABLE NIKNO FREE OR PRESCRIBED
            DO 585 N=1,NSDIS
              IF(NFVAR(N).EQ.NIKNO) GO TO 587
  585       CONTINUE
            GO TO 588
C...NIKNO IS PRESCRIBED
  587       SPD = SPDIS(N)
            DO 590 J=1,NFRON
  590       GLOAD(J) = GLOAD(J) - SPD*EQUAT(J,NBUFA)
            GO TO 540
C
C...ELIMINATE FREE VARIABLE - FIRST DEAL WITH RHS THEN LHS COEFS.
  588       EQPIV = EQRHS(NBUFA)/PIVOT
            DO 530 J=1,NFRON
              EQJ = EQUAT(J,NBUFA)
            GLOAD(J) = GLOAD(J) - EQJ*EQPIV
  530       CONTINUE
C
  540       EQUAT(II,NBUFA) = PIVOT
            NACVA(II) = 0
            GO TO 560
  550     CONTINUE
C...ELIMINATION COMPLETE
C
C...REDUCE FRONT WIDTH IF POSSIBLE
  560     IF(NACVA(NFRON).NE.0) GO TO 570
          NFRON = NFRON-1
          IF(NFRON.GT.0) GO TO 560
  570   CONTINUE
  580 CONTINUE
C...ELEMENT LOOP COMPLETED IN FORWARD ELIMINATION
C
      ENDFILE (21)
      BACKSPACE (21)
      BACKSPACE (19)
C
C
C...START BACK SUBSTITUTION.
      DO 520 IE=1,KELVA
        IF(NBUFA.EQ.0) THEN
          BACKSPACE (19)
          READ (19) EQUAT,EQRHS,NPIVO,NAMEV
          BACKSPACE (19)
          NBUFA = MBUFA
C...RESOLUTIONS ONLY
          BACKSPACE (21)
          READ (21) EQRHS
          BACKSPACE (21)
        ENDIF
        I = NPIVO(NBUFA)
        NIKNO = NAMEV(NBUFA)
C
C...CHECK IF DISP. PRESCRIBED
        DO 507 N=1,NSDIS
          IF(NFVAR(N).EQ.NIKNO) GO TO 510
  507   CONTINUE
C...DISP. NOT PRESCRIBED - COMPUTE IT
        PIVOT = EQUAT(I,NBUFA)
        EQUAT(I,NBUFA) = 0.0d0
        EQR = EQRHS(NBUFA)
        DO 505 J=1,NFMAX
  505   EQR = EQR - VECRV(J)*EQUAT(J,NBUFA)
        VECRV(I) = EQR/PIVOT
        GO TO 515
C...DISP. IS PRESCRIBED - PUT REACTIONS (WHICH INCL APP. LOADS) IN ASLOD
  510   VECRV(I) = SPDIS(N)
        EQR = EQRHS(NBUFA)
        DO 512 J=1,NFMAX
  512   EQR = EQR - VECRV(J)*EQUAT(J,NBUFA)
        ASLOD(NIKNO) = -EQR
  515   ASDIS(NIKNO) = VECRV(I)
        NBUFA = NBUFA - 1
  520 CONTINUE
C
C...ADVISE WHAT MAX FRONT WIDTH WAS
      WRITE(6,600) NFMAX
      RETURN
C
  600 FORMAT(/' FRONT COMPLETED. MAX FRONT WIDTH WAS',I4)
  610 FORMAT(1P,/' MIN. PIVOT =',G12.4,', MAX. PIVOT =',G12.4)
  650 FORMAT(1P,' WARNING IN FRONT, PIVOT =',G12.4,' FOR EQN',I4,
     1 ' IS DANGEROUS')
  700 FORMAT(/' STOPPED AT START OF FRONT.',
     1' MFRON TOO LARGE IN RELATION TO MSTIF.')
  701 FORMAT(/' STOPPED IN FRONT, PIVOT TOO SMALL FOR EQN',I4)
  702 FORMAT(/' STOPPED IN FRONT, NFRON =',I4,' EXCEEDS MFRON IN EL.'
     1,I4)
  703 FORMAT(/' STOPPED IN FRONT, MSTIF (=',I4,') EXCEEDED IN ELEMENT'
     1,I4,', NEEDS TO BE AT LEAST',I5)
C
C...ERROR MESSAGES
  500 WRITE(6,700)
              call prompt
      STOP
  501 WRITE(6,701) NIKNO
              call prompt
      STOP
  502 WRITE(6,702) NFRON,L
              call prompt
      STOP
  503 WRITE(6,703) MSTIF,L,NSTIF
              call prompt
      STOP
      END
C
C
      subroutine result(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1                  nelem,npoin,lnods,lnodn,lsidn,coord,nkode,
     3         lpros,props,aslod,asdis,lmtyp,ndof,nvar,strsg,tload)
c
c...output the results to file
c
      dimension aslod(mtotv),props(msets,mcomp),tload(mtotv),
     1         lnods(melem,mnode),coord(mpoin,2),endis(16),encod(2,8),
     2        shape(8),deriv(2,8),cartd(2,8),dmatx(4,4),  
     3       bmatx(4,16),gploc(2,4),lsidn(melem),smapf(8),
     4      gpwts(4),gpsig(4),gpeps(4),gpcod(2),dmapf(2,8),
     5      nkode(mpoin),asdis(mtotv),lnodn(melem),lpros(melem)
     6     ,lmtyp(melem),ndof(mpoin),nvar(mpoin),strsg(4,4,melem)
c
      real*8 sum
c
c...in axisymmetry, there are 4 stress components, otherwise 3
      if (mstyp.eq.3) then
        nstres = 4
      else
        nstres = 3
      endif
c...output displacements and reactions
c...output to screen (unit 6):
      write(6,*)
      write(6,*)'RESULTS'
      write(6,*)'======='
      write(6,*)
      write(6,610)
  610 format(//' nodal reactions and displacements'
     1/40('-')//1x,'node',
     26x,'x-reaction',5x,'y-reaction',5x,'  moment  ',4x,
     3'x-displt',7x,'y-displt',9x,'rotation')
c...output to results file (unit 16):
      write(16,*)
      write(16,*)'RESULTS'
      write(16,*)'======='
      write(16,*)
      write(16,610)
c...and output to end of data file, for postprocessor:
      write(17,*)'  node  x-displt  y-displt  rotation'
      do 20 np=1,npoin  
        i = nvar(np)
        ndofnp = ndof(np)
        dispx = asdis(i+1)
        dispy = asdis(i+2)
        reacx = aslod(i+1)
        reacy = aslod(i+2)
        if (ndofnp.eq.3) then
          rotat = asdis(i+3)
          bmomt = aslod(i+3)
          write(17,612) np,dispx,dispy,rotat
  612     format(i5,3e10.3)
          write(6,615) np,reacx,reacy,bmomt,dispx,dispy,rotat 
          write(16,615) np,reacx,reacy,bmomt,dispx,dispy,rotat
  615     format(1x,i3,3(5x,e10.3),3(5x,e10.3)) 
  616     format(1x,i3,2(5x,e10.3),15x,2(5x,e10.3))       
        else
          write(17,612) np,dispx,dispy
          write(6,616) np,reacx,reacy,dispx,dispy 
          write(16,616) np,reacx,reacy,dispx,dispy 
        endif
        if (mod(np,25).eq.0) call prompt
   20 continue
c
      call prompt
c...element Gauss pt stressess
  620 format(//'Element Gauss pt. stresses:'/27('-')/)
      write(6,620)
      write(16,620)
      write(17,*)
     1' elt gp    x-coord   y-coord   sigma-x   sigma-y   tau-xy'
      do 100 l=1,nelem
        if (mod(l,3).eq.0) call prompt
        write(6,630) l
        write(16,630) l
  630   format(//' element',i5/1x,12('-')/)
        lset = lpros(l)
        nnode = lnodn(l)
        nside = lsidn(l)
        nodsid = nnode*10 + nside
c...skip, for 1D elements
        if (nside.eq.1) goto 100
        ym = props(lset,1)
        pr = props(lset,2)
c...store nodal coords, displts for elt in encod, endis
        nevab = 0
        i = 0
        do 25 n=1,nnode
          np = lnods(l,n)
          ndofnp = ndof(np)
c...ignore any extra dof's in 2D elements
          if (ndofnp.gt.2) ndofnp = 2
          nevab = nevab + ndofnp
          encod(1,n) = coord(np,1)
          encod(2,n) = coord(np,2)
cp...have to use nvar
          ii = nvar(np)
          endis(i+1) = asdis(ii + 1)
          endis(i+2) = asdis(ii + 2)
          i = i + 2
   25   continue
c
        call dmat(mstyp,ym,pr,dmatx)
c...find Gauss-pt stresses & strains
        write(6,640) (i,i=1,3)
        write(16,640) (i,i=1,3)
  640   format(/' gauss-pt stresses and strains'/6x,'x-coord',8x,
     1  'y-coord',10x,3('stress (',i1,')',5x))
        call gauss(nnode,nside,gploc,gpwts,ngaus)
c...loop over each Gauss point
        do 80 ij=1,ngaus
          s = gploc(1,ij)
          t = gploc(2,ij)
          call sfr(nnode,nside,s,t,shape,deriv)
          IF (NODSID.EQ.53.OR.NODSID.EQ.32.
     1      OR.NODSID.EQ.43) THEN
C...MAPPING FUNCTIONS FOR INFINITE ELEMENT
	      CALL MAPFUN(nnode,nside,s,t,smapf,dmapf)
          ELSE
	      DO 423 N=1,NNODe
	        SMAPF(N) = SHAPE(N)
	        DO 423 M=1,2
	          DMAPF(M,N) = DERIV(M,N)
  423       CONTINUE
          ENDIF
c
          call jacob(nnode,nside,encod,shape,deriv,cartd,djacb
     1               ,smapf,dmapf)
c...get Gauss point coords in gpcod
          do 27 m=1,2
            sum = 0.d0
            do 26 n=1,nnode
   26       sum = sum + dble(encod(m,n))*dble(smapf(n))
            gpcod(m) = sum
   27     continue
          if (mstyp.eq.3) radius = gpcod(1)
          call bmat(nnode,cartd,mstyp,radius,shape,bmatx)
c...find strains as bmatx*displts
          do 30 i=1,nstres
            sum = 0.d0
            do 29 k=1,nevab
   29       sum = sum + dble(bmatx(i,k))*dble(endis(k))
            gpeps(i) = sum
   30     continue
c...stresses = dmatx*strains
          do 40 i=1,nstres
            sum = 0.d0 
            do 39 k=1,nstres
   39       sum = sum + dble(dmatx(i,k))*dble(gpeps(k))
            gpsig(i) = sum + strsg(i,ij,l)
            strsg(i,ij,l) = gpsig(i)
   40     continue
c...if a tensile strength has been input as mat. prop. 4, test:
          lset = lpros(l)
          sigt = props(lset,4)
          mkode = 0
c...find minor principal stress
          sx = gpsig(1)
          sy = gpsig(2)
          tau = gpsig(3)
          smean = 0.5*(sx + sy)
          sdev = 0.5*(sx - sy)
          sroot = sqrt(sdev*sdev + tau*tau)
          sig1 = smean + sroot
          sig3 = smean - sroot
c...if sigt=0.0 it means no strength has been specified
          if (sigt.gt.0.0.and.sig3.lt.-sigt) mkode = 1
c...only stresses printed; modify this to print strains too
          write(17,'(i5,i2,i3,2e10.3,4f10.5)') 
     1       l,ij,mkode,(gpcod(i),i=1,2),(gpsig(i),i=1,nstres)
          write(6,650) (gpcod(i),i=1,2),(gpsig(i),i=1,nstres)
          write(16,650) (gpcod(i),i=1,2),(gpsig(i),i=1,nstres)
          if (mkode.ne.0) then
            write(6,*)'Gauss-pt yielded: minor stress =',sig3
            write(16,*)'Gauss-pt yielded: minor stress =',sig3
          endif
  650     format(3x,2(e10.3,5x),3(5x,f10.5))
   80   continue
c...end of Gauss loop
  100 continue
c...end of element loop
c
      write(6,*)
      write(6,*)'END OF RESULTS'
      write(6,*)
      write(16,*)
      write(16,*)'END OF RESULTS'
      write(16,*)
      return
      end
c
c
c
c...subroutines for STIFF     
c
c
      subroutine gauss(nnode,nside,gploc,gpwts,ngaus)
      dimension gploc(2,4),gpwts(4)
c...set up Gauss pts & weights for reduced integration
c
C
      R13 = 1./SQRT(3.0)
C
      DO 5 J=1,4
        GPWTS(J) = 0.0
        DO 5 I=1,2
          GPLOC(I,J) = 0.0
    5 CONTINUE
C
      nodsid = 10*nnode + nside
c
c...Gauss rules for 1D and joint elements included, for completeness
      if (nodsid.eq.21) then
C           1 POINT INTEG. OF 2 NODED BAR
          ngaus = 1
	  GPWTS(1) = 2.
	  GPLOC(1,1) = 0.
c
C        3 POINT INTEGRATION OF JOINT 
      else if (nodsid.eq.62.or.nodsid.eq.42) then
          ngaus = 3
	  R35 = SQRT(0.6)
	  D59 = 5.0/9.0
          GPLOC(1,1) = -R35
	  GPLOC(1,2) = 0.0
	  GPLOC(1,3) =  R35
	  GPWTS(1) = D59
	  GPWTS(2) = 8.0/9.0
	  GPWTS(3) = D59
C-
C        2 PT INTEGRATION OF quadratic (3-noded) BAR 
      else if (nodsid.eq.31) then
          ngaus = 2
	  GPLOC(1,1) = -R13
	  GPLOC(1,2) =  R13
	  gpwts(1) = 1.0
          gpwts(2) = 1.0
C-
      else if (nodsid.eq.33) then
C           1 POINT INTEG. OF 3 NODED TRIANGLE
          ngaus = 1
	  GPWTS(1) = 0.5
	  GPLOC(1,1) = 1./3.
	  GPLOC(2,1) = 1./3.
c
      else if (nodsid.eq.44.or.nodsid.eq.43) then
C        ONE-POINT INTEGRATION OF LINEAR QUADRILATERAL
          ngaus = 1
	  GPLOC(1,1) = 0.
	  GPLOC(2,1) = 0.
	  GPWTS(1) = 4.0
c
      else if (nodsid.eq.63) then
C        3 PT INTEG. OF TRIANGLE BY MIDSIDE RULE 
          ngaus = 3
	  GPLOC(1,1) = 0.5
	  GPLOC(1,2) = 0.5
	  GPLOC(1,3) = 1.0e-5
	  GPLOC(2,1) = 1.0e-5
	  GPLOC(2,2) = 0.5
	  GPLOC(2,3) = 0.5
	  DO 45 I=1,ngaus
            GPWTS(I) = 1./6.
   45     continue
c
      else if (nodsid.eq.84.or.nodsid.eq.53.or.nodsid.eq.32) then
C        2X2 INTEG. OF QUADRILATERAL, finite or infinite
          ngaus = 4
	  GPLOC(1,1) = -R13
	  GPLOC(1,2) = -R13
	  GPLOC(1,3) =  R13
	  GPLOC(1,4) =  R13
	  GPLOC(2,1) = -R13
	  GPLOC(2,2) =  R13
	  GPLOC(2,3) = -R13
	  GPLOC(2,4) =  R13
          do 50 i=1,4
            gpwts(i) = 1.0
   50     continue
c
      else if (nodsid.eq.64) then
C        3 PT. INTEGRATION OF 6-NODED QUAD IN S-DIRECTION
          NGAUS = 3
	  R35 = SQRT(0.6)
	  D59 = 5.0/9.0
	  GPLOC(1,1) = -R35
	  GPLOC(1,2) = 0.0
	  GPLOC(1,3) = R35
	  GPLOC(2,1) = 0.0
	  GPLOC(2,2) = 0.0
	  GPLOC(2,3) = 0.0
	  GPWTS(1) = 2.*D59
	  GPWTS(2) = 16.0/9.0
	  GPWTS(3) = 2.*D59
C...IF ONLY 2-PT INTEGRATION OF THIN QUADRILATERAL IS
C...REQUIRED, USE CODING BELOW:
C           ngaus = 2
C	    GPLOC(2,1) = 0.
C	    GPLOC(2,2) = 0.
C	    GPLOC(1,1) = -R13
C	    GPLOC(1,2) =  R13
C	    GPWTS(1)   =  2.0
C	    GPWTS(2)   =  2.0
c
      endif
      return
      end
c
c
      subroutine dmat(mstyp,ym,pr,dmatx)
      dimension dmatx(4,4)
      do 5 i=1,4
        do 5 j=1,4
          dmatx(i,j) = 0.0
    5 continue
c...ym,pr are Young's mod., Poisson's ratio
c...g is the elastic shear modulus
      g = 0.5*ym/(1.0 + pr)
      if (mstyp.eq.1) then
c...form the D matrix for plane stress
        f = 0.5*ym/(1.0 - pr)
        dxx = f + g
        dxy = f - g
      else
c...form the D matrix for plane strain
c...assumes thickness = 1.0
        dxy = g*pr/(0.5 - pr)
        dxx = 2.*g + dxy
      endif
      dmatx(1,1) = dxx
      dmatx(1,2) = dxy
      dmatx(2,1) = dxy
      dmatx(2,2) = dxx
      dmatx(3,3) = g
      if (mstyp.eq.3) then
c...extra row and column, for axisymmetric D-matrix
        dmatx(1,4) = dxy
        dmatx(2,4) = dxy
        dmatx(4,1) = dxy
        dmatx(4,2) = dxy
        dmatx(4,4) = dxx
      endif
      return
      end
c
c
      subroutine bmat(nnode,cartd,mstyp,radius,shape,bmatx)
      dimension cartd(2,8),bmatx(4,16),shape(8)
c...form the B matrix
      do 5 j=1,16
      do 5 i=1,4
        bmatx(i,j) = 0.0
    5 continue
c
      j = 0
      do 10 n=1,nnode
        i = j + 1
        j = i + 1
        bmatx(1,i) = -cartd(1,n)
        bmatx(2,j) = -cartd(2,n)
        bmatx(3,i) = -cartd(2,n)
        bmatx(3,j) = -cartd(1,n)
        if (mstyp.eq.3) then
c...final row, for axisymmetry
          bmatx(4,i) = -shape(n)/radius
        endif
c...negative sign for 'compression +ve' sign convention
   10 continue
      return
      end
c
c
c
      subroutine sfr(nnode,nside,s,t,shape,deriv)
      dimension shape(8),deriv(2,8)
c...evaluate the shape functions and 1st derivs at local coords s,t
c
      nodsid = 10*nnode + nside
c
      if (nodsid.eq.62.or.nodsid.eq.31) then
C...3 NODE LINE ELEMENT AND JOINT ELEMENT
      SS = S*S
      SHAPE(1) = (SS-S)*.5
      SHAPE(2) =  1.-SS
      SHAPE(3) = (SS+S)*.5
      DERIV(1,1) = S-.5
      DERIV(1,2) = -S*2.
      DERIV(1,3) = S+.5
      RETURN
C
      else if (nodsid.eq.21) then
C...2-NODE LINE ELEMENT
      SHAPE(1) = (1.-S)/2.
      SHAPE(2) = (1.+S)/2.
      DERIV(1,1) = -0.5
      DERIV(1,2) =  0.5
      RETURN
C
      else if (nodsid.eq.33) then
C...LINEAR TRIANGLE 
      SHAPE(1) = 1.-S-T
      SHAPE(2) = S
      SHAPE (3) = T
      DERIV(1,1) = -1.
      DERIV(1,2) =  1.
      DERIV(1,3) =  0.
      DERIV(2,1) = -1.
      DERIV(2,2) =  0.
      DERIV(2,3) =  1.
      RETURN
C
      else if (nodsid.eq.44) then
C...LINEAR QUADRILATERAL
      ST = S*T
      SHAPE(1) = (1.-S-T+ST)*.25
      SHAPE(2) = (1.+S-T-ST)*.25
      SHAPE(3) = (1.+S+T+ST)*.25
      SHAPE(4) = (1.-S+T-ST)*.25
      DERIV(1,1) = -(1.-T)*.25
      DERIV(1,2) =  (1.-T)*.25
      DERIV(1,3) =  (1.+T)*.25
      DERIV(1,4) = -(1.+T)*.25
      DERIV(2,1) = -(1.-S)*.25
      DERIV(2,2) = -(1.+S)*.25
      DERIV(2,3) =  (1.+S)*.25
      DERIV(2,4) =  (1.-S)*.25
      RETURN
C
      else if (nodsid.eq.63) then
C...6 NODED TRIANGLE
      ST = S+T
      S2T = 2.*S*T
      SHAPE(1) = 1.-ST*(3.-2.*ST)
      SHAPE(2) = 4.*S*(1.-ST)
      SHAPE(3) = S*(2.*S-1.)
      SHAPE(4) = 2.*S2T
      SHAPE(5) = T*(2.*T-1.)
      SHAPE(6) = 4*T*(1.-ST)
      DERIV(1,1) = -3.+4.*ST
      DERIV(1,2) = 4.-8.*S-4.*T
      DERIV(1,3) = 4.*S-1.
      DERIV(1,4) = 4.*T
      DERIV(1,5) = 0.
      DERIV(1,6) = -4.*T
      DERIV(2,1) = -3.+4.*ST
      DERIV(2,2) = -4.*S
      DERIV(2,3) = 0.
      DERIV(2,4) = 4.*S
      DERIV(2,5) = 4.*T-1.
      DERIV(2,6) = 4.-4.*S-8.*T
      RETURN
C
c
      else if (nodsid.eq.64) then
C...SIX-NODED QUADRILATERAL
      SM1 = (1.-S)/2.
      SP1 = (1.+S)/2.
      TM1 = (1.-T)/2.
      TP1 = (1.+T)/2.
      SS1 = 1. - S*S
      SHAPE(1) = -S*SM1*TM1
      SHAPE(2) =  SS1*TM1
      SHAPE(3) =  S*SP1*TM1
      SHAPE(4) =  S*SP1*TP1
      SHAPE(5) =  SS1*TP1
      SHAPE(6) = -S*SM1*TP1
      DERIV(1,1) = -(0.5-S)*TM1
      DERIV(1,2) = -S*(1.-T)
      DERIV(1,3) =  (0.5+S)*TM1
      DERIV(1,4) =  (0.5+S)*TP1
      DERIV(1,5) = -S*(1.+T)
      DERIV(1,6) = -(0.5-S)*TP1
      S2 = 0.5*S
      DERIV(2,1) =  S2*SM1
      DERIV(2,2) = -0.5*SS1
      DERIV(2,3) = -S2*SP1
      DERIV(2,4) =  S2*SP1
      DERIV(2,5) =  0.5*SS1
      DERIV(2,6) = -S2*SM1
      RETURN
C
      else if (nodsid.eq.84) then
C...8 NODED QUADRILATERAL
      S2 = S*2.
      T2 = T*2.
      SS = S*S
      TT = T*T
      ST = S*T
      SST= SS*T
      STT= S*TT
      ST2= ST*2.
      SHAPE(1)  =  (-1.0+ST+SS+TT-SST-STT)*.25
      SHAPE(2)  =  (1.0-T-SS+SST)*.5
      SHAPE(3)  =  (-1.0-ST+SS+TT-SST+STT)*.25
      SHAPE(4)  =  (1.0+S-TT-STT)*.5
      SHAPE(5)  =  (-1.0+ST+SS+TT+SST+STT)*.25
      SHAPE(6)  =  (1.0+T-SS-SST)*.5
      SHAPE(7)  =  (-1.0-ST+SS+TT+SST-STT)*.25
      SHAPE(8)  =  (1.0-S-TT+STT)*.5
      DERIV(1,1) = (T+S2-ST2-TT)*.25
      DERIV(1,2) = -S+ST
      DERIV(1,3) = (-T+S2-ST2+TT)*.25
      DERIV(1,4) = (1.0-TT)*.5
      DERIV(1,5) = (T+S2+ST2+TT)*.25
      DERIV(1,6) = -S-ST
      DERIV(1,7) = (-T+S2+ST2-TT)*.25
      DERIV(1,8) = (-1.0+TT)*.5
      DERIV(2,1) = (S+T2-SS-ST2)*.25
      DERIV(2,2) = (-1.0+SS)*.5
      DERIV(2,3) = (-S+T2-SS+ST2)*.25
      DERIV(2,4) = -T-ST
      DERIV(2,5) = (S+T2+SS+ST2)*.25
      DERIV(2,6) = (1.0-SS)*.5
      DERIV(2,7) = (-S+T2+SS-ST2)*.25
      DERIV(2,8) = -T+ST
      RETURN
C
      else if (nodsid.eq.53) then
C...QUADRATIC INFINITE ELT: 8-NODED QUAD, WITH 3 NODES (ETA=1)
C   AT INFINITY
      S2 = S*2.
      T2 = T*2.
      SS = S*S
      TT = T*T
      ST = S*T
      SST= SS*T
      STT= S*TT
      ST2= ST*2.
      SHAPE(1)  =  (-1.0+ST+SS+TT-SST-STT)*.25
      SHAPE(2)  =  (1.0-T-SS+SST)*.5
      SHAPE(3)  =  (-1.0-ST+SS+TT-SST+STT)*.25
      SHAPE(4)  =  (1.0+S-TT-STT)*.5
      SHAPE(5)  =  (1.0-S-TT+STT)*.5
      DERIV(1,1) = (T+S2-ST2-TT)*.25
      DERIV(1,2) = -S+ST
      DERIV(1,3) = (-T+S2-ST2+TT)*.25
      DERIV(1,4) = (1.0-TT)*.5
      DERIV(1,5) = (-1.0+TT)*.5
      DERIV(2,1) = (S+T2-SS-ST2)*.25
      DERIV(2,2) = (-1.0+SS)*.5
      DERIV(2,3) = (-S+T2-SS+ST2)*.25
      DERIV(2,4) = -T-ST
      DERIV(2,5) = -T+ST
      RETURN
C
      else if (nodsid.eq.32) then
C...3-NODED DOUBLY INFINITE ELEMENT
      S2 = S*2.
      T2 = T*2.
      SS = S*S
      TT = T*T
      ST = S*T
      SST= SS*T
      STT= S*TT
      ST2= ST*2.
      SHAPE(1)  =  (-1.0+ST+SS+TT-SST-STT)*.25
      SHAPE(2)  =  (1.0-T-SS+SST)*.5
      SHAPE(3)  =  (1.0-S-TT+STT)*.5
      DERIV(1,1) = (T+S2-ST2-TT)*.25
      DERIV(1,2) = -S+ST
      DERIV(1,3) = (-1.0+TT)*.5
      DERIV(2,1) = (S+T2-SS-ST2)*.25
      DERIV(2,2) = (-1.0+SS)*.5
      DERIV(2,3) = -T+ST
      RETURN
C
      else if (nodsid.eq.43) then
C...FOUR-NODED INFINITE ELT, NODES AT XI=1 AT INFINITY
      SM1 = (1.-S)/2.
      SP1 = (1.+S)/2.
      TM1 = (1.-T)/2.
      TP1 = (1.+T)/2.
      SS1 = 1. - S*S
      SHAPE(2) = -S*SM1*TM1
      SHAPE(3) =  SS1*TM1
      SHAPE(4) =  SS1*TP1
      SHAPE(1) = -S*SM1*TP1
      DERIV(1,2) = -(0.5-S)*TM1
      DERIV(1,3) = -S*(1.-T)
      DERIV(1,4) = -S*(1.+T)
      DERIV(1,1) = -(0.5-S)*TP1
      S2 = 0.5*S
      DERIV(2,2) =  S2*SM1
      DERIV(2,3) = -0.5*SS1
      DERIV(2,4) =  0.5*SS1
      DERIV(2,1) = -S2*SM1
      RETURN
C
      else if (nodsid.eq.42) then
C...INFINITE JOINT ELEMENT
        SS = S*S
        SHAPE(1) = (SS-S)*0.5
        SHAPE(2) = 1.-SS
        DERIV(1,1) = S - 0.5
        DERIV(1,2) = -2.0*S
        RETURN
      endif
C
      END
c
c
c
      subroutine mapfun(nnode,nside,s,t,smapf,dmapf)
      dimension smapf(8),dmapf(2,8)
c
C...MAPPING FUNCTIONS FOR INFINITE ELT (BASED ON 8-NODED QUAD)
      nodsid = nnode*10 + nside
C
      SM = 1.-S
      TM = 1.-T
      SP = 1.+S
      TP = 1.+T
      SM2 = SM*SM
      TM2 = TM*TM
      SMTM = SM*TM
      SM2TM = SM*SMTM
      SMTM2 = SMTM*TM
C
      if (nodsid.eq.53) then
C...5-NODED INF. ELT.
        SMAPF(1) = SM*(-SP-T)/TM
        SMAPF(2) = 2.*(1.-S*S)/TM
        SMAPF(3) = SP*(-SM-T)/TM
        SMAPF(4) = 0.5*TP*SP/TM
        SMAPF(5) = 0.5*TP*SM/TM
        DMAPF(1,1) = (2.*S+T)/TM
        DMAPF(1,2) = -4.*S/TM
        DMAPF(1,3) = (2.*S-T)/TM
        DMAPF(1,4) = 0.5*TP/TM
        DMAPF(1,5) = -0.5*TP/TM
        DMAPF(2,1) = -SM*(2.+S)/TM2
        DMAPF(2,2) = 2.*(1.-S*S)/TM2
        DMAPF(2,3) = -SP*(2.-S)/TM2
        DMAPF(2,4) = SP/TM2
        DMAPF(2,5) = SM/TM2
        RETURN
C
      else if (nodsid.eq.32) then
C...THREE-NODED DOUBLY INFINITE ELEMENT
        SMAPF(1) = (S*T - 3.*(1.+S+T))/SMTM
        SMAPF(2) = 2.*SP/SMTM
        SMAPF(3) = 2.*TP/SMTM
        DMAPF(1,1) = -2.*(T+3.)/SM2TM
        DMAPF(1,2) = 4./SM2TM
        DMAPF(1,3) = 2.*TP/SM2TM
        DMAPF(2,1) = -2.*(S+3.)/SMTM2
        DMAPF(2,2) = 2.*SP/SMTM2
        DMAPF(2,3) = 4./SMTM2
        RETURN
C
      else if (nodsid.eq.43) then
C...FOUR-NODED INF ELT, NODES AT XI=1 AT INFINITY
        SMAPF(1) = -S*TP/SM
        SMAPF(2) = -S*TM/SM
        SMAPF(3) = 0.5*TM*SP/SM
        SMAPF(4) = 0.5*TP*SP/SM
        DMAPF(1,1) = TP/SM2
        DMAPF(1,2) = TM/SM2
        DMAPF(1,3) = TM/SM2
        DMAPF(1,4) = TP/SM2
        DMAPF(2,1) = -S/SM
        DMAPF(2,2) =  S/SM
        DMAPF(2,3) = -0.5*SP/SM
        DMAPF(2,4) =  0.5*SP/SM
      RETURN
C
      else if (nodsid.eq.42) then
C...INFINITE JOINT ELEMENT (ONE-DIMENSIONAL)
        SMAPF(1) = -2.*S/SM
        SMAPF(2) = SP/SM
        DMAPF(1,1) = -2.0/SM2
        DMAPF(1,2) = 2.0/SM2
        RETURN
      endif
C
      return
      END
c
c     
      subroutine jacob(nnode,nside,encod,shape,deriv,cartd,djacb,
     1                 smapf,dmapf)
      dimension shape(8),deriv(2,8),cartd(2,8),encod(2,8),xjacm(2,2),
     1         xjaci(2,2),smapf(8),dmapf(2,8)
c...evaluate the Jacobian matrix, and its determinant
      real*8 sum,dj,xj
C
      nodsid = nnode*10 + nside
      ndime = 2
c...distinguish between 1D and 2D elements 
c...(although not used in 1D elts in this program)
      NJDIM = 2
      IF (nside.eq.1) NJDIM = 1
c
C...CREATE JACOBIAN MATRIX
      DO 5 J=1,ndime
      DO 5 I=1,NJDIM
      SUM = 0.d0
      DO 4 N=1,NNODE
    4 SUM = SUM + dble(DMAPF(I,N))*dble(ENCOD(J,N))
      XJACM(I,J) = SUM
    5 CONTINUE
C
C...CALCULATE DETERMINANT AND INVERSE JACOBIAN
C
      if (njdim.eq.1) then
C...LINE OR JOINT ELEMENT 
cb...not used in eladv
        DJ = 0.d0
        DO 12 I=1,ndime
          XJ = XJACM(1,I)
          DJ = DJ + XJ*XJ
   12   continue
        DJACB = SQRT(DJ)
        DO 15 N=1,NNODE
	  CARTD(1,N) = DERIV(1,N)/DJACB
          CARTD(2,N) = 0.0
   15   continue
C...CARTD CONTAINS DN/DS
c
      else
C
C...AREA ELEMENT 
        DJACB = XJACM(1,1)*XJACM(2,2) - XJACM(1,2)*XJACM(2,1)
        if (djacb.lt.1.0e-12) then
          write(6,*)'ERROR: determinant of Jacobian is -ve or zero'
        write(6,*)'Element is inside out, or nodes defined clockwise'
              call prompt
          stop
        endif
        XJACI(1,1) = XJACM(2,2)/DJACB
        XJACI(2,2) = XJACM(1,1)/DJACB
        XJACI(1,2) =-XJACM(1,2)/DJACB
        XJACI(2,1) =-XJACM(2,1)/DJACB
c
C...CALCULATE CARTESIAN DERIVATIVES
        DO 50 I=1,NDIME
          DO 50 N=1,NNODE
            SUM = 0.d0
            DO 49 J=1,NJDIM
   49         SUM = SUM + dble(XJACI(I,J))*dble(DERIV(J,N))
            CARTD(I,N) = SUM
   50   CONTINUE
      endif
      RETURN
      END
c
c
c
      subroutine input(ichan,melem,mpoin,mnode,msets,mcomp,nelem,
     1      npoin,lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props
     2     ,ndofn,npros,title,ndof)
c
c...this reads basic mesh data 
c
      implicit integer(i-o)
      dimension lnods(melem,mnode),lnodn(melem),lsidn(melem)
     1,nkode(mpoin),lpros(melem),coord(mpoin,2),cordn(2),
     2props(msets,mcomp),lmtyp(melem),ndof(mpoin)
c
c
      character*80 title
c
      rewind (ichan)
      read(ichan,501,err=99,end=199) title
      write(6,*) 'TITLE: ',title
c
      read(ichan,*)
      read(ichan,401,err=99,end=199) npoin,nelem,ndofn,npros
      if (npoin.gt.mpoin) then
        write(6,*)'ERROR: ',npoin,' nodes exceeds maximum.'
        write(6,*)'Increase MPOIN and recompile.'
              call prompt
        stop
      endif
      if (nelem.gt.melem) then
        write(6,*)'ERROR: ',nelem,' elements exceeds maximum.'
        write(6,*)'Increase MELEM and recompile.'
              call prompt
        stop
      endif
c
      do 11 i=1,melem
        lmtyp(i) = 0
        lnodn(i) = 0
        lsidn(i) = 0
        lpros(i) = 0
        do 11 j=1,mnode
          lnods(i,j) = 0
   11 continue
      do 12 i=1,mpoin
        nkode(i) = 0
        do 12 j=1,2
          coord(i,j) = 0.0
   12 continue
      do 13 i=1,msets
        do 13 j=1,mcomp
          props(i,j) = 0.0
   13 continue
c
c...read original elt-node connections
      k = 0
      read(ichan,*)
   15 k = k + 1
      read(ichan,421,err=99,end=199) lcard,nside,lmtype,nset,
     1lnum,(lnods(lnum,n),n=1,mnode)
      if (nside.eq.0) nside=4
c...default element material type is 1
      if (lmtype.eq.0) lmtype = 1
c...default property set is 1
      if (nset.eq.0) nset = 1
c...count no. of nodes in element
      nnode = 0
      do 21 n=1,mnode
        if (lnods(lnum,n).eq.0) goto 22
        nnode = nnode + 1
   21 continue
   22 lsidn(lnum) = nside
      lnodn(lnum) = nnode
      lmtyp(lnum) = lmtype
      lpros(lnum) = nset
      if (lcard.eq.0) goto 15
      if (k.ne.nelem) then
        write(6,*) 'WARNING: no. of elements differs from NELEM'
        write(6,*) 'NELEM changed to ',k
        nelem = k
      endif
c
c...read nodal coord cards, & store in coord
      read(ichan,*)
   23 read(ichan,441,err=99,end=199) lcard,kode,np,(cordn(i),i=1,2)
     1,ndofnp
c...default is 2 d.o.f.s at node
      if (ndofnp.eq.0) ndofnp = 2
      if (np.gt.npoin) then
         write(6,*) 'WARNING: node no. ',np,' exceeds NPOIN'
         write(6,*) 'NPOIN increased to ',np
         npoin = np
      endif
      if (np.le.0.or.npoin.gt.mpoin) then
         write(6,*) 'ERROR: node no.',np,' exceeds MPOIN or is negative'
              call prompt
         stop
      endif
      do 24 i=1,2
        coord(np,i) = cordn(i)
   24 continue
      nkode(np) = kode
      ndof(np) = ndofnp
      if (lcard.eq.0) goto 23
c
c
      read(ichan,*)
   30 read(ichan,491,err=99,end=199) lcard,nset,icomp,value
      if (nset.gt.0) props(nset,icomp) = value
      if (lcard.eq.0) goto 30
c
      return
   99 write(6,*)'ERROR: wrong format in data file'
              call prompt
      stop
  199 write(6,*)'ERROR: end of file encountered during input'
              call prompt
      stop 
c
c
c...input formats
  501 format (a80)
  401 FORMAT(12I5)
  421 FORMAT(I1,I4,I2,I3,10I5)
  441 FORMAT(I1,I4,I5,2g10.3,I5)
  491 format(i1,i4,i5,10x,g10.3)
c
      end
c
c
c
c
      subroutine output(ichan,melem,mpoin,mnode,msets,mcomp,
     1            nelem,npoin,lnods,lnodn,lsidn,lmtyp,coord,
     2      nkode,lpros,props,ndofn,npros,title,ndof)
      character title*50
c
      dimension lnods(melem,mnode),lnodn(melem),lsidn(melem)
     1,coord(mpoin,2),nkode(mpoin),lmtyp(melem)
     2,lpros(melem),props(msets,mcomp),ndof(mpoin)
c
      write(ichan,600) title
c
      write(ichan,*)'  npoin nelem ndofn npros'
      write(ichan,612) npoin,nelem,ndofn,npros
c
      lcard = 0
      write(ichan,*)'nside typ/set elt       nodes'  
      do 155 l=1,nelem
        if (l.eq.nelem) lcard=1
        nnode = lnodn(l)
        nside = lsidn(l)
        lmtype = lmtyp(l)
        nset = lpros(l)
        write(ichan,621) lcard,nside,lmtype,nset,l,
     1                   (lnods(l,n),n=1,nnode)
  155 continue
c
      lcard = 0
      write(ichan,*)'  kode node  x-coord  y-coord ndof'
      do 160 n=1,npoin
        if (n.eq.npoin) lcard=1
        kode = nkode(n)
        ndofnp = ndof(n)
        if (ndofnp.eq.1) then
        write(ichan,641) lcard,kode,n,(coord(n,i),i=1,2),ndofnp
        else if (ndofnp.eq.2) then
        write(ichan,642) lcard,kode,n,(coord(n,i),i=1,2),ndofnp
        else if (ndofnp.eq.3) then
        write(ichan,643) lcard,kode,n,(coord(n,i),i=1,2),ndofnp
        else 
        write(ichan,644) lcard,kode,n,(coord(n,i),i=1,2),ndofnp
        endif
  160 continue
c
c...write mat props to file
      lcard = 0
      write(ichan,*)'  set cmpnt        value'
      do 175 nset=1,npros
        do 175 icomp=1,mcomp
          value = props(nset,icomp)
          if (abs(value).gt.0.0) then
            write(ichan,691) lcard,nset,icomp,value
          endif
  175 continue
      lcard = 1
      write(ichan,691) lcard
c
  600 format(a50)
  612 format(5I5)
  621 FORMAT(I1,I4,I2,I3,10I5)
  641 format(i1,3x, i1, i5,2f10.5,i5)
  642 format(i1,2x,i2.2,i5,2f10.5,i5)
  643 format(i1,1x,i3.3,i5,2f10.5,i5)
  644 format(i1,   i4.4,i5,2f10.5,i5)
  691 format(i1,i4,3x,i2,10x,e10.3)
      return
      end
c
c
c
      subroutine print(ichan,melem,mpoin,mnode,msets,mcomp,
     1nelem,npoin,lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,
     2            props,ndofn,npros,title,ndof)
      implicit integer(i-o)
      character title*50
      dimension lnods(melem,mnode),lnodn(melem),lsidn(melem)
     1,coord(mpoin,2),nkode(mpoin),lmtyp(melem),ndof(mpoin)
     2,lpros(melem),props(msets,mcomp),kard(14)
c
      write(ichan,*)
      write(ichan,*)'MESH DETAILS'
      write(ichan,*)'============'
      write(ichan,*)
c
        write(ichan,*)
        write(ichan,*)
        write(ichan,*)'Title is:'
        write(ichan,*)
        write(ichan,*) title
        write(ichan,*)
        write(ichan,*)
c
        write(ichan,*)'Integer problem constants'
        write(ichan,*)'-------------------------'
        write(ichan,*)
	write(ichan,603)'npoin = ',npoin,'   nelem = ',nelem
        write(ichan,*)
        if (abs(ndofn).eq.2) 
     1        write(ichan,*)'Elasticity problem'
        if (abs(ndofn).eq.1) write(ichan,*)'Poisson/Laplace problem'
        if (ndofn.lt.0) then
          write(ichan,*)'Advanced-level analysis'
        else
          write(ichan,*)'Basic-level analysis'
        endif
  603   format(a8,i5,a11,i5)
	write(ichan,*)
	write(ichan,*)
c
      if(ichan.eq.6) call prompt
      write(ichan,*) 'Element-node connection data'
      write(ichan,*) '----------------------------'
      write(ichan,*)
     1'Element  No. of  No. of  Matl     Nodes around element'
      write(ichan,*)
     1'number   sides   nodes   type'
      write(ichan,*)
      do 30 lnum=1,nelem
        if (ichan.eq.6.and.mod(lnum,25).eq.0) call prompt
        nside = lsidn(lnum)
        nnode = lnodn(lnum)
        lmtype = lmtyp(lnum)
	write(ichan,612)lnum,nside,nnode,lmtype,
     1                  (lnods(lnum,n),n=1,nnode)
  612   format(i5,i7,i8,i8,5x,8i5)
   30 continue
      write(ichan,*)
c
      if (ichan.eq.6) call prompt
      write(ichan,*)'Nodal Coordinates'
      write(ichan,*)'-----------------'
      write(ichan,*)
      iflag = 0
      write(ichan,*)' Node   Coordinates      ndof'
      write(ichan,*)
      do 55 n=1,npoin
        ndofnp = ndof(n)
      if (ichan.eq.6.and.mod(n,25).eq.0) call prompt
      write(ichan,'(i5,2(3x,f10.3),i5)')n,(coord(n,m),m=1,2)
     1,ndofnp
   55 continue
c
      if (ichan.eq.6) call prompt
c...note that this coding for displaying fixities is only
c...valid if there are 1 or 2 dof's per node!
      if (abs(ndofn).eq.2) then
        write(ichan,*)'Fixed degrees of freedom:'
        write(ichan,*)'-------------------------'
        write(ichan,*)
        write(ichan,*)
        write(ichan,*)'Fixed in x-direction (KODE=10):'
        write(ichan,*)'-------------------------------'
        kode = 10
        nin = 1
   71   call scan(kard,nin,nout,kode,nkode,npoin)
        write(ichan,'(14i5)')(kard(i),i=1,14)
        nin = nout + 1
        if (nin.le.npoin) goto 71
        write(ichan,*)
        write(ichan,*)'Fixed in y-direction (KODE=01):'
        write(ichan,*)'-------------------------------'
        kode = 01
        nin = 1
   73   call scan(kard,nin,nout,kode,nkode,npoin)
        write(ichan,'(14i5)')(kard(i),i=1,14)
        nin = nout + 1
        if (nin.le.npoin) goto 73
        write(ichan,*)'Fixed in both directions (KODE=11):'
        write(ichan,*)'-----------------------------------'
        kode = 11
        nin = 1
   75   call scan(kard,nin,nout,kode,nkode,npoin)
        write(ichan,'(14i5)')(kard(i),i=1,14)
        nin = nout + 1
        if (nin.le.npoin) goto 75
      else
        write(ichan,*)'Dirichlet nodes (KODE=1):'
        write(ichan,*)'-------------------------'
        kode = 1
        nin = 1
   77   call scan(kard,nin,nout,kode,nkode,npoin)
        write(ichan,'(14i5)')(kard(i),i=1,14)
        nin = nout + 1
        if (nin.le.npoin) goto 77
      endif
c
      if (ichan.eq.6) call prompt
      write(ichan,*)'Material Properties'
      write(ichan,*)'-------------------'
      write(ichan,*)
      iflag = 0
      do 90 nset=1,npros
      write(ichan,*)
      write(ichan,*)'Elts in mat. prop. set ',nset,':'
      nin = 1
   85 call scan(kard,nin,nout,nset,lpros,nelem)
      write(ichan,'(14i5)')(kard(i),i=1,14)
      nin = nout + 1
      if (nin.le.nelem) goto 85
   90 continue
      write(ichan,*)
      write(ichan,*)'The non-zero mat. properties are:'
      do 327 iset=1,npros
      write(ichan,*)
      write(ichan,*)'Set number ',iset
      write(ichan,*)
      do 327 j=1,mcomp
      val = props(iset,j)
      if (abs(val).le.0.0) goto 327
      write(ichan,*)'Component ',j,' has value ',val
  327 continue
      write(ichan,*)
      write(ichan,*)'*** End of mesh details ***'
      write(ichan,*)
      write(ichan,*)
      return
      end
c      
c
c
c
      function lenstr(string)
      character*(*) string
c...finds the length of a character string
      lc = len(string)
      do 10 i=lc,1,-1
        if (string(i:i).ne.' ') then
          lenstr = i
          return
        endif
   10 continue
      lenstr = 0
      return
      end 
c
c
      subroutine scan(kard,lin,lout,lvalue,list,ldim)
      dimension list(ldim),kard(14)
c
      lout = lin
      ki = 0
      do 5 i=1,14
      kard(i) = 0
    5 continue
      do 20 li=lin,ldim
      lout = li
      lvali = list(li)
      if (lvali.eq.lvalue) then
         ki = ki + 1
         if (ki.le.14) then
            kard(ki) = li
         else
            lout = lout - 1
            return
         endif
      endif
   20 continue
      return
      end
c
c
      subroutine yesno(char,askim)
      character char*1
      character*(*) askim
c
    5 lc = lenstr(askim)
      write(6,*) askim(1:lc)
      read(5,'(a1)') char
      if (char.eq.'y'.or.char.eq.'Y') then
        char = 'y'
      else if (char.eq.'n'.or.char.eq.'N') then
        char = 'n'
      else
       write(6,*)'Type y for yes or n for no: '
       goto 5
      endif
      return
      end
c
c
c
c
      subroutine askfil(infile,iflag)
      character*80 infile,intemp
c
      iflag = 0
      write(6,*)'Type the name of the .dat input file: '
      read(5,500) intemp
  500 format(a8)
      lenin = lenstr(intemp)
      infile = intemp(1:lenin)//'.dat'
      lenin = lenin + 4
      return
      end
c
c
c
      subroutine prompt
      character*1 char
      write(6,*) 'Press Return to continue!'
      read(5,'(a1)') char
      return
      end
c
c