c...Plane beam analysis "main engine"
c...for use with FELIPE version 3
c... - sample datafile is framexa1.dat or framexb1.dat
c
c
c...If you want to re-compile this program, you must link it with
c...the file felsb.obj, compiled from felsb.for (provided in the
c...FELIPE folder). The file felsb.for contains additional subroutines 
c...used by all the 'main engines' except eladv.for. 
c...If you don't use linking, you should paste the contents of 
c...the file felsb.for to the end of this file, before compiling
c...and running it.
c
      program frame
c
      parameter(melem=200,mpoin=400,mnode=2,
     1msets=9,mcomp=3,mtotv=1200)
c
c...these parameters fix the dimensions of the arrays:
c...melem = max. no. of elements
c...mpoin = max. no. of nodes
c...mnode = max. no. of nodes in a single element
c...msets = max. no. of mat. prop. sets (also used for loading sets)
c...mcomp = max. no. of mat. props (e.g. Young's mod)
c...mtotv = max no. of degrees of freedom (= 3*mpoin for beams)
c
c
      character char*1
      character*80 title,infile,outfil,prtfil
c
      dimension lnods(melem,mnode),lnodn(melem),lsidn(melem),
     1  coord(mpoin,2),nkode(mpoin),aslod(mtotv),ndof(mpoin),
     2        lpros(melem),props(msets,mcomp),asdis(mtotv),
     4      surft(msets,2),lmtyp(melem),diagk(mtotv),
     6      nvar(mpoin),z(mtotv),r(mtotv),p(mtotv),y(mtotv)
c...lnods(l,n) = global no. of n'th node of element l
c...lnodn(l) = no. of nodes in element l ( = 2 for beam element)
c...lsidn = no. of sides in element l    ( = 1 for beam element)
c...coord(np,1), coord(np,2) = x,y coords of node np
c...nkode(np) = fixity code of node np
c...aslod is global nodal load vector
c...ndof(np) = no. of dof.s at node np 
cb...nvar(np) indicates the global d.o.f. at which node np starts
c...lpros(l) = mat. prop. set no. of element l
c...props(m,i) = value of i'th mat. property in mat. prop. set m
c...asdis is global nodal displacement vector
c...astmp is a scratch vector used in the equation solution substitutions
c...surft(m,1),surft(m,2) = normal,tangential surface tractions in set m
c...gstif holds assembled global stiffness matrix as a vector in skyline storage
c...      or as a matrix using band storage
c...nminvc(i)=j means that G(i,j) is the first nonzero entry in row i of gstif
c...lpoint(i)=ii means that G(i,i) is stored at gstif(ii) in skyline storage
c...lmtyp(l) = material type of elt l (not used in this program)
c...diagk holds the diagonal of the global stiffness matrix, to use as
c...     preconditioner in the conjugate gradient solution
c...z,r,p,y are scratch vectors used during the c.g. solution 
c
      write(6,*)'Program FRAME: solves beam and frame problems.'
      write(6,*)'Handles 2-noded cubic beam elements.'
      write(6,*)'Dimensioned for',melem,' elements, and',mpoin,' nodes.'
      write(6,*)
c
c...read an existing .dat file
c
    1 continue
c
c...get the input file name in infile
      call askfil(infile,iflag)
      if (iflag.eq.1) then
        call yesno(char,'Do you want to exit?')
        if (char.eq.'y') stop
        goto 1
      endif
c
      write(6,*)'Opening file ',infile,'...'
      lenin = lenstr(infile)
      open(15,file=infile(1:lenin),status='old',err=999)
      write(6,*)
      lm4 = lenin - 4
      prtfil = infile(1:lm4)//'.prt'
      write(6,*)'Creating file ',prtfil(1:lenin),' for results...'
      write(6,*)
      open(16,file=prtfil(1:lenin),status='unknown',err=999)      
      outfil = infile(1:lm4)//'.out'
      write(6,*)'Creating file ',outfil(1:lenin),' for output...'
      write(6,*)
      open(17,file=outfil(1:lenin),status='unknown',err=999)      
c
c...read the input mesh data, and echo it to the output files
      call input(15,melem,mpoin,mnode,msets,mcomp,nelem,
     1 npoin,lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props
     2         ,ndofn,npros,title,ndof)
      call print(16,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1 lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2           npros,title,ndof)
      call output(17,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1  lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2            npros,title,ndof)
      call yesno(char,'Do you want the mesh details listed?')
      if (char.eq.'y') then
        call print(6,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1  lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2           npros,title,ndof)
      endif
c
      iflag = 0
      do 2 l=1,nelem
        nnode = lnodn(l)
        nside = lsidn(l)
        nodsid = 10*nnode + nside
        if (nodsid.ne.21) then
          iflag = 1
          write(6,*)'ERROR: Cannot handle element ',l
          write(6,*)'which has ',nside,' sides and ',nnode,' nodes'
        endif
    2 continue
      if (iflag.eq.1) then
        call prompt
        stop
      endif
c
      cgtol = 1.e-5
c
cp...create the pointer array nvar. Needed because a node may
cp...have 2 or 3 d.o.f.s; this info stored in ndof
      nvar(1) = 0
      ntotv = 0
      do 5 np=1,npoin
        ndofnp = ndof(np)
        nvar(np) = ntotv
        ntotv = ntotv + ndofnp
    5 continue
cp...Now local dof i of node np is global dof nvar(np)+i
cp...and ntotv is total no. of degrees of freedom
c
c
c...open a scratch file to write the element stiffness matrices on
      open(unit=18,status='scratch',form='unformatted')  
      write(6,*)'Scratch file opened ...'
      write(6,*)
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine STIFF...'
      call stiff(melem,mpoin,mnode,msets,mcomp,mstyp,
     1           nelem,npoin,lnods,lnodn,coord,lpros,props)
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine LOAD...'
      call load(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1          nelem,npoin,lnods,lnodn,coord,
     2          surft,aslod,lpros,props)
c
      call prompt
c...store the diagonal of K, to use as the preconditioner
      write(6,*)
      write(6,*)'Starting subroutine ASSDK...'
      call assdk(melem,mpoin,mnode,mtotv,
     1    nelem,npoin,lnods,lnodn,ndof,nvar,nkode,diagk)
c
c...solve using conjugate gradients!
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine PCGSOL...'
      call pcgsol(mtotv,melem,mpoin,mnode,ntotv,nelem,
     1           npoin,lnods,lnodn,diagk,nkode,ndof,nvar,
     2    aslod,asdis, z , r , p , y ,cgtol,itscg,rnorm0)
c
c...calculate nodal reactions by forming gstif*asdis in diagk,
c...then add this to negative of equivalent nodal loads in aslod
      call prompt
      write(6,*)
      write(6,*)'Calculating reactions at nodes...'
      call ematgv(18,mtotv,melem,mpoin,mnode,ntotv,nelem,
     1         npoin,lnods,lnodn,ndof,nvar,asdis,diagk)
      do 10 i=1,ntotv
        aslod(i) = -aslod(i) + diagk(i)
   10 continue
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine RESULT...'
      call result(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1            nelem,npoin,lnods,lnodn,coord,nkode,lpros,props,
     2            aslod,asdis)
c
      write(6,*)
      write(6,*)'FRAME completed. Results are in file ',prtfil(1:lenin)
      write(6,*)
      write(6,*)
      close(15)
      close(16)
      close(17)
      call prompt
      stop
  999 write(6,*)
     1'Cannot open file with this name. Terminating program.'
      call prompt
      stop
      end
c
c
c      
      subroutine stiff(melem,mpoin,mnode,msets,mcomp,mstyp,
     1             nelem,npoin,lnods,lnodn,coord,lpros,props)
c      
c...calculates the element stiffness matrices, and writes them
c...to the scratch file.
c
      dimension props(msets,mcomp),lnods(melem,mnode),encod(2,8),
     1         estif(6,6),
     2        lpros(melem),lnodn(melem),coord(mpoin,2)
c...encod holds the nodal coords for the nodes around the current element
c...estif will be the stiffness matrix for the current element
c
      rewind 18
c
c...start loop over all elements
      do 80 l=1,nelem
c
c...extract Young's modulus be and moment of inertia bi from props      
      lset = lpros(l)
      be = props(lset,1)
      bi = props(lset,2)
c...also extract cross-sectional area ba of beam
      ba = props(lset,3)
      if (ba.eq.0.0) ba = 1.0
c
c...put nodal coords in encod
      nnode = lnodn(l)
      do 15 n=1,nnode
        nn = lnods(l,n)
        encod(1,n) = coord(nn,1)
        encod(2,n) = coord(nn,2)
   15 continue
c...zero stiffness matrix
      do 20 j=1,6
      do 20 i=1,6
        estif(i,j) = 0.0
   20 continue
c
c...find length, and angle of inclination of beam
      dx = encod(1,2) - encod(1,1)
      dy = encod(2,2) - encod(2,1)
      bl = sqrt(dx*dx + dy*dy)
      alpha = atan2(dy,dx)
      c = cos(alpha)
      s = sin(alpha)
      c2 = c*c
      s2 = s*s
      cs = c*s
      bl2 = bl*bl
      bil = bi/bl
      bil2 = bi/bl2
c
c...now assemble stiffness matrix
      estif(1,1) =   ba*c2 + 12.*bil2*s2 
      estif(1,2) =  (ba    - 12.*bil2)*cs
      estif(1,3) =         -  6.*bil*s
      estif(1,4) = -(ba*c2 + 12.*bil2*s2)
      estif(1,5) = -(ba    - 12.*bil2)*cs
      estif(1,6) =         -  6.*bil*s
      estif(2,2) =   ba*s2 + 12.*bil2*c2
      estif(2,3) =            6.*bil*c
      estif(2,4) = -(ba    - 12.*bil2)*cs
      estif(2,5) = -(ba*s2 + 12.*bil2*c2)
      estif(2,6) =            6.*bil*c
      estif(3,3) =            4.*bi
      estif(3,4) =            6.*bil*s
      estif(3,5) =         -  6.*bil*c
      estif(3,6) =            2.*bi
      estif(4,4) =   ba*c2 + 12.*bil2*s2
      estif(4,5) =  (ba    - 12.*bil2)*cs
      estif(4,6) =            6.*bil*s
      estif(5,5) =   ba*s2 + 12.*bil2*c2
      estif(5,6) =         -  6.*bil*c
      estif(6,6) =            4.*bi
c
c...all terms are multiplied by E/l
      bel = be/bl
      do 25 i=1,6
        do 25 j=i,6
          estif(i,j) = bel*estif(i,j)
   25 continue
c
c...now fill in lower triangle of estif
      do 30 i=2,6
        do 30 j=1,i-1
          estif(i,j) = estif(j,i)
   30 continue
c
c...write estif to file
      write (18) estif
      write(6,*)'Stiffness matrix for elt ',l,' written to file.'
   80 continue
c...end of element loop
c
      write(6,*)'Elt stiffness matrices written to file'
      endfile 18
c
      return
      end
c
c
      subroutine load(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1    nelem,npoin,lnods,lnodn,coord,surft,aslod,lpros,props)
c
c...compile global nodal load vector
c
      dimension aslod(mtotv),surft(msets,2),props(msets,mcomp),
     1     lnods(melem,mnode),press(2,3),npals(3),lpros(melem),
     2     coord(mpoin,2),encod(2,3),lnodn(melem)
c
      ntotv = npoin*3
c
      write(6,*)'Applied Loads'
      write(6,*)'-------------'
      write(6,*)
      write(16,*)'Applied Loads'
      write(16,*)'-------------'
      write(16,*)
c...initialize the global load vector
      do 10 i=1,mtotv
        aslod(i) = 0.0
   10 continue
c
c...there are some dummy lines in the input data, which are read but ignored
      read(15,*,end=99)
c...npset>0 if there are surface tractions. nbset is not used.
      read (15,510) nbset,npset
  510 format(2i5)
c
c...nodal loads (ld=1, ignored)
      write(6,*)'Nodal loads:'
      write(6,*)'-----------'
      write(6,*)
      write(6,*)'  Node   x-component   y-component'
      write(16,*)'Nodal loads:'
      write(16,*)'-----------'
      write(16,*)
      write(16,*)'  Node   x-component   y-component'
      read(15,*)
   15 read(15,520) lcard,ld,np,fx,fy
  520 format(i1,i4,i5,2e10.3)
      if (np.ne.0) then
c...the i'th component of node np must be converted to global degree 
c...of freedom component
c...remember we have three d.o.f.s per node: (u,v,theta)
c...corresponding to forces/moments: (fx,fy,M)
        nm = 3*(np - 1)
c...add nodal loads into aslod
        aslod(nm + 1) = aslod(nm + 1) + fx
        aslod(nm + 2) = aslod(nm + 2) + fy
        write(6,610) np,fx,fy
        write(16,610) np,fx,fy
  610   format(i5,2(4x,e10.3))
        if (lcard.eq.0) goto 15
      endif
      write(6,*)
      write(16,*)
c
c
c...surface tractions
      if (npset.gt.0) then
        read(15,*)
        write(6,*)'Surface Tractions'
        write(6,*)'-----------------'
        write(6,*)
        write(6,*)' set   normal cmpnt   tangential cmpnt' 
        write(16,*)'Surface Tractions'
        write(16,*)'-----------------'
        write(16,*)
        write(16,*)' set   normal cmpnt   tangential cmpnt' 
c
c...first read surface traction sets
   25   read(15,530) lcard,iset,i1,i2,pn,pt
  530   format(i1,i4,3x,2i2,8x,2e10.3)
        if (iset.gt.0) then 
          surft(iset,1) = pn
          surft(iset,2) = pt
          write(6,620)  iset,pn,pt
          write(16,620) iset,pn,pt
  620     format(i3,2(5x,e10.3))
        endif
        if (lcard.eq.0) goto 25
c
c...now apply these tractions to edges
        read(15,*)
        write(6,*)
        write(6,*)'   elt  end-nodes   set-numbers'
        write(16,*)
        write(16,*)'   elt   end-nodes   set-numbers'
   30   read(15,540) lcard,nnas,l,n1,iset1,n2,iset2
c...nnas is the number of nodes along the edge (=2 for simple beam elt)
  540   format(i1,i4,i5,2(i5,i2,1x))
        if (l.gt.0) then
          write(6,630) l,n1,n2,iset1,iset2
          write(16,630) l,n1,n2,iset1,iset2
  630     format(i5,2(2x,2i5))
c...put the tractions in press. 
c...m=1 for normal, and m=2 for tangential tractions
          do 35 m=1,2
            press(m,1) = surft(iset1,m)
            press(m,2) = surft(iset2,m)
   35     continue
c...store the node numbers along side in npals
          npals(1) = n1
          npals(2) = n2
c...extract nodal coords and store in encod
          do 60 n=1,2
            np = npals(n)
            do 55 m=1,2
   55         encod(m,n) = coord(np,m)
   60     continue
c
c...find the angle of inclination of the beam
          dx = encod(1,2) - encod(1,1)
          dy = encod(2,2) - encod(2,1)
          bl = sqrt(dx*dx + dy*dy)
          alpha = atan2(dy,dx)
          cosa = cos(alpha)
          sina = sin(alpha)
c...equiv nodal forces in normal and tangential directions at node 1:
c...note: normal forces use the cubic shape fns, while tangential forces
c...distributed loading also creates moments at nodes
c...      use the linear shape fns, hence the different coeffts
          fn = (7.0*press(1,1) + 3.0*press(1,2))/20.
          ft = (2.0*press(2,1) +     press(2,2))/6.
          bm = (3.0*press(1,1) + 2.0*press(1,2))/60. 
c...now resolve into x and y direction forces
          fx = -fn*sina + ft*cosa
          fy =  fn*cosa + ft*sina
          np1 = npals(1)
          nm = 3*(np1 - 1)
c...add nodal loads into aslod
          aslod(nm + 1) = aslod(nm + 1) + fx*bl
          aslod(nm + 2) = aslod(nm + 2) + fy*bl
          aslod(nm + 3) = aslod(nm + 3) + bm*bl*bl
c...now deal with node 2
          fn =  (3.0*press(1,1) + 7.0*press(1,2))/20.
          ft =  (    press(2,1) + 2.0*press(2,2))/6.
          bm = -(2.0*press(1,1) + 3.0*press(1,2))/60. 
          fx = -fn*sina + ft*cosa
          fy =  fn*cosa + ft*sina
          np2 = npals(2)
          nm = 3*(np2 - 1)
          aslod(nm + 1) = aslod(nm + 1) + fx*bl
          aslod(nm + 2) = aslod(nm + 2) + fy*bl
          aslod(nm + 3) = aslod(nm + 3) + bm*bl*bl
        endif
        if (lcard.eq.0) goto 30
c...go back for next loaded edge
      endif
c
c...report the non-zero loads assembled in aslod
      call prompt
      write(6,*) 
      write(6,*)'Non-zero entries in global load vector:'
      write(6,*)'  node  comp.     value'
      write(6,*)
      write(16,*) 
      write(16,*)'Non-zero entries in global load vector:'
      write(16,*)'  node  comp.     value'
      write(16,*)
      do 95 i=1,ntotv
        ai = aslod(i)
c...zero any tiny loads caused by rounding errors
        if (abs(ai).lt.1.e-5) then
          aslod(i) = 0.0
        else
          n = (i-1)/3 + 1
          m = i - 3*(n-1)
          write(6,650) n,m,ai
          write(16,650) n,m,ai
  650     format(i5,i5,5x,e10.3)
        endif
   95 continue
c
      write(6,*)
      write(6,*)'End of load data.'
      write(6,*)
      write(16,*)
      write(16,*)'End of load data.'
      write(16,*)
      return
   99 write(6,*)'ERROR: this input file has no loading data.'
      call prompt
      stop
      end
c      
c
      subroutine result(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1                  nelem,npoin,lnods,lnodn,coord,nkode,
     3                  lpros,props,aslod,asdis)
c
c...output the results to file
c
      dimension aslod(mtotv),props(msets,mcomp),
     1         lnods(melem,mnode),coord(mpoin,2),
     5      nkode(mpoin),asdis(mtotv),lnodn(melem),lpros(melem)
c...output displacements and reactions
c...output to screen (unit 6):
      write(6,*)
      write(6,*)'RESULTS'
      write(6,*)'======='
      write(6,*)
      write(6,610)
  610 format(//' nodal reactions, displacements, rotations'
     1/40('-')//3x,'node',
     212x,'x-reaction',5x,'y-reaction',5x,'  moment  ',9x,
     3'x-displt',7x,'y-displt',9x,'rotation')
c...output to results file (unit 16):
      write(16,*)
      write(16,*)'RESULTS'
      write(16,*)'======='
      write(16,*)
      write(16,610)
c...and output to end of data file, for postprocessor:
      write(17,*)'  node  x-displt  y-displt  rotation'
      do 20 n=1,npoin  
        i = 3*(n - 1)
        dispx = asdis(i+1)
        dispy = asdis(i+2)
        rotat = asdis(i+3)
        reacx = aslod(i+1)
        reacy = aslod(i+2)
        bmomt = aslod(i+3)
        write(17,612) n,dispx,dispy,rotat
  612   format(i5,3f10.5)
        write(6,615) n,reacx,reacy,bmomt,dispx,dispy,rotat 
        write(16,615) n,reacx,reacy,bmomt,dispx,dispy,rotat
  615   format(3x,i3,6x,3(5x,e10.3),5x,3(5x,e10.3))    
        if (mod(n,25).eq.0) call prompt
   20 continue
c
      write(6,*)
      write(6,*)'END OF RESULTS'
      write(6,*)
      write(16,*)
      write(16,*)'END OF RESULTS'
      write(16,*)
      return
      end
c      
c
      subroutine assdk(melem,mpoin,mnode,mtotv,
     1    nelem,npoin,lnods,lnodn,ndof,nvar,nkode,diagk)
c
c...assembles the diagonal of the stiffness matrices, 
c...read from the scratch file, into the global vector diagk
c
      dimension lnods(melem,mnode),estif(6,6),kd(3),
     1         ndof(mpoin),nvar(mpoin),
     1         lnodn(melem),nkode(mpoin),diagk(mtotv)
c
c...initialize diagk
      do 20 i=1,mtotv
        diagk(i) = 0.0
   20 continue
c
      rewind 18
c
c...loop through all elements
      do 50 l=1,nelem
         read (18) estif
         nnode = lnodn(l)
c...deal with elt nodes ni in turn
c... these will be global nodes nni
        ni=0
        ndi = 0
   25   ni = ni + 1
        nni = lnods(l,ni)
        ndofnn = ndof(nni)
        nvarnn = nvar(nni)
c...put the entries from estif into diag(gstif)
        do 40 mi=1,ndofnn
            im = ndi + mi
            iim = nvarnn + mi
            diagk(iim) = diagk(iim) + estif(im,im)
   40   continue
        ndi = ndi + ndofnn
        if (ni.lt.nnode) goto 25
   50 continue
c...end of element loop
c
c...add a large number to the diagonal coefft of fixed d.o.f.s
      do 70 nn=1,npoin
        kode = nkode(nn)
        ndofnn = ndof(nn)
        nvarnn = nvar(nn)
        if (ndofnn.eq.2) then       
          kd(1) = kode/10
          kd(2) = kode - kd(1)*10
        else
          kd(1) = kode/100
          kod = kode - kd(1)*100
          kd(2) = kod/10
          kd(3) = kod - kd(2)*10
        endif
        do 60 m=1,ndofnn
          if (kd(m).eq.0) goto 60
          nposn = nvarnn + m
          diagk(nposn) = diagk(nposn) + 10.0e20
   60   continue
   70 continue
c
      return
      end
c
c
c
      subroutine pcgsol(mtotv,melem,mpoin,mnode,ntotv,nelem,
     1           npoin,lnods,lnodn,diagk,nkode,ndof,nvar,
     2        g , v , z , r , p , y ,cgtol,itscg,rnorm0)
c
      dimension nkode(mpoin),diagk(mtotv),lnods(melem,mnode),
     1   lnodn(melem),kd(3),ndof(mpoin),nvar(mpoin),
     2   v(mtotv),g(mtotv),z(mtotv),r(mtotv),p(mtotv),y(mtotv)
c
c...solve Kv=g using diagk as preconditioner
c...multiplications K.v are performed element-by-element, 
c...reading the element matrices of K from the scratch file
c
c...initial guess; 
      do 2 i=1,ntotv
       v(i) = 0.0
       r(i) = g(i)
       z(i) = 0.0
    2 continue
c
c...zero the fixed d.o.f.s in r
      do 6 np=1,npoin
        ndofnp = ndof(np)
        nvarnp = nvar(np)
        kode = nkode(np)
        if (ndofnp.eq.2) then       
          kd(1) = kode/10
          kd(2) = kode - kd(1)*10
        else
          kd(1) = kode/100
          kod = kode - kd(1)*100
          kd(2) = kod/10
          kd(3) = kod - kd(2)*10
        endif
        do 4 m=1,ndofnp
          if (kd(m).eq.0) goto 4
          nposn = nvarnp + m
          r(nposn) = 0.0
    4   continue
    6 continue
c
c...z = {Kdiag}^-1 * r
      do 15 i=1,ntotv
        z(i) = r(i)/diagk(i)
   15 continue
c
      do 16 i=1,ntotv
        p(i) = z(i)
   16 continue
      ztr = 0.0
      do 17 i=1,ntotv
        zi = z(i)
        ri = r(i)
        ztr = ztr + zi*ri
   17 continue
      call norm2(mtotv,ntotv,r,rnorm0)
      itscg = 0
c
      write(6,*)'PGSOL: iteration',itscg,',norm = ',rnorm0
c
   25 itscg = itscg + 1

c...create y which is K*p
      call ematgv(18,mtotv,melem,mpoin,mnode,ntotv,nelem,
     1         npoin,lnods,lnodn,ndof,nvar, p , y )
c
c...zero the fixed d.o.f.s in y
      do 26 np=1,npoin
        ndofnp = ndof(np)
        nvarnp = nvar(np)
        kode = nkode(np)
        if (ndofnp.eq.2) then       
          kd(1) = kode/10
          kd(2) = kode - kd(1)*10
        else
          kd(1) = kode/100
          kod = kode - kd(1)*100
          kd(2) = kod/10
          kd(3) = kod - kd(2)*10
        endif
        do 24 m=1,ndofnp
          if (kd(m).eq.0) goto 24
          nposn = nvarnp + m
          y(nposn) = 0.0
   24   continue
   26 continue
c
      pty = 0.0
      do 28 i=1,ntotv
        pty = pty + p(i)*y(i)
   28 continue
      alpha = ztr/pty      
      do 30 i=1,ntotv
       v(i) = v(i) + alpha*p(i)
   30 continue
c
      do 40 i=1,ntotv
        r(i) = r(i) - alpha*y(i)
   40 continue
c
      call norm2(mtotv,ntotv,r,rnorm)
c
      write(6,*)'       iteration',itscg,',norm = ',rnorm
c
      if(rnorm.lt.cgtol) goto 200
c
c...z = {Kdiag}^-1 * r
      do 45 i=1,ntotv
        z(i) = r(i)/diagk(i)
   45 continue
c
      ztr0 = ztr
      ztr = 0.0
      do 60 i=1,ntotv
        ztr = ztr + z(i)*r(i)
   60 continue
      beta = ztr/ztr0
c
      do 130 i=1,ntotv
       p(i) = z(i) + beta*p(i)
  130 continue
c
      goto 25
c
  200 continue
      write(6,*)'Solution converged after ',itscg,' iterations.'
      write(6,*)
      return
      end
c
c
      subroutine norm2(mdim,ndim,v,vnorm)
      dimension v(mdim)
c
c...calculates the Euclidean norm of the vector v, returns it in vnorm
      sum = 0.0
      do 5 i=1,ndim
        vi = v(i)
        sum = sum + vi*vi
    5 continue
      vnorm = sqrt(sum)
      return
      end
c
c
      subroutine ematgv(ichan,mtotv,melem,mpoin,mnode,ntotv,nelem,
     1         npoin,lnods,lnodn,ndof,nvar,asvin,asvout)
      dimension asvin(mtotv),asvout(mtotv),estif(6,6),lgdof(6)
     1     ,lnods(melem,mnode),lnodn(melem)
     2     ,ndof(mpoin),nvar(mpoin)
c
c...multiply element matrices on file, 
c...by global vector asvin, to get global vector asvout
      rewind ichan
c
      do 10 i=1,ntotv
        asvout(i) = 0.0
   10 continue
c
c...loop through all elements
      do 50 l=1,nelem
        nnode = lnodn(l)
c...set up an array to relate the global node no.s to the local ones
        idof = 0
        do 15 n=1,nnode
          np = lnods(l,n)
          nvarnp = nvar(np)
          ndofnp = ndof(np)
          do 12 i=1,ndofnp
            idof = idof + 1
            iidof = nvarnp + i
            lgdof(idof) = iidof
   12     continue
   15   continue
        nevab = idof
c...the i'th d.o.f. of elt matrix corresponds to global d.o.f. lgdof(i)
c...note that this will not work in CONSL.FOR, where an elt matrix may not
c...involve all d.o.f.s of a node (interface of structural and soil elts)
c 
        read (ichan) estif
c
        do 27 i=1,nevab
          sum = 0.0
          do 26 j=1,nevab
            jj = lgdof(j)
            sum = sum + estif(i,j)*asvin(jj)
   26     continue
          ii = lgdof(i)
          asvout(ii) = asvout(ii) + sum
   27   continue
c
   50 continue
c...end of element loop
c
      return
      end
c
c