c...thermoelasticity "main engine"
c...for use with FELIPE version 3 - sample file is thermqd2.dat
c
c
c...If you want to re-compile this program, you must link it with
c...the file felsb.obj, compiled from felsb.for (provided in the
c...FELIPE folder). The file felsb.for contains additional subroutines 
c...used by all the 'main engines' except eladv.for. 
c...If you don't use linking, you should paste the contents of 
c...the file felsb.for to the end of this file, before compiling
c...and running it.
c
      program therm
cp...program for steady-state solution of 2D thermoelasticity problems.
cp...uses 4-noded or 8-noded quad elts where corner nodes have (u,v,p)
cp...degrees of freedom (p=temperature). Example of a 
cp...coupled problem, and of additional d.o.f.s per node. Also uses
cp...a Gaussian elimination solver for the nonsymmetric stiffness
cp...matrix (stored as a band). Theory: solve
cp
cp   /         T  \ /   \     /   \
cp   |  K     Q   | | u |     | f |
cp   |            | |   |  =  |   |
cp   |  0    -H   | | p |     | r |
cp   \            / \   /     \   /
cp...see manual for definition of terms. p is the temperature.
cp...Changes to elast.for are prefaced by cp... comment lines 
c
      parameter(melem=400,mpoin=800,mnode=8,
     1msets=9,mcomp=6,mtotv=2400,mbw=500)
c...these parameters fix the dimensions of the arrays:
c...melem = max. no. of elements
c...mpoin = max. no. of nodes
c...mnode = max. no. of nodes in a single element
c...msets = max. no. of mat. prop. sets (also used for loading sets)
c...mcomp = max. no. of mat. props (e.g. Young's mod, Poisson ratio)
c...mtotv = max no. of degrees of freedom (corner nodes have 3 dofs)
c...mbw  = max. bandwidth of global stiffness matrix
cp...We have to use bandwidth and store whole band, because K is not 
cp...symmetric!
c
c
      character*1 char,dbdcmp(8)
      character*80 title,infile,outfil,prtfil
c
      dimension lnods(melem,mnode),lnodn(melem),lsidn(melem),
     1  coord(mpoin,2),nkode(mpoin),aslod(mtotv),
     2        lpros(melem),props(msets,mcomp),asdis(mtotv),
     4      surft(msets,2),gstif(mtotv,mbw),lmtyp(melem)
     5  ,ndof(mpoin),nvar(mpoin),dbdcrd(8),dbdval(8)
c...lnods(l,n) = global no. of n'th node of element l
c...lnodn(l) = no. of nodes in element l ( = 8 for 8-noded quad.)
c...lsidn = no. of sides in element l    ( = 4 for 8-noded quad.)
c...coord(np,1), coord(np,2) = x,y coords of node np
c...nkode(np) = fixity code of node np
c...aslod is global nodal load vector
c...ndof(np) = no. of dof.s at node np 
cp..nvar(np) is a pointer array for global dof of node np
c...lpros(l) = mat. prop. set no. of element l
c...props(m,i) = value of i'th mat. property in mat. prop. set m
c...asdis is global nodal displacement vector
c...
c...surft(m,1),surft(m,2) = normal,tangential surface tractions in set m
c...gstif holds assembled global stiffness matrix in compact form
c...lmtyp(l) = material type 
c...dbdcmp,dbdval,dbdcrd are component, coord, value for Dirichlet
c...    boundary value sets. 
c
      write(6,*)'Program THERM: 2D thermoelasticity problems.'
      write(6,*)'Handles 4-noded or 8-noded quadrilateral elements,'
      write(6,*)'with (u,v,t) d.o.f.s at all nodes.'
      write(6,*)'Dimensioned for',melem,' elements, and',mpoin,' nodes.'
      write(6,*) 
c
c...read an existing .dat file
c
    1 continue
c
c...get the input file name in infile
      call askfil(infile,iflag)
      if (iflag.eq.1) then
        call yesno(char,'Do you want to exit?')
        if (char.eq.'y') stop
        goto 1
      endif
c
      write(6,*)'Opening file ',infile,'...'
      lenin = lenstr(infile)
      open(15,file=infile(1:lenin),status='old',err=999)
      write(6,*)
      lm4 = lenin - 4
      prtfil = infile(1:lm4)//'.prt'
      write(6,*)'Creating file ',prtfil(1:lenin),' for results...'
      write(6,*)
      open(16,file=prtfil(1:lenin),status='unknown',err=999)      
      outfil = infile(1:lm4)//'.out'
      write(6,*)'Creating file ',outfil(1:lenin),' for output...'
      write(6,*)
      open(17,file=outfil(1:lenin),status='unknown',err=999)      
c
c...read the input mesh data, and echo it to the output files
      call input(15,melem,mpoin,mnode,msets,mcomp,nelem,
     1 npoin,lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props
     2         ,ndofn,npros,title,ndof)
      call print(16,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1 lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2           npros,title,ndof)
      call output(17,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1  lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2            npros,title,ndof)
      call yesno(char,'Do you want the mesh details listed?')
      if (char.eq.'y') then
        call print(6,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1   lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2           npros,title,ndof)
      endif
c
      iflag = 0
      do 2 l=1,nelem
        nnode = lnodn(l)
        nside = lsidn(l)
        nodsid = 10*nnode + nside
        if (nodsid.ne.84.and.nodsid.ne.44) then
          iflag = 1
          write(6,*)'ERROR: Cannot handle element ',l
          write(6,*)'which has ',nside,' sides and ',nnode,' nodes'
        endif
    2 continue
      if (iflag.eq.1) then
        call prompt
        stop
      endif
c
cp...create the pointer array nvar. Needed when there are variable
cp...dof's per node. Not strictly needed in THERM, but will be needed
cp...in POREL. The no of dof's in node n is stored in ndof(n)
      nvar(1) = 0
      nvtot = 0
      do 5 np=1,npoin
        ndofnp = ndof(np)
        nvar(np) = nvtot
        nvtot = nvtot + ndofnp
    5 continue
cp...Now local dof i of node np is global dof nvar(np)+i
cp...and nvtot is total no. of degrees of freedom
c
      write(6,*)'Is this a plane stress or plane strain analysis?'
    3 write(6,*)'Type 1 for plane stress, or 2 for plane strain...'
      read(5,'(i1)',err=3) mstyp
      write(16,*)
      if (mstyp.eq.1) then
        write(6,*)'Plane stress analysis'
        write(16,*)'Plane stress analysis'
      else if (mstyp.eq.2) then
        write(6,*)'Plane strain analysis'
        write(16,*)'Plane strain analysis'
      else
        write(6,*)'Invalid input!'
        goto 3
      endif        
      write(16,*)
c
c...open a scratch file to write the element stiffness matrices on
      open(unit=18,status='scratch',form='unformatted')  
      write(6,*)'Scratch file opened ...'
      write(6,*)
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine STIFF...'
      call stiff(melem,mpoin,mnode,msets,mcomp,mstyp,
     1     nelem,npoin,lnods,lnodn,lsidn,coord,lpros,props
     2           ,lmtyp,ndof,nvar)
cp...previous line added for this and following routines
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine LOAD...'
      call load(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1          nelem,npoin,lnods,lnodn,coord,
     2          surft,aslod,lmtyp,ndof,nvar,lpros,props)
c
c
c...read data for Dirichlet boundaries, if present
      read(15,*)
      read(15,'(2i5)')nbrad,nbdir
c
      if (nbrad.gt.0) then
        write(6,*) 'ERROR: nbrad > 0.'
        write(6,*)'This program cannot handle radiating boundaries.'
        call prompt
        stop
      endif
c
      if (nbdir.gt.0) then
c...read Dirichlet boundary values, if present, from end of .dat file
        call bndry(ndbdry,dbdcmp,dbdcrd,dbdval)
      endif
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine ASSMB...'
      call assmb(melem,mpoin,mnode,mtotv,mbw,
     1           nelem,npoin,nbw,lnods,lnodn,
     2           gstif,lmtyp,ndof,nvar)
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine SOLVE...'
      call solve(mpoin,mtotv,mbw,npoin,nbw,nkode,
     1           coord,aslod,asdis,gstif,ndof,nvar,
     2           ndbdry,dbdcmp,dbdcrd,dbdval)
c
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine RESULT...'
      call result(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1   nelem,npoin,lnods,lnodn,lsidn,coord,nkode,lpros,props,
     2            aslod,asdis,lmtyp,ndof,nvar)
c
      write(6,*)
      write(6,*)'THERM completed. Results are in file ',prtfil(1:lenin)
      write(6,*)
      write(6,*)
      call prompt
      stop
  999 write(6,*)
     1'Cannot open file with this name. Terminating program.'
      call prompt
      stop
      end
c
c
c
      subroutine bndry(ndbdry,dbdcmp,dbdcrd,dbdval)
c
c...reads the values of U for planes on the Dirichlet boundary.
c
      character*1 char,dbdcmp(8)
      dimension dbdcrd(8),dbdval(8)
c
      i = 0
      read(15,'(a80)',end=5,err=5)
      write(6,*)'Dirichlet boundary values:'
      write(6,*)'--------------------------'
      write(6,*)
    3 read(15,503)lcard,char,plane,value,indof
  503 format(i1,1x,a1,2x,g10.3,5x,g10.3,i5)
c...we assume that indof = 3 (=temperature dof) for this program
      indof = 3
      if (char.eq.' ') goto 5
      write(6,603) 'T =',value,' on Dirichlet bdry where ',
     1  char,' =',plane
  603 format(a3,g10.3,a25,a1,a2,e10.4)
      i = i + 1
      dbdcmp(i) = char
      dbdcrd(i) = plane
      dbdval(i) = value
      if (i.lt.8.and.lcard.eq.0) goto 3
    5 ndbdry = i
      if (ndbdry.eq.0) then
        write(6,*)'No Dirichlet boundary values - the inbuilt'
        write(6,*)'boundary value function bdryfn(x,y) will be used.' 
      else
        write(6,*)'The inbuilt boundary value function will be used'
        write(6,*)'for any Dirichlet nodes not defined above.'
      endif
      write(6,*)
      write(6,*)
c 
      return
      end
c
c
c
      function bdryfn(x,y)
c...Dirichlet boundary function (used if no values specified) 
c...The example here is g(x,y) = x*x - y*y.  Fill in your own choice!!
        bdryfn = x*x - y*y
      return
      end
c
c
c      
      subroutine stiff(melem,mpoin,mnode,msets,mcomp,mstyp,
     1    nelem,npoin,lnods,lnodn,lsidn,coord,lpros,props
     2             ,lmtyp,ndof,nvar)
c      
c...calculates the element stiffness matrices, and writes them
c...to the scratch file.
c
      dimension props(msets,mcomp),lnods(melem,mnode),encod(2,8),
     1         estif(16,16),bmatx(3,16),dmatx(3,3),dbmat(3,16),
     2        gpwts(4),lpros(melem),lnodn(melem),coord(mpoin,2),
     3       shape(8),gploc(2,4),gpcod(2),deriv(2,8),cartd(2,8),
     4      lmtyp(melem),ndof(mpoin),nvar(mpoin),qmatx(8,16),
     5     hmatx(8,8),xmatx(24,24),lsidn(melem)
c
cp...for thermoelastic elements, stiffness matrix is 
cp
cp   /         T  \  Form K in estif(16,16)    
cp   |  K    -Q   |  Form Q in qmatx(8,16)
cp   |            |  Form H in hmatx(8,8)
cp   |  0     H   |  Assemble whole matrix in xmatx(24,24)
cp   \            /  Write xmatx to file 
cp
cp
c
c...encod holds the nodal coords for the nodes around the current element
c...estif will be the stiffness matrix for the current element
c...bmatx, dmatx are the B and D matrices, and dbmat is D*B
c...gpwts,gploc are the Gauss point weights and locations (in xi,eta coords)
c...gpcod will be the x,y coords of the current Gauss point
c...shape holds the shape functions evaluated at the Gauss point
c...deriv holds the derivatives (w.r.t. xi, eta) of the shape functions
c...cartd holds the Cartesian derivatives (w.r.t x,y) of shape functions
c
      rewind 18
c
c...start loop over all elements
      do 80 l=1,nelem
c
c...extract Young's modulus & Poisson's ratio from props      
        lset = lpros(l)
        ym = props(lset,1)
        pr = props(lset,2)
        thick = props(lset,3)
        if (thick.eq.0.0) thick = 1.0
cp...get elt material type:
        lmtype = lmtyp(l)
        conduc = props(lset,4)
        expand = props(lset,5)
cp...conduc = conductivity coefft, 
cp...expand = thermal expansion coefft
c
c...put nodal coords in encod
        nnode = lnodn(l)
        nside = lsidn(l)
        do 15 n=1,nnode
          nn = lnods(l,n)
          encod(1,n) = coord(nn,1)
          encod(2,n) = coord(nn,2)
   15   continue
c...zero stiffness matrices
        do 16 j=1,8
          do 16 i=1,8
            hmatx(i,j) = 0.0
   16   continue
        do 20 j=1,16
          do 17 i=1,8
            qmatx(i,j) = 0.0
   17     continue
          do 20 i=1,16
            estif(i,j) = 0.0
   20   continue
        do 21 i=1,24
          do 21 j=1,24
            xmatx(i,j) = 0.0
   21   continue  
c
cp...set up dimensions, to help generalisation to other elt types
        nevab = 2*nnode
        nsvab = nevab + nnode
cp...for 8-noded elements,
cp...nevab = 16, the dimension of the elastic stiffness matrix estif
cp...nsvab = 24, the dimension of xmatx (including temperature dof's)
c
cp...first form K in estif
c
c...get D matrix 
        call dmat(mstyp,ym,pr,dmatx)
c
c...put local coords of Gauss-pts in gploc
cp...ngaus = 1 for linear, 2 for quadratic elements
        call gauss(nnode,gploc,gpwts,ngaus)
c
c...start loop through gauss-pts for contribution to estif
        do 75 ij=1,ngaus
c...put local coords of gauss-pt in s,t
          s = gploc(1,ij)
          t = gploc(2,ij)
c...get shape fns (shape) & derivs (deriv) in local coords
          call sfr(nnode,s,t,shape,deriv)
c...get xy-coords of gauss-pt into gpcod, using shape fn interpolation
          do 25 m=1,2
            sum = 0.0
            do 24 n=1,nnode
   24       sum = sum + encod(m,n)*shape(n)
            gpcod(m) = sum
   25     continue
c...get xy-derivs (cartd) & determinant of Jacobian (djacb)
          call jacob(nnode,encod,shape,deriv,cartd,djacb)
c...get B matrix 
          call bmat(nnode,cartd,bmatx)
c...form matrix product D*B
          do 40 j=1,nevab
            do 40 i=1,3
              sum=0.0
              do 39 k=1,3
   39         sum = sum + dmatx(i,k)*bmatx(k,j)
              dbmat(i,j) = sum
   40     continue
c...B transpose *d*b* djacb = contribution to estif
        dvolu = djacb*gpwts(ij)*thick
        do 50 j=1,nevab
        do 50 i=1,nevab
          sum = 0.0
          do 49 k=1,3
   49     sum = sum + bmatx(k,i)*dbmat(k,j)
          estif(i,j) = estif(i,j) + sum*dvolu
   50   continue
c
cp...now form qmatx and hmatx, for thermoelastic elements
cp
cp                       T  T  
cp...form Q matrix as   N  M  B 
cp                  T    T
cp (actually, form Q  = B  M N, dimensions 16*3 3*2 2*8 ) 
cp...where M selects only direct stresses (so k=1,2 below) 
cp
cp      | bx  0  |
cp  M = | 0   by |
cp      | 0   0  |
cp
          do 55 j=1,nnode
            do 55 i=1,nevab
              sum=0.0
cp...we don't need to use k=3 below, since bottom row of M is zero
              do 54 k=1,2
   54           sum = sum + bmatx(k,i)*expand*shape(j)
              qmatx(j,i) = qmatx(j,i) + sum*dvolu
   55     continue
cp                                       T
cp...now form hmatx as  conductivity * N'  N'
          do 60 j=1,nnode
          do 60 i=1,nnode
            sum = 0.0
            do 59 k=1,2
   59       sum = sum + conduc*cartd(k,i)*cartd(k,j)
            hmatx(i,j) = hmatx(i,j) + sum*dvolu
   60     continue
cp
   75 continue
c...end of gauss loop
c
cp...for thermoelastic matrices, assemble xmatx from K, L and H.
cp...arrange so that temperature dof of corner nodes
cp...comes after displacement dofs for that node
        jcol = 0
        np   = 0
        do 155 n=1,nnode
          jcol1 = jcol + 1
          jend = jcol + 2
cp...top left submatrix
          do 140 j=jcol1,jend
            jmnp = j - np
            do 135 i=1,nevab
              xmatx(i,j) = estif(i,jmnp) 
  135       continue
c...zero submatrix in bottom left
  140     continue
          np = np + 1
          jcol = jend + 1
cp...top right submatrix
          do 145 i=1,nevab
            xmatx(i,jcol) = qmatx(np,i)
  145     continue
cp...bottom right submatrix
          do 150 i=nevab+1,nsvab
            imnvb = i - nevab
            xmatx(i,jcol) = hmatx(imnvb,np)
  150     continue
  155   continue
cp...now swap rows around correspondingly
        np = 0
        irow = 0
        do 170 n=1,nnode
          irow = irow + 2
          np = np + 1
          do 165 j=1,nsvab
            nvbnp = nevab + np
cp..store the temperature row for this node in temp
            tempx = xmatx(nvbnp,j)
cp..move all subsequent displt rows down, to make room
            i = nevab + np - 1
  160       continue
            xmatx(i+1,j) = xmatx(i,j)
            i = i - 1
            if (i.ge.irow+1) goto 160
cp...insert the temp row in the space ater the displt rows for this node
            xmatx(irow+1,j) = tempx
  165     continue
          irow = irow + 1
  170   continue
cp...end of formation of xmatx
c
cp...note that if we had a quadratic element, with
cp... temperature dof's only at corner nodes, we would
cp...have to use linear shape fn's for these. The more general coding for this
cp...situation can be found in porel.for
c
c...write stiffness matrix to file
        write(18) xmatx
c
      write(6,*)'Stiffness matrix for elt ',l,' written to file.'
c
   80 continue
c...end of element loop
c
      write(6,*)'Elt stiffness matrices written to file'
      endfile 18
c
      return
      end
c
c
      subroutine load(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1                nelem,npoin,lnods,lnodn,coord,surft,
     2                aslod,lmtyp,ndof,nvar,lpros,props)
c
c...compile global nodal load vector
c
      dimension aslod(mtotv),surft(msets,2),
     1     lnods(melem,mnode),press(2,3),npals(3),gpwts(2),
     2     coord(mpoin,2),gploc(2),shape(3),deriv(3),xjacm(2),      
     3   gpres(2),fcomp(2),encod(2,3),lnodn(melem),lpros(melem)
     4  ,lmtyp(melem),ndof(mpoin),nvar(mpoin),props(msets,mcomp)
c
      ntotv = nvar(npoin) + ndof(npoin)
c
      write(6,*)'Applied Loads'
      write(6,*)'-------------'
      write(6,*)
      write(16,*)'Applied Loads'
      write(16,*)'-------------'
      write(16,*)
c...initialize the global load vector
      do 10 i=1,mtotv
        aslod(i) = 0.0
   10 continue
c
c...there are some dummy lines in the input data, which are read but ignored
      read(15,*,end=99)
c...npset>0 if there are surface tractions. nbset is not used.
      read (15,510) nbset,npset
  510 format(2i5)
c
c...nodal loads (ld=1, ignored)
      write(6,*)'Nodal loads:'
      write(6,*)'-----------'
      write(6,*)
      write(6,*)'  Node   x-component   y-component'
      write(16,*)'Nodal loads:'
      write(16,*)'-----------'
      write(16,*)
      write(16,*)'  Node   x-component   y-component'
      read(15,*)
   15 read(15,520) lcard,ld,np,fx,fy
  520 format(i1,i4,i5,2e10.3)
      if (np.ne.0) then
c...the i'th component of node np must be converted to global degree 
c...of freedom component
cp...use nvar for this!
        nm = nvar(np)
c...add nodal loads into aslod
        aslod(nm + 1) = aslod(nm + 1) + fx
        aslod(nm + 2) = aslod(nm + 2) + fy
        write(6,610) np,fx,fy
        write(16,610) np,fx,fy
  610   format(i5,2(4x,e10.3))
        if (lcard.eq.0) goto 15
      endif
      write(6,*)
      write(16,*)
c
c
c...surface tractions
      if (npset.gt.0) then
        read(15,*)
        write(6,*)'Surface Tractions'
        write(6,*)'-----------------'
        write(6,*)
        write(6,*)' set   normal cmpnt   tangential cmpnt' 
        write(16,*)'Surface Tractions'
        write(16,*)'-----------------'
        write(16,*)
        write(16,*)' set   normal cmpnt   tangential cmpnt' 
c
c...first read surface traction sets
   25   read(15,530) lcard,iset,i1,i2,pn,pt
  530   format(i1,i4,3x,2i2,8x,2e10.3)
        if (iset.gt.0) then 
          surft(iset,1) = pn
          surft(iset,2) = pt
          write(6,620)  iset,pn,pt
          write(16,620) iset,pn,pt
  620     format(i3,2(5x,e10.3))
        endif
        if (lcard.eq.0) goto 25
c
c...now apply these tractions to edges
        read(15,*)
        write(6,*)
        write(6,*)'   elt  end-nodes   set-numbers'
        write(16,*)
        write(16,*)'   elt   end-nodes   set-numbers'
   30   read(15,540) lcard,nnas,l,n1,iset1,n2,iset2
c...nnas is the number of nodes along the edge (=3 for 8-noded quad.)
  540   format(i1,i4,i5,2(i5,i2,1x))
        if (l.gt.0) then
c...put the tractions in press. Interpolate linearly for the
c...tractions at the midside node, in quadratic edge case.
          if (nnas.eq.2) then
cp...linear edge
            do 33 i=1,2
              press(i,1) = surft(iset1,i)
              press(i,2) = surft(iset2,i)
   33       continue
            npals(1) = n1
            npals(2) = n2
          else 
cp...quadratic edge
            do 35 i=1,2
              press(i,1) = surft(iset1,i)
              press(i,3) = surft(iset2,i)
              press(i,2) = (press(i,1) + press(i,3))/2.
   35       continue
c...store the node numbers along side in npals
            npals(1) = n1
            npals(3) = n2
c...we have to find the midside node no. from lnods
            nnode = lnodn(l)
            do 42 n=1,nnode
              if (lnods(l,n).eq.npals(1)) goto 45
   42       continue
c...if loop completed, npals(1) is not in element 1
            write(6,700) npals(1),l
  700       format('ERROR: node',i4,' is not in element ',i4,'.')
            call prompt
            stop
   45       k = n + 1
            npals(2) = lnods(l,k)
          endif
          write(6,630) l,n1,n2,iset1,iset2
          write(16,630) l,n1,n2,iset1,iset2
  630     format(i5,2(2x,2i5))
c
c...extract nodal coords and store in encod
          do 60 n=1,nnas
            np = npals(n)
            do 55 m=1,2
   55         encod(m,n) = coord(np,m)
   60     continue
c
c...extract material thickness, for integration of load over 
c...2D element edge (which is actually a face in 3D)
          lset = lpros(l)
          thick = props(lset,3)
          if (thick.eq.0.0) thick = 1.0
c
c...enter 1D Gauss loop along side (2-pt integration)
          if (nnas.eq.2) then
            ngaus = 1
            gploc(1) = 0.0
            gpwts(1) = 2.0
          else
            ngaus = 2
            r13 = 1.0/sqrt(3.0)
            gploc(1) = -r13
            gploc(2) = r13
            gpwts(1) = 1.0
            gpwts(2) = 1.0
          endif
          do 90 ij=1,ngaus
            s = gploc(ij)
            if (nnas.eq.2) then
              shape(1) = 0.5*(1.-s)
              shape(2) = 0.5*(1.+s)
              deriv(1) = -0.5
              deriv(2) = 0.5
            else
              ss = s*s
              shape(1) = (ss-s)*0.5
              shape(2) = 1.0 - ss
              shape(3) = (ss + s)*0.5
              deriv(1) = s - 0.5
              deriv(2) = -2.0*s
              deriv(3) = s + 0.5
            endif
c...get jacobian xjacm
            do 75 j=1,2 
              sum = 0.0
              do 72 n=1,nnas
   72           sum = sum + deriv(n)*encod(j,n)
              xjacm(j) = sum
   75       continue
c...find gauss pt pressures, & convert to components in local coords
            do 80 m=1,2
              sum = 0.0
              do 77 n=1,nnas
   77           sum = sum + press(m,n)*shape(n)
              gpres(m) = sum
   80       continue
            darea = gpwts(ij)*thick
            fcomp(1) = (xjacm(1)*gpres(2) - xjacm(2)*gpres(1))*darea
            fcomp(2) = (xjacm(1)*gpres(1) + xjacm(2)*gpres(2))*darea
c...accumulate equivalent nodal forces in aslod
            do 85 n=1,nnas
              np = npals(n)
cp...use nvar to identify global dof's
              nvnp = nvar(np)
              do 82 m=1,2
                nm = nvnp + m
                aslod(nm) = aslod(nm) + fcomp(m)*shape(n)
   82         continue
   85       continue
   90     continue
c...end of Gauss loop
c
        endif
        if (lcard.eq.0) goto 30
c...go back for next loaded edge
c
c...report the non-zero loads assembled in aslod
        call prompt
        write(6,*) 
        write(6,*)'Non-zero entries in global load vector:'
        write(6,*)' d.o.f.     value'
        write(6,*)
        write(16,*) 
        write(16,*)'Non-zero entries in global load vector:'
        write(16,*)' d.o.f.     value'
        write(16,*)
        do 95 i=1,ntotv
          ai = aslod(i)
c...zero any tiny loads caused by rounding errors
          if (abs(ai).lt.1.e-5) then
            aslod(i) = 0.0
          else
            write(6,650) i,ai
            write(16,650) i,ai
  650       format(i5,5x,e10.3)
          endif
  95    continue
      endif
c
      write(6,*)
      write(6,*)'End of load data.'
      write(6,*)
      write(16,*)
      write(16,*)'End of load data.'
      write(16,*)
      return
   99 write(6,*)'ERROR: this input file has no loading data.'
      call prompt
      stop
      end
c      
c
      subroutine assmb(melem,mpoin,mnode,mtotv,mbw,
     1                 nelem,npoin,nbw,lnods,lnodn,
     2                 gstif,lmtyp,ndof,nvar)
c
c...assembles the elt stiffness matrices, read from the scratch file,
c...into the global matrix gstif, stored compactly using its banded
c...structure. 
cp...As K is nonsymmetric, we have to store the entire band.
cp...entry ( i , j ) maps to gstif( i , j-i+nhbw+1 )
cp...note the bandwidth nbw = 2*nhbw+1 (nhbw=half-bandwidth)
c
      dimension lnods(melem,mnode),gstif(mtotv,mbw),
     1         lnodn(melem),lmtyp(melem),ndof(mpoin),nvar(mpoin)
     2        ,xmatx(24,24)
c
      ntotv = nvar(npoin) + ndof(npoin)
cp...ntotv is total no. of dof's
c...first find the halfbandwidth; mdiff is the max. difference in  
c...global dof's within the same element
      mdiff = 0
      do 15 l=1,nelem
        maxdof = 0
        mindof = ntotv
        nnode = lnodn(l)
        do 10 n=1,nnode
          ln = lnods(l,n)
          lnvar = nvar(ln)
          lndof = ndof(ln)
          lnmin = lnvar + 1
          lnmax = lnvar + lndof
          if (lnmax.gt.maxdof) maxdof = lnmax
          if (lnmin.lt.mindof) mindof = lnmin
   10   continue
        ldiff = maxdof - mindof
        if (ldiff.gt.mdiff) mdiff = ldiff
   15 continue
cp...then the halfbandwidth is ...
      nhbw = mdiff
cp...so the bandwidth is...
      nbw = 2*nhbw + 1
      write(6,*)' Bandwidth = ',nbw
      if (nbw.gt.mbw) then
        write(6,*)'ERROR: bandwidth is ',nbw
        write(6,*)
     1'This exceeds parameter mbw. Increase mbw and recompile.'
        call prompt
        stop
      endif
c
c...initialize gstif
      do 20 i=1,mtotv
      do 20 j=1,mbw
      gstif(i,j) = 0.0
   20 continue
c
      rewind 18
c
c...loop through all elements
      do 50 l=1,nelem
cp...read xmatx
          read(18) xmatx
c
        nnode = lnodn(l)
c...we will allow for variable dof's per node, and use ndof and nvar
c
c...deal with elt nodes ni in rows, nj in columns
c... these will be global nodes nni, nnj
        ni=0
        i = 0
   25   ni = ni + 1
        nni = lnods(l,ni)
        idof = ndof(nni)
        ii = nvar(nni)
        nj = 0
        j = 0
   30   nj = nj + 1
        nnj = lnods(l,nj)
        jdof = ndof(nnj)
        jj = nvar(nnj)
c...put the node-block entries from xmatx into gstif
        do 40 mi=1,idof
          do 35 mj=1,jdof
            im = i + mi
            jm = j + mj
            iim = ii + mi
            jjm = jj + mj
            jjmg = jjm - iim + nhbw + 1
            xij = xmatx(im,jm)
            gstif(iim,jjmg) = gstif(iim,jjmg) + xij
   35     continue
   40   continue
        j = j + jdof
        if (nj.lt.nnode) goto 30
        i = i + idof
        if (ni.lt.nnode) goto 25
        write(6,*)'Stiffness matrix of element ',l,' assembled.'
   50 continue
c...end of element loop
c
      return
      end
c
c
      subroutine solve(mpoin,mtotv,mbw,npoin,nbw,nkode,
     1   coord,aslod,asdis,gstif,ndof,nvar,
     2   ndbdry,dbdcmp,dbdcrd,dbdval)
      dimension gstif(mtotv,mbw),aslod(mtotv),asdis(mtotv),
     1     nkode(mpoin),ndof(mpoin),nvar(mpoin),
     2   dbdcrd(8),dbdval(8),coord(mpoin,2)
      character*1 dbdcmp(8)
c
c...solve matrix system  gstif*asdis=aslod
c
      ntotv = nvar(npoin) + ndof(npoin)
c...add 10e12 to diagonal elements for fixed dof, and
c...modify rows for Dirichlet boundary nodes  
      call fixdof(mtotv,mbw,ntotv,nbw,mpoin,npoin,gstif,nkode,
     1    coord,aslod,ndof,nvar,ndbdry,dbdcmp,dbdcrd,dbdval)
c...solution of equations by Gaussian Elimination
      write(6,*)
      write(6,*)'Starting equation solution: reduction...'
      call reduce(mtotv,mbw,ntotv,nbw,gstif,aslod)
      write(6,*)'Starting back-substitution...'
      call backsu(mtotv,mbw,ntotv,nbw,gstif,aslod,asdis)
c...calculate reactions at fixed dofs, put reactions into aslod
      write(6,*)'Starting calculation of reactions ...'
      call reactn(mtotv,mbw,mpoin,ntotv,nbw,npoin,
     1gstif,asdis,aslod,nkode,ndof,nvar)
      return
      end
c
c
      subroutine result(melem,mpoin,mnode,msets,mcomp,mtotv,mstyp,
     1            nelem,npoin,lnods,lnodn,lsidn,coord,nkode,
     3            lpros,props,aslod,asdis,lmtyp,ndof,nvar)
c
c...output the results to file
c
      dimension aslod(mtotv),props(msets,mcomp),
     1         lnods(melem,mnode),coord(mpoin,2),endis(16),encod(2,8),
     2        shape(8),deriv(2,8),cartd(2,8),dmatx(3,3),  
     3       bmatx(3,16),gploc(2,4),lsidn(melem),
     4      gpwts(4),gpsig(3),gpeps(3),gpcod(2),
     5      nkode(mpoin),asdis(mtotv),lnodn(melem),lpros(melem)
     6     ,lmtyp(melem),ndof(mpoin),nvar(mpoin)
c
c...output displacements and reactions
c...output to screen (unit 6):
      write(6,*)
      write(6,*)'RESULTS'
      write(6,*)'======='
      write(6,*)
      write(6,610)
  610 format(//' nodal reactions, displacements, temperatures'
     1/40('-')//3x,'node',
     212x,'x-reaction',5x,'y-reaction',9x,'x-displt',7x,'y-displt'
     3,9x,' temprtur')
c...output to results file (unit 16):
      write(16,*)
      write(16,*)'RESULTS'
      write(16,*)'======='
      write(16,*)
      write(16,610)
c...and output to end of data file, for postprocessor:
      write(17,*)'  node  x-displt  y-displt  temperature'
      do 20 n=1,npoin  
        i = nvar(n)
        ndofnn = ndof(n)
        dispx = asdis(i+1)
        dispy = asdis(i+2)
        if (ndofnn.eq.3) then 
          theta = asdis(i+3)
          write(17,'(i5,3e10.3)') n,dispx,dispy,theta
        else        
          write(17,'(i5,2e10.3)') n,dispx,dispy
        endif
        if (mod(n,25).eq.0) call prompt
        if (ndofnn.eq.2) then
          write(6,615) n,aslod(i+1),aslod(i+2),dispx,dispy 
          write(16,615) n,aslod(i+1),aslod(i+2),dispx,dispy
        else
          write(6,615) n,aslod(i+1),aslod(i+2),dispx,dispy
     1                  ,theta
          write(16,615) n,aslod(i+1),aslod(i+2),dispx,dispy
     1                  ,theta 
        endif 
  615   format
     1   (3x,i3,11x,e10.3,5x,e10.3,10x,e10.3,2(5x,e10.3))    
   20 continue
c
      call prompt
c...element Gauss pt stressess
  620 format(//'Element Gauss pt. stresses:'/27('-')/)
      write(6,620)
      write(16,620)
      write(17,*)
     1' elt gp    x-coord   y-coord   sigma-x   sigma-y   tau-xy'
      do 100 l=1,nelem
        if (mod(l,3).eq.0) call prompt
        write(6,630) l
        write(16,630) l
  630   format(//' element',i5/1x,12('-')/)
        lset = lpros(l)
        nnode = lnodn(l)
        nside = lsidn(l)
cp...only consider the 2 displacement dof's, so ...        
        nevab = 2*nnode
        ym = props(lset,1)
        pr = props(lset,2)
c...store nodal coords, displts for elt in encod, endis
        do 25 n=1,nnode
          nn = lnods(l,n)
          encod(1,n) = coord(nn,1)
          encod(2,n) = coord(nn,2)
cp...we only want to extract the 2 displt dof's
          i = 2*(n-1)
cp...have to use nvar for their location in global vector
          ii = nvar(nn)
          endis(i+1) = asdis(ii + 1)
          endis(i+2) = asdis(ii + 2)
   25   continue
        call dmat(mstyp,ym,pr,dmatx)
c...find Gauss-pt stresses & strains
        write(6,640) (i,i=1,3)
        write(16,640) (i,i=1,3)
  640   format(/' gauss-pt stresses and strains'/6x,'x-coord',8x,
     1  'y-coord',10x,3('stress (',i1,')',5x))
        call gauss(nnode,gploc,gpwts,ngaus)
c...loop over each Gauss point
        do 80 ij=1,ngaus
          s = gploc(1,ij)
          t = gploc(2,ij)
          call sfr(nnode,s,t,shape,deriv)
          call jacob(nnode,encod,shape,deriv,cartd,djacb)
          call bmat(nnode,cartd,bmatx)
c...get Gauss point coords in gpcod
          do 27 m=1,2
            sum = 0.0
            do 26 n=1,nnode
   26       sum = sum + encod(m,n)*shape(n)
            gpcod(m) = sum
   27     continue
c...find strains as bmatx*displts
          do 30 i=1,3
            sum = 0.0
            do 29 k=1,nevab
   29       sum = sum + bmatx(i,k)*endis(k)
            gpeps(i) = sum
   30     continue
c...stresses = dmatx*strains
          do 40 i=1,3
            sum = 0.0 
            do 39 k=1,3
   39       sum = sum + dmatx(i,k)*gpeps(k)
            gpsig(i) = sum
   40     continue
c...only stresses printed; modify this to print strains too
          write(17,'(i5,i2,3x,5e10.3)') 
     1       l,ij,(gpcod(i),i=1,2),(gpsig(i),i=1,3)
          write(6,650) (gpcod(i),i=1,2),(gpsig(i),i=1,3)
          write(16,650) (gpcod(i),i=1,2),(gpsig(i),i=1,3)
  650     format(3x,2(e10.3,5x),5x,3(e10.3,5x))
   80   continue
c...end of Gauss loop
  100 continue
c...end of element loop
c
      write(6,*)
      write(6,*)'END OF RESULTS'
      write(6,*)
      write(16,*)
      write(16,*)'END OF RESULTS'
      write(16,*)
      return
      end
c
c
c...subroutines for STIFF     
c
c
      subroutine gauss(nnode,gploc,gpwts,ngaus)
      dimension gploc(2,4),gpwts(4)
c...set up Gauss pts & weights
      do 5 i=1,4
        gpwts(i) = 0.0
        do 5 j=1,2
          gploc(j,i) = 0.0
    5 continue
      if (nnode.eq.4) then
cp...1-point integration for linear quadrilateral elt
        ngaus = 1
	GPLOC(1,1) = 0.
	GPLOC(2,1) = 0.
	GPWTS(1) = 4.0
      else
cp...2*2 integration of quadratic elt
        ngaus = 4
        r13 = 1.0/sqrt(3.0)
        gploc(1,1) = -r13
        gploc(1,2) = -r13
        gploc(1,3) = r13
        gploc(1,4) = r13
        gploc(2,1) = -r13
        gploc(2,2) = r13
        gploc(2,3) = -r13
        gploc(2,4) = r13
        do 50 i=1,4
          gpwts(i) = 1.0
   50   continue
      endif
      return
      end
c
c
      subroutine dmat(mstyp,ym,pr,dmatx)
      dimension dmatx(3,3)
c...ym,pr are Young's mod., Poisson's ratio
c...g is the elastic shear modulus
      g = 0.5*ym/(1.0 + pr)
      if (mstyp.eq.1) then
c...form the D matrix for plane stress
        f = 0.5*ym/(1.0 - pr)
        dxx = f + g
        dxy = f - g
      else
c...form the D matrix for plane strain
        dxy = g*pr/(0.5 - pr)
        dxx = 2.*g + dxy
      endif
      dmatx(1,1) = dxx
      dmatx(1,2) = dxy
      dmatx(1,3) = 0.0
      dmatx(2,1) = dxy
      dmatx(2,2) = dxx
      dmatx(2,3) = 0.0
      dmatx(3,1) = 0.0
      dmatx(3,2) = 0.0
      dmatx(3,3) = g
      return
      end
c
c
      subroutine bmat(nnode,cartd,bmatx)
      dimension cartd(2,8),bmatx(3,16)
c...form the B matrix
      do 5 j=1,16
      do 5 i=1,3
        bmatx(i,j) = 0.0
    5 continue
      j = 0
      do 10 n=1,nnode
        i = j + 1
        j = i + 1
        bmatx(1,i) = -cartd(1,n)
        bmatx(2,j) = -cartd(2,n)
        bmatx(3,i) = -cartd(2,n)
        bmatx(3,j) = -cartd(1,n)
c...negative sign for 'compression +ve' sign convention
   10 continue
      return
      end
c
c
      subroutine sfr(nnode,s,t,shape,deriv)
      dimension shape(8),deriv(2,8)
c...evaluate the shape functions and 1st derivs at local coords s,t
c
      if (nnode.eq.4) then
cp...linear quadrialteral
        ST = S*T
        SHAPE(1) = (1.-S-T+ST)*.25
        SHAPE(2) = (1.+S-T-ST)*.25
        SHAPE(3) = (1.+S+T+ST)*.25
        SHAPE(4) = (1.-S+T-ST)*.25
        DERIV(1,1) = -(1.-T)*.25
        DERIV(1,2) =  (1.-T)*.25
        DERIV(1,3) =  (1.+T)*.25
        DERIV(1,4) = -(1.+T)*.25
        DERIV(2,1) = -(1.-S)*.25
        DERIV(2,2) = -(1.+S)*.25
        DERIV(2,3) =  (1.+S)*.25
        DERIV(2,4) =  (1.-S)*.25
      else
        s2 = 2.0*s
        t2 = 2.0*t
        ss = s*s
        tt = t*t
        st = s*t
        sst = ss*t
        stt = s*tt
        st2 = 2.0*st
        shape(1) = (-1.0 + st + ss + tt - sst - stt)*0.25
        shape(2) = (1.0 - t - ss + sst)*0.5
        shape(3) = (-1.0 - st + ss + tt - sst + stt)*0.25
        shape(4) = (1.0 + s - tt - stt)*0.5
        shape(5) = (-1.0 + st + ss + tt + sst + stt)*0.25
        shape(6) = (1.0 + t - ss - sst)*0.5
        shape(7) = (-1.0 - st + ss + tt + sst - stt)*0.25
        shape(8) = (1.0 - s - tt + stt)*0.5
        deriv(1,1) = (t + s2 - st2 - tt)*0.25
        deriv(1,2) = -s + st
        deriv(1,3) = (-t + s2 - st2 + tt)*0.25
        deriv(1,4) = (1.0 - tt)*0.5
        deriv(1,5) = (t + s2 + st2 + tt)*0.25
        deriv(1,6) = -s - st
        deriv(1,7) = (-t + s2 + st2 - tt)*0.25
        deriv(1,8) = (-1.0 + tt)*0.5
        deriv(2,1) = (s + t2 - ss - st2)*0.25
        deriv(2,2) = (-1.0 + ss)*0.5
        deriv(2,3) = (-s + t2 - ss + st2)*0.25
        deriv(2,4) = -t - st              
        deriv(2,5) = (s + t2 + ss + st2)*0.25
        deriv(2,6) = (1.0 - ss)*0.5
        deriv(2,7) = (-s + t2 + ss - st2)*0.25
        deriv(2,8) = -t + st
      endif
c
      return
      end
c
c     
      subroutine jacob(nnode,encod,shape,deriv,cartd,djacb)
      dimension shape(8),deriv(2,8),cartd(2,8),encod(2,8),xjacm(2,2),
     1         xjaci(2,2)
c...evaluate the Jacobian matrix, and its determinant
      do 5 j=1,2
      do 5 i=1,2
        sum = 0.0
        do 4 n=1,nnode
    4   sum = sum + deriv(i,n)*encod(j,n)
        xjacm(i,j) = sum
    5 continue
c...xjacm is the Jacobian matrix. find det & inverse by Cramer        
      djacb = xjacm(1,1)*xjacm(2,2) - xjacm(1,2)*xjacm(2,1)
      if (djacb.lt.1.0e-12) then
        write(6,*)'ERROR: determinant of Jacobian is -ve or zero'
        write(6,*)'Element is inside out, or nodes defined clockwise'
        call prompt
        stop
      endif
      xjaci(1,1) = xjacm(2,2)/djacb
      xjaci(2,2) = xjacm(1,1)/djacb
      xjaci(1,2) = -xjacm(1,2)/djacb
      xjaci(2,1) = -xjacm(2,1)/djacb
c...get Cartesian derivs of shape functions using J inverse
      do 50 i=1,2
      do 50 n=1,nnode
        sum = 0.0
        do 49 j=1,2
   49   sum = sum + xjaci(i,j)*deriv(j,n)
        cartd(i,n) = sum
   50 continue
      return
      end
c
c
c...subroutines for SOLVE
c
c
      subroutine fixdof(mtotv,mbw,ntotv,nbw,mpoin,npoin,gstif,nkode,
     1    coord,aslod,ndof,nvar,ndbdry,dbdcmp,dbdcrd,dbdval)
c
      character*1 dbdcmp(8),xychar
      dimension gstif(mtotv,mbw),kd(3),nkode(mpoin),coord(mpoin,2),
     1  aslod(mtotv),dbdcrd(8),dbdval(8),ndof(mpoin),nvar(mpoin)
c
c...note that as nbw is odd, we simply divide by 2 in integer arithmetic
c...to get the half-bandwidth nhbw
      nhbw = nbw/2
      nhbw1 = nhbw + 1
c...modify the diagonal coefft of fixed d.o.f.s
      do 20 nn=1,npoin
        kode = nkode(nn)
cp...have to use ndof(nn) rather than 2 for ndof
        ndofnn = ndof(nn)
        nvarnn = nvar(nn)
        if (ndofnn.eq.2) then
          nkd = 2
          kd(1) = kode/10
          kd(2) = kode - kd(1)*10
        else
          nkd = 3
          kd(1) = kode/100
          k2 = kode - kd(1)*100
          kd(2) = k2/10
          kd(3) = k2 - kd(2)*10
        endif
c...only consider the displt dof's; temperature dof dealt with afterwards
        do 10 m=1,2
          if (kd(m).eq.0) goto 10
          nposn = nvarnn + m
          gstif(nposn,nhbw1) = gstif(nposn,nhbw1) + 10.0e12
   10   continue
c
c...now check if temperature dof exists and is on Dirichlet bdry
        if (nkd.eq.3) then
          kd3 = kd(3)
          if (kd3.eq.0) goto 20
c...node on the Dirichlet boundary
          nn3 = nvarnn + 3
          x = coord(nn,1)
          y = coord(nn,2)
c...find potential to apply to this node
c...default is potential from function bdryfn
          g = bdryfn(x,y)
          if (ndbdry.gt.0) then
            do 15 i=1,ndbdry
c...check each boundary condition against coordinates of node
              xychar = dbdcmp(i)
              xyval = dbdcrd(i)
              if ((xychar.eq.'x'.and.abs(x-xyval).lt.1.d-6).or.
     1        (xychar.eq.'y'.and.abs(y-xyval).lt.1.d-6)) 
     2         g = dbdval(i)
   15       continue
          endif
          if (abs(g).lt.1.d-6) then
            write(6,*)'Zero temperature fixed at node ',nn
c...value is zero. Add a large no. to the diagonal coefft
            gstif(nn3,nhbw1) = gstif(nn3,nhbw1) + 10.0e12
            aslod(nn3) = 0.0
          else
            write(6,*)'Nonzero temperature fixed at node ',nn
c...multiply diagonal coefft by a large no.
            gstif(nn3,nhbw1) = gstif(nn3,nhbw1)*1.0e12
c...and change r.h.s. component accordingly
            aslod(nn3) = g*gstif(nn3,nhbw1)
          endif
        endif
c
   20 continue
      return
      end
c
c
      subroutine reduce(mtotv,mbw,ntotv,nbw,gstif,aslod)
      dimension gstif(mtotv,mbw),aslod(mtotv)
c...Gaussian elimination, using compact storage of band
c...K(i,j) is stored at gstif(i,j-i+nhbw+1) for j>=i
c
      nhbw = nbw/2
      nhbw1 = nhbw + 1
      do 20 i=1,ntotv-1
        k1 = i + 1
        k2 = i + nhbw 
        if (k2.gt.ntotv) k2 = ntotv
        pivot = gstif(i,nhbw1)
        if (abs(pivot).lt.1.0e-12) then
          write(6,*)'ERROR: zero or negative pivot in row',i
          write(6,*)'Pivot = ',pivot
          call prompt
          stop
        endif
        do 10 k=k1,k2
          fact = gstif(k,i-k+nhbw1)/pivot
          j1 = i + 1
          j2 = i + nhbw 
          if (j2.gt.ntotv) j2 = ntotv
          do 5 j=j1,j2
            ji = j - i + nhbw1
            jk = j - k + nhbw1
            gstif(k,jk) = gstif(k,jk) - fact*gstif(i,ji)
    5     continue
          aslod(k) = aslod(k) - fact*aslod(i)
   10   continue
   20 continue
      return
      end
c
c
      subroutine backsu(mtotv,mbw,ntotv,nbw,gstif,aslod,asdis)
      dimension gstif(mtotv,mbw),aslod(mtotv),asdis(mtotv)
c...back-substitution phase
      nhbw = nbw/2
      nhbw1 = nhbw + 1
      asdis(ntotv) = aslod(ntotv)/gstif(ntotv,nhbw1)
      do 10 i=ntotv-1,1,-1
        pivot = gstif(i,nhbw1)
        if (abs(pivot).lt.1.0e-12) then
          write(6,*)'ERROR: zero pivot in row',i
          call prompt
          stop
        endif
        j1 = i + 1
        j2 = i + nhbw
        if (j2.gt.ntotv) j2 = ntotv
        do 5 j=j1,j2
          ji = j - i + nhbw1
          aslod(i) = aslod(i) - gstif(i,ji)*asdis(j)
    5   continue
        asdis(i) = aslod(i)/pivot
   10 continue
      return
      end
c
c
      subroutine reactn(mtotv,mbw,mpoin,ntotv,nbw,npoin,
     1    gstif,asdis,aslod,nkode,ndof,nvar)
c...calculate reactions, and store them in aslod.
c...zero the other entries in aslod.
      dimension gstif(mtotv,mbw),asdis(mtotv),
     1          kd(3),nkode(mpoin),aslod(mtotv)
     1          ,ndof(mpoin),nvar(mpoin)

c
      nhbw = nbw/2 
      nhbw1 = nhbw + 1
      do 5 i=1,ntotv
        aslod(i) = 0.0
    5 continue
      do 20 nn=1,npoin
        kode = nkode(nn)
cp...have to use ndof(nn) rather than 2 for ndof
        ndofn = ndof(nn)
        nvarnn = nvar(nn)
        if (ndofn.eq.2) then
          nkd = 2
          kd(1) = kode/10
          kd(2) = kode - kd(1)*10
        else
          nkd = 3
          kd(1) = kode/100
          k2 = kode - kd(1)*100
          kd(2) = k2/10
          kd(3) = k2 - kd(2)*10
        endif
c...only consider displt dof's
        do 10 m=1,2
          if(kd(m).eq.0) goto 10
          nposn = nvarnn + m
          aslod(nposn) = -gstif(nposn,nhbw1)*asdis(nposn)
          asdis(nposn) = 0.0
   10   continue
   20 continue
      return
      end
c
c
