c...Plane strain elasto-viscoplasticity "main engine"
c...for use with FELIPE version 3 - sample datafile is vpcyl4.dat
      program vplas
c
c
c...If you want to re-compile this program, you must link it with
c...the file felsb.obj, compiled from felsb.for (provided in the
c...FELIPE folder). The file felsb.for contains additional subroutines 
c...used by all the 'main engines' except eladv.for. 
c...If you don't use linking, you should paste the contents of 
c...the file felsb.for to the end of this file, before compiling
c...and running it.
c
cv...This program performs elast-viscoplasticity analyses.
cv...It is developed from the plasticity program plast.for
cv...Added code for viscoplasticity is preceded by comment lines
cv...which begin with cv (like these lines!)
cv...VPLAS allows for non-associated flow: therefore a 
cv...frontal solution routine for nonsymmetric matrices is used.
cv...plane strain conditions, with material thickness = 1.0
c
CV...THE PROGRAM (AND FELSUB.FOR) SHOULD BE COMPILED IN DOUBLE PRECISION:
CV...     ftn77  /dreal vplas.for         in Salford Fortran77
c
      parameter(melem=800,mpoin=2000,mnode=8,
     1msets=9,mcomp=6,mtotv=4000,mfixv=150,
     2mfron=150,mstif=22500,mbufa=100)
c...these parameters fix the dimensions of the arrays:
c...melem = max. no. of elements
c...mpoin = max. no. of nodes
c...mnode = max. no. of nodes in a single element
c...msets = max. no. of mat. prop. sets (also used for loading sets)
c...mcomp = max. no. of mat. props = 6 for (visco)plasticity
c...mtotv = max no. of degrees of freedom (= 2*mpoin for elasticity)
cv...mfixv = max no. of fixed degrees of freedom
cv...mfron = max frontwidth in solver FRONT
cv...mstif = mfron*mfron in non-symmetric case
c
c
      character char*1
      character*80 title,infile,outfil,prtfil
c
      dimension lnods(melem,mnode),lnodn(melem),lsidn(melem),
     1 coord(mpoin,2),nkode(mpoin),aslod(mtotv),ndof(mpoin),
     2        lpros(melem),props(msets,mcomp),asdis(mtotv),
     4      surft(msets,2),lmtyp(melem)
c...lnods(l,n) = global no. of n'th node of element l
c...lnodn(l) = no. of nodes in element l ( = 8 for 8-noded quad.)
c...lsidn = no. of sides in element l    ( = 4 for 8-noded quad.)
c...coord(np,1), coord(np,2) = x,y coords of node np
c...nkode(np) = fixity code of node np
c...aslod is global nodal load vector
c...ndof(np) = no. of dof.s at node np (not used in this program)
c...lpros(l) = mat. prop. set no. of element l
c...props(m,i) = value of i'th mat. property in mat. prop. set m
c...asdis is global nodal displacement vector
c...surft(m,1),surft(m,2) = normal,tangential surface tractions in set m
c...lmtyp(l) = material type of elt l (not used in this program)
c
      real*8 pival,eqrhs(mbufa)
     1      ,vecrv(mfron),gstif(mstif),gload(mfron),equat(mfron,mbufa)
      DIMENSION NPIVO(MBUFA),NAMEV(MBUFA),NACVA(MFRON),nfvar(mfixv),
     1          spdis(mfixv),nvar(mpoin)
cb...arrays above are used in the FRONT frontal solver
c 
      dimension strsg(3,4,melem),strsg1(3,4,melem),totld(mtotv),
     1          accld(mtotv),accds(mtotv),aslod1(mtotv),resld(mtotv),
     2          fincs(9),mohrg(4,melem),vivel(3,4,melem)
cc..strsg holds stresses at Gauss points (Cartesian)
cc..totld is the total load to be applied
cc..accld is the accumulated load applied so far
cc..aslod is the load being applied in the current increment
cc..accds is the accumulated displacements 
cc..resld is the residual force vector
cc..fincs is the set of load increments
cc..mohrg is the yield code at Gauss pts: 0=elastic, 1=plastic
cv...vivel stores the viscoplastic strain velocities ("epsilon dot")
c
c
cc..tolerances for convergence tests:
      rtoler = 0.1
      stoler = 1.e-8
cc..rtoler is pl. strain velocities tolerance (%) for the timesteps
cc..stoler is tolerance for stresses in local (stress) iterations
cc..stoler is used in subroutine stress
c
      write(6,*)'Program VPLAS: elasto-viscoplasticity (Mohr-Coulomb).'
      write(6,*)'Handles 8-noded quadrilateral elements.'
      write(6,*)'Allows for non-associated flow.'
      write(6,*)'Dimensioned for',melem,' elements, and',mpoin,' nodes.'
      write(6,*)
c
c...read an existing .dat file
c
    1 continue
c
c...get the input file name in infile
      call askfil(infile,iflag)
      if (iflag.eq.1) then
        call yesno(char,'Do you want to exit?')
        if (char.eq.'y') stop
        goto 1
      endif
c
      write(6,*)'Opening file ',infile,'...'
      lenin = lenstr(infile)
      open(15,file=infile(1:lenin),status='old',err=999)
      write(6,*)
      lm4 = lenin - 4
      prtfil = infile(1:lm4)//'.prt'
      write(6,*)'Creating file ',prtfil(1:lenin),' for results...'
      write(6,*)
      open(16,file=prtfil(1:lenin),status='unknown',err=999)      
      outfil = infile(1:lm4)//'.out'
      write(6,*)'Creating file ',outfil(1:lenin),' for output...'
      write(6,*)
      open(17,file=outfil(1:lenin),status='unknown',err=999)      
c
c...read the input mesh data, and echo it to the output files
      call input(15,melem,mpoin,mnode,msets,mcomp,nelem,
     1 npoin,lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props
     2           ,ndofn,npros,title,ndof)
      call print(16,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1 lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2           npros,title,ndof)
      call output(17,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1 lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2            npros,title,ndof)
      call yesno(char,'Do you want the mesh details listed?')
      if (char.eq.'y') then
        call print(6,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1  lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2           npros,title,ndof)
      endif
c
      iflag = 0
      do 3 l=1,nelem
        nnode = lnodn(l)
        nside = lsidn(l)
        nodsid = 10*nnode + nside
        if (nodsid.ne.84) then
          iflag = 1
          write(6,*)'ERROR: Cannot handle element ',l
          write(6,*)'which has ',nside,' sides and ',nnode,' nodes'
        endif
    3 continue
      if (iflag.eq.1) then
        call prompt
        stop
      endif
c
c...iprint tells if we have output any results yet
      iprint = 0
c 
cc..zero the accumulating variables, stresses, etc.
cv...and vivel
c
      do 10 i=1,mtotv
        accds(i) = 0.0
        accld(i) = 0.0
        resld(i) = 0.0
   10 continue
      do 11 i=1,9
        fincs(i) = 0.0
   11 continue  
      do 12 l=1,melem
        do 12 ij=1,4
          do 12 i=1,3
            strsg(i,ij,l) = 0.0
            strsg1(i,ij,l) = 0.0
            vivel(i,ij,l) = 0.0
   12 continue
      do 13 l=1,melem
        do 13 ij=1,4
          mohrg(ij,l) = 0
   13 continue
c
cp...create the pointer array nvar. Needed because a node may
cp...have 2 or 3 d.o.f.s; this info stored in ndof
cp...also store info on fixed d.o.f.s in nfvar and spdis,
cp...rather than in nkode
      ntotv = 0
      nfixv = 0
      do 5 np=1,npoin
        ndofnp = ndof(np)
        nvar(np) = ntotv
        kode = nkode(np)
C...SORT OUT KODE AND STORE FIXED VARIABLE NO. IN NFVAR()
        KOD = KODE
        DO 4 M=1,NDOFNP
          KD = KOD/10
          K = KOD - KD*10
          KOD = KD
          IF(K.ne.0) then
            NFIXV=NFIXV+1
            IF (NFIXV.GT.MFIXV) THEN
          WRITE(6,*)'No. of fixed dof.s exceeds parameter MFIXV'
              write(6,*)'Increase MFIXV and recompile.'
              call prompt
              stop
            endif
            M1 = NDOFNP + 1 - M
            NFVAR(NFIXV) = ntotv + M1
            spdis(nfixv) = 0.0
          endif
    4   continue
        ntotv = ntotv + ndofnp
    5 continue
      nsdis = nfixv
cp...s/r LOAd will add spec displts, total no will be held in nsdis
cp...Now local dof i of node np is global dof nvar(np)+i
cp...and nfvar(i) is global dof of i'th fixed dof
cp...and ntotv is total no. of degrees of freedom
c
cc..read the equivalent load vector, into totld, and then
cc..read the load increment sizes
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine LOAD...'
      call load(melem,mpoin,mnode,msets,mcomp,mtotv,
     1          nelem,npoin,lnods,lnodn,coord,
     2  surft,totld)
c
c
c...read boundary data header (no boundary set data expected)
      read(15,*)
      read(15,*)nbrad,nbdir
      if (nbrad.gt.0.or.nbdir.gt.0) then
        write(6,*)
     1  'ERROR: boundary set data present - cannot cope!'
        call prompt
        stop
      endif
c...read data defining load increments
      read(15,*) 
      read(15,*) nincs,(fincs(i),i=1,nincs)
cv...read the timestepping control parameters
      read(15,*)
      read(15,*) dtint,ftime,allow,theta
cv...theta is the time discretization parameter: =0 for explicit algorithm,
cv...                    between 0.5 and 1.0 for stable implicit algorithm.
cv...dtint is the initial timestep length
cv...if no new g.pts. have yielded, next timestep is FTIME times previous one
cv...                    (DNEXT = FTIME*DTIME)
cv...max. timestep is ALLOW times initial one (ALLOW*DTINT)
c
c
c...open a scratch file to write the element stiffness matrices on
      open(unit=18,status='scratch',form='unformatted')  
cv...open further scratch files for use in FRONT
      OPEN(19,STATUS='SCRATCH',FORM='UNFORMATTED')
      OPEN(20,STATUS='SCRATCH',FORM='UNFORMATTED')
      OPEN(21,STATUS='SCRATCH',FORM='UNFORMATTED')
      PIVAL = 1.d-8
c
c
c
cc..start of load increment loop
cc..============================
c
cv...initialize the timestepping
      ttime = 0.0
      dtime = dtint
c
      irsol = 0
cv...irsol = 0 for first solution, in front or afront
cv...irsol = 1 for re-soln with new matrix having same structure
cv...irsol = 2 for re-solution with already-decomposed matrix
      isymm = 1
cv...isymm = 1 for symmetric matrix, isymm=0 for nonsymm matrix
c 
      do 200 inc=1,nincs
        finc = fincs(inc)
        write(6,*)' Load increment ',inc
        write(6,*)
        write(6,*)' Increment size: ',finc
        write(6,*)
cc..increment load, adding residual from previous increment
        do 20 i=1,ntotv
          asdis(i) = 0.0
          aslod(i) = finc*totld(i)
          accld(i) = accld(i) + aslod(i)
          aslod(i) = aslod(i) + resld(i)
   20   continue
cv...get norm of the loads (to use in testing if steady state reached)
        call norm2(mtotv,ntotv,aslod,fnorm)
        do 21 l=1,melem
          do 21 ij=1,4
            do 21 i=1,3
              strsg1(i,ij,l) = strsg(i,ij,l)
   21   continue
cc..strsg will stay fixed as the start-of-timestep stress
cc..strsg1 will hold the end-of-timestep stress 
c
        kount = 0
cc..start of global timestepping 
cc..============================
  100   continue
          kount = kount + 1
cc..note: stiffness depends on current stress state
          call stiff(melem,mpoin,mnode,msets,mcomp,
     1           nelem,npoin,lnods,lnodn,coord,lpros,props
     2           ,strsg,mohrg,dtime,theta)
c
          do 110 i=1,ntotv
            aslod1(i) = aslod(i)
  110     continue
c          call solve(mpoin,mtotv,mhbw,npoin,nhbw,nkode,
c     1           aslod1,asdis,gstif)
c
cb...the next subroutine replaces ASSMB and SOLVE in ELAST.FOR
c          call prompt
          if (isymm.eq.1) then
            write(6,*)
            write(6,*)'Starting subroutine FRONT...'
            CALL FRONT(MTOTV,MPOIN,MELEM,MFIXV,MNODE,MFRON,MSTIF,
     1NPOIN,NELEM,NSDIS,LNODS,LNODN,NDOF,NVAR,NFVAR,ASLOD1,ASDIS,SPDIS,
     2VECRV,GSTIF,GLOAD,EQUAT,NACVA,
     3PIVAL,IRSOL,EQRHS,NPIVO,NAMEV,MBUFA)
            irsol = 1
cv...if time parameter theta=0.0, only elastic matrix is used
            if (theta.eq.0.0) irsol = 2
          else
cv...non-symmetric matrix eqn must be solved
            CALL AFRONT(MTOTV,MPOIN,MELEM,MFIXV,MNODE,MFRON,MSTIF,
     1NPOIN,NELEM,NSDIS,LNODS,LNODN,NDOF,NVAR,NFVAR,ASLOD1,ASDIS,SPDIS,
     2VECRV,GSTIF,GLOAD,EQUAT,NACVA,
     3PIVAL,IRSOL,EQRHS,NPIVO,NAMEV,MBUFA)
          endif
c
c
cc..now evaluate new stresses, from this displt increment, in strsg1
cv..In v.p. program, mohrg is not updated here
          write(6,*)'Evaluating new stresses...'     
          call stress(melem,mpoin,mnode,msets,mcomp,mtotv,
     1            nelem,npoin,lnods,lnodn,coord,lpros,props,
     2            asdis,strsg,strsg1,mohrg,stoler
     3            ,vivel,dtime,theta)
c
cv...update stresses and displacements
        do 160 i=1,ntotv
          accds(i) = accds(i) + asdis(i)
  160   continue
        do 161 l=1,nelem
          do 161 ij=1,4
            do 161 i=1,3
              strsg(i,ij,l) = strsg1(i,ij,l)
  161   continue
c
cv...calculate new plastic strain velocities in vivel, & update mohrg
          write(6,*)'Updating yield...'
          call strain(melem,msets,mcomp,nelem,lpros,props,
     1                    strsg,vivel,mohrg,dtime,theta,newfl)
c
cv...set next timestep - newfl tells if new g.pt. has yielded
          ttime = ttime + dtime
          dnext = ftime*dtime
          if (newfl.ne.0) dnext = dtint
          if (dnext.gt.allow*dtint) dnext = allow*dtint
          dtime = dnext
c
cv...also set irsol and isymm: 
cv...matrix becomes nonsymm if any yielding has started
          if (isymm.eq.1.and.newfl.ne.0.and.theta.ne.0.0) isymm=0
          if (newfl.eq.0.and.irsol.eq.0) then
            irsol = 2
            isymm = 1
          else
            irsol = 1
          endif
          write(6,*)'irsol, isymm:',irsol,isymm
c
cv...now form new r.h.s. for next timestep in aslod
          write(6,*)'Forming new right-hand side vector...'
          call newrhs(melem,mpoin,mnode,mtotv,msets,mcomp,
     1           nelem,npoin,ntotv,lpros,props,lnods,lnodn,coord,
     2           theta,dtime,nkode,strsg,aslod,vivel,mohrg)
c
cv...check if steady state reached
          write(6,*)'Checking for steady state...'
          call steady(melem,nelem,vivel,mohrg,kount,first,ratio)
          write(6,*)'Time = ',ttime,' pl. strain ratio = ',ratio
          if (ratio.gt.rtoler) goto 100
c
cc..end of timestepping for this increment
cc..======================================
  150   continue
        write(6,*) kount,' global timesteps'
        write(6,*)
c
cc..compute residual force vector
        call residu(melem,mpoin,mnode,mtotv,nelem,npoin,ntotv,
     1              lnods,lnodn,coord,nkode,accld,strsg,resld)
c
        if (inc.lt.nincs) then
c...output results if required
          call yesno(char,'Do you want the results output?')
          if (char.eq.'y') then
            if (iprint.eq.1) then
              write(6,666) inc,ttime
              write(16,666) inc,ttime
              write(17,666) inc,ttime
            endif
            iprint = 1
            call result(melem,mpoin,mnode,msets,mcomp,mtotv,
     1            nelem,npoin,lnods,lnodn,coord,nkode,lpros,props,
     2            accld,accds,strsg,mohrg)
          endif
        endif
c
  200 continue
cc..end of load increment loop
cc..==========================
c
      if (iprint.eq.1) then
        write(6,666) nincs,ttime
        write(16,666) nincs,ttime
        write(17,666) nincs,ttime
  666   format(' End of load inc ',i5,'. Time = ',e10.3)
      endif
      call result(melem,mpoin,mnode,msets,mcomp,mtotv,
     1            nelem,npoin,lnods,lnodn,coord,nkode,lpros,props,
     2            accld,accds,strsg,mohrg)
c
      write(6,*)
      write(6,*)'VPLAS completed. Results are in file ',prtfil(1:lenin)
      write(6,*)
      write(6,*)
      call prompt
      stop
  999 write(6,*)
     1'Cannot open file with this name. Terminating program.'
      call prompt
      stop
      end
c
c
c      
      subroutine stiff(melem,mpoin,mnode,msets,mcomp,
     1             nelem,npoin,lnods,lnodn,coord,lpros,props,
     2             strsg,mohrg,dtime,theta)
c      
c...calculates the element stiffness matrices, and writes them
c...to the scratch file.
c
      dimension props(msets,mcomp),lnods(melem,mnode),encod(2,8),
     1         estif(16,16),bmatx(3,16),dmatx(3,3),dbmat(3,16),
     2        gpwts(4),lpros(melem),lnodn(melem),coord(mpoin,2),
     3       shape(8),gploc(2,4),gpcod(2),deriv(2,8),cartd(2,8)
cc..extra arrays in plasticity
     4  ,strsg(3,4,melem),gpsig(3),demat(3,3),mohrg(4,melem)
c...encod holds the nodal coords for the nodes around the current element
c...estif will be the stiffness matrix for the current element
c...bmatx, dmatx are the B and D matrices, and dbmat is D*B
c...gpwts,gploc are the Gauss point weights and locations (in xi,eta coords)
c...gpcod will be the x,y coords of the current Gauss point
c...shape holds the shape functions evaluated at the Gauss point
c...deriv holds the derivatives (w.r.t. xi, eta) of the shape functions
c...cartd holds the Cartesian derivatives (w.r.t x,y) of shape functions
c
cc..demat is the elastic stiffness matrix
c
      rewind 18
c
c...start loop over all elements
      do 80 l=1,nelem
c
c...extract Young's modulus & Poisson's ratio from props      
      lset = lpros(l)
      ym = props(lset,1)
      pr = props(lset,2)
cc..and the Mohr-Coulomb plasticity parameters k and sigma_c
      triax = props(lset,3)
      sigc = props(lset,4)
cv...and gamma
      gamma = props(lset,5)
      dilat = props(lset,6)
      thick = 1.0
c
c...put nodal coords in encod
      nnode = lnodn(l)
      do 15 n=1,nnode
      nn = lnods(l,n)
      encod(1,n) = coord(nn,1)
      encod(2,n) = coord(nn,2)
   15 continue
c...zero stiffness matrix
      do 20 j=1,16
      do 20 i=1,16
      estif(i,j) = 0.0
   20 continue
c
c...get elastic D matrix
      call dmat(ym,pr,demat)
c
c...put local coords of 2x2 guass-pts in gploc
      call gauss(gploc,gpwts,ngaus)
c
c...start loop through gauss-pts for contribution to estif
      do 75 ij=1,ngaus
c...put local coords of gauss-pt in s,t
      s = gploc(1,ij)
      t = gploc(2,ij)
c...get shape fns (shape) & derivs (deriv) in local coords
      call sfr(s,t,shape,deriv)
c...get xy-coords of gauss-pt into gpcod, using shape fn interpolation
      do 25 m=1,2
      sum = 0.0
      do 24 n=1,8
   24 sum = sum + encod(m,n)*shape(n)
   25 gpcod(m) = sum
c...get xy-derivs (cartd) & derterminant of Jacobian (djacb)
      call jacob(encod,shape,deriv,cartd,djacb)
c...get B matrix 
      call bmat(cartd,bmatx)
c
cc..extract stress at Gauss point into gpsig
      do 30 i=1,3
        gpsig(i) = strsg(i,ij,l)
   30 continue
c
      if (mohrg(ij,l).eq.0) then
cc..still elastic - use demat
        do 32 i=1,3
          do 32 j=1,3
            dmatx(i,j) = demat(i,j)
   32   continue
      else
cv..calculate elasto-plastic D matrix
        call dmatvp(triax,dilat,sigc,gamma,theta,dtime,
     1              gpsig,demat,dmatx)
      endif
cc..end of plasticity section
c
c...form matrix product D*B
      dvolu = thick*djacb*gpwts(ij)
      do 40 j=1,16
      do 40 i=1,3
      sum=0.0
      do 39 k=1,3
   39 sum = sum + dmatx(i,k)*bmatx(k,j)
   40 dbmat(i,j) = sum
c...B transpose *d*b* djacb = contribution to estif
      do 50 j=1,16
      do 50 i=1,16
      sum = 0.0
      do 49 k=1,3
   49 sum = sum + bmatx(k,i)*dbmat(k,j)
   50 estif(i,j) = estif(i,j) + sum*dvolu
   75 continue
c...end of gauss loop
c
c...write estif to file
      write (18) estif
   80 continue
c...end of element loop
c
      endfile 18
c
      return
      end
c
c
      subroutine load(melem,mpoin,mnode,msets,mcomp,mtotv,
     1                nelem,npoin,lnods,lnodn,coord,surft,aslod)
c
c...compile global nodal load vector
c
      dimension aslod(mtotv),surft(msets,2),
     1     lnods(melem,mnode),press(2,3),npals(3),
     2     coord(mpoin,2),gploc(2),shape(3),deriv(3),xjacm(2),      
     3     gpres(2),fcomp(2),encod(2,3),lnodn(melem)
c
      ntotv = npoin*2
c
      write(6,*)'Applied Loads'
      write(6,*)'-------------'
      write(6,*)
      write(16,*)'Applied Loads'
      write(16,*)'-------------'
      write(16,*)
c...initialize the global load vector
      do 10 i=1,mtotv
      aslod(i) = 0.0
   10 continue
c
c...there are some dummy lines in the input data, which are read but ignored
      read(15,*,end=99)
c...npset>0 if there are surface tractions. nbset is not used.
      read (15,510) nbset,npset
  510 format(2i5)
c
c...nodal loads (ld=1, ignored)
      write(6,*)'Nodal loads:'
      write(6,*)'-----------'
      write(6,*)
      write(6,*)'  Node   x-component   y-component'
      write(16,*)'Nodal loads:'
      write(16,*)'-----------'
      write(16,*)
      write(16,*)'  Node   x-component   y-component'
      read(15,*)
   15 read(15,520) lcard,ld,np,fx,fy
  520 format(i1,i4,i5,2e10.3)
      if (np.ne.0) then
c...the i'th component of node np must be converted to global degree 
c...of freedom component
        nm = 2*(np - 1)
c...add nodal loads into aslod
        aslod(nm + 1) = aslod(nm + 1) + fx
        aslod(nm + 2) = aslod(nm + 2) + fy
        write(6,610) np,fx,fy
        write(16,610) np,fx,fy
  610   format(i5,2(4x,e10.3))
        if (lcard.eq.0) goto 15
      endif
      write(6,*)
      write(16,*)
c
c
c...surface tractions
      if (npset.gt.0) then
        read(15,*)
        write(6,*)'Surface Tractions'
        write(6,*)'-----------------'
        write(6,*)
        write(6,*)' set   normal cmpnt   tangential cmpnt' 
        write(16,*)'Surface Tractions'
        write(16,*)'-----------------'
        write(16,*)
        write(16,*)' set   normal cmpnt   tangential cmpnt' 
c
c...first read surface traction sets
   25   read(15,530) lcard,iset,i1,i2,pn,pt
  530   format(i1,i4,3x,2i2,8x,2e10.3)
        if (iset.gt.0) then 
          surft(iset,1) = pn
          surft(iset,2) = pt
          write(6,620)  iset,pn,pt
          write(16,620) iset,pn,pt
  620     format(i3,2(5x,e10.3))
        endif
        if (lcard.eq.0) goto 25
c
c...now apply these tractions to edges
        read(15,*)
        write(6,*)
        write(6,*)'   elt  end-nodes   set-numbers'
        write(16,*)
        write(16,*)'   elt   end-nodes   set-numbers'
   30   read(15,540) lcard,nnas,l,n1,iset1,n2,iset2
c...nnas is the number of nodes along the edge (=3 for 8-noded quad.)
  540   format(i1,i4,i5,2(i5,i2,1x))
        if (l.gt.0) then
          write(6,630) l,n1,n2,iset1,iset2
          write(16,630) l,n1,n2,iset1,iset2
  630     format(i5,2(2x,2i5))
c...put the tractions in press. Interpolate linearly for the
c...tractions at the midside node.
          do 35 i=1,2
            press(i,1) = surft(iset1,i)
            press(i,3) = surft(iset2,i)
            press(i,2) = (press(i,1) + press(i,3))/2.
   35     continue
c...store the node numbers along side in npals
          npals(1) = n1
          npals(3) = n2
c...we have to find the midside node no. from lnods
          nnode = lnodn(l)
          do 42 n=1,nnode
            if (lnods(l,n).eq.npals(1)) goto 45
   42     continue
c...if loop completed, npals(1) is not in element 1
          write(6,700) npals(1),l
  700     format('ERROR: node',i4,' is not in element ',i4,'.')
          call prompt
          stop
   45     k = n + 1
          npals(2) = lnods(l,k)
c
c...extract nodal coords and store in encod
          do 60 n=1,3
            np = npals(n)
            do 55 m=1,2
   55         encod(m,n) = coord(np,m)
   60     continue
c
c...enter 1D Gauss loop along side (2-pt integration)
          r13 = 1.0/sqrt(3.0)
          gploc(1) = -r13
          gploc(2) = r13
          gpwts = 1.0
          do 90 ij=1,2
            s = gploc(ij)
            ss = s*s
            shape(1) = (ss-s)*0.5
            shape(2) = 1.0 - ss
            shape(3) = (ss + s)*0.5
            deriv(1) = s - 0.5
            deriv(2) = -2.0*s
            deriv(3) = s + 0.5
c...get jacobian xjacm
            do 75 j=1,2 
              sum = 0.0
              do 72 n=1,3
   72           sum = sum + deriv(n)*encod(j,n)
              xjacm(j) = sum
   75       continue
c...find gauss pt pressures, & convert to components in local coords
            do 80 m=1,2
              sum = 0.0
              do 77 n=1,3
   77           sum = sum + press(m,n)*shape(n)
              gpres(m) = sum
   80       continue
            fcomp(1) = (xjacm(1)*gpres(2) - xjacm(2)*gpres(1))*gpwts
            fcomp(2) = (xjacm(1)*gpres(1) + xjacm(2)*gpres(2))*gpwts
c...accumulate equivalent nodal forces in aslod
            do 85 n=1,3
              np = npals(n)
              do 82 m=1,2
                nm = 2*(np - 1) + m
                aslod(nm) = aslod(nm) + fcomp(m)*shape(n)
   82         continue
   85       continue
   90     continue
c...end of Gauss loop
c
        endif
        if (lcard.eq.0) goto 30
c...go back for next loaded edge
c
c...report the non-zero loads assembled in aslod
        call prompt
        write(6,*) 
        write(6,*)'Non-zero entries in global load vector:'
        write(6,*)'  node  comp.     value'
        write(6,*)
        write(16,*) 
        write(16,*)'Non-zero entries in global load vector:'
        write(16,*)'  node  comp.     value'
        write(16,*)
        do 95 i=1,ntotv
          ai = aslod(i)
c...zero any tiny loads caused by rounding errors
          if (abs(ai).lt.1.e-5) then
            aslod(i) = 0.0
          else
            n = (i+1)/2
            m = i - 2*(n-1)
            write(6,650) n,m,ai
            write(16,650) n,m,ai
  650       format(i5,i5,5x,e10.3)
          endif
  95    continue
      endif
c
      write(6,*)
      write(6,*)'End of load data.'
      write(6,*)
      write(16,*)
      write(16,*)'End of load data.'
      write(16,*)
      return
   99 write(6,*)'ERROR: this input file has no loading data.'
      call prompt
      stop
      end
c      
c
      SUBROUTINE FRONT(MTOTV,MPOIN,MELEM,MFIXV,MNODE,MFRON,MSTIF,
     1NPOIN,NELEM,NSDIS,LNODS,LNODN,NDOF,NVAR,NFVAR,ASLOD,ASDIS,SPDIS,
     2VECRV,GSTIF,GLOAD,EQUAT,NACVA,
     3PIVAL,IRSOL,EQRHS,NPIVO,NAMEV,MBUFA)
C************************************************************************
C...solves (assembled ESTIF)*ASDIS = ASLOD  by frontal algorithm. For
C...symmetric matrices. Code based on code from FINEPACK package, by
c...kind permission of David Naylor. 
C************************************************************************
      real*8 eqr,spd,eqj,pivot,pivmin,pivmax,pival,eqpiv,SMALL
     1      ,vecrv(mfron),gstif(mstif),gload(mfron),equat(mfron,mbufa)
     2      ,eqrhs(mbufa),eqjp
      DIMENSION LNODS(MELEM,MNODE),LNODN(MELEM),NDOF(MPOIN)
     1  ,NFVAR(MFIXV),ASLOD(MTOTV),ASDIS(MTOTV),SPDIS(MFIXV)
     3  ,ESTIF(16,16),NPIVO(MBUFA),NAMEV(MBUFA)
     4  ,NACVA(MFRON),LOCEL(16),NDEST(16),NVAR(MPOIN)
C...VECRV...NDEST INTERNAL USE ONLY
C
C...FUNCTION STATEMENT
      NFUNC(I,J) = (J*J - J)/2 + I
c...this relates the (i,j)th matrix entry to the vector storage
      IF(NFUNC(MFRON,MFRON).GT.MSTIF) GO TO 500
C
      SMALL = 1.d-6
C...CHANGE SIGN OF LAST APPEARANCE OF EACH NODE
C...NEEDED FOR INITIAL SOLUTION (IRSOL=0) ONLY
C
      IF (IRSOL.EQ.0) THEN
        DO 40 NP=1,NPOIN
          KLAST = 0
          DO 30 L=1,NELEM
            NNODE = LNODN(L)
            DO 20 N=1,NNODE
              IF(LNODS(L,N).EQ.NP) THEN
                KLAST = L
                NLAST = N
              ENDIF
   20       CONTINUE
   30     CONTINUE
          IF(KLAST.NE.0) LNODS(KLAST,NLAST) = -NP
   40   CONTINUE
      ENDIF
C
C...ZERO ARRAYS
      DO 50 I=1,MSTIF
   50 GSTIF(I) = 0.0D0
      DO 60 I=1,MFRON
        GLOAD(I) = 0.0D0
        VECRV(I) = 0.0D0
   60 NACVA(I) = 0
      DO 70 I=1,MBUFA
        EQRHS(I) = 0.0D0
        DO 70 J=1,MFRON
          EQUAT(J,I) = 0.0D0
   70 CONTINUE
      PIVMAX = 0.0D0
      PIVMIN = 1.0D9
      NFMAX = 0
      NBUFA = 0
C
      REWIND (18)
      REWIND (19)
      REWIND (20)
C
C...NEXT SECTION SOLVES USING NEW STIFFNESS MATRIX
      IF (IRSOL.GT.1) GOTO 522
C...ENTER MAIN ASSEMBLY - REDUCTION LOOP
      NFRON = 1
      KELVA = 0
      DO 380 L=1,NELEM
        NNODE = LNODN(L)
C...NEXT BIT ONLY NEEDED FOR INITIAL SOLUTION
        IF (IRSOL.EQ.0) THEN
C...SET UP LOCEL TO RELATE ELEMENT VARIABLE POSITIONS TO OVERALL POSNS
          NEVAB = 0
          DO 170 N=1,NNODE
            NP = LNODS(L,N)
            IS = 1
            IF(NP.GT.0) GO TO 160
            IS = -1
            NP = -NP
  160       NDOFNP = NDOF(NP)
            NIK = NVAR(NP)
            DO 165 I=1,NDOFNP
              NEVAB = NEVAB + 1
  165       LOCEL(NEVAB) = (NIK+I)*IS
            LNODS(L,N) = NP
  170     CONTINUE
C...THE ROLE OF LNODS FOR LOCATING LAST NODE APPEARANCE HAS BEEN TRANS-
C...FERRED TO LOCEL. HENCEFORTH WE FORGET D.OF F. AT NODE AND CONSIDER
C...ONLY THE VARIABLE POSITION (IE EQN NO.)
C
C...LOOK FOR EXISTING DESTINATIONS
          KEVAB = 0
          DO 210 IE=1,NEVAB
            NIKNO = IABS(LOCEL(IE))
            KEXIS = 0
            DO 180 I=1,NFRON
              IF(NIKNO.NE.NACVA(I)) GO TO 180
              KEVAB = KEVAB+1
              KEXIS = 1
              NDEST(KEVAB) = I
  180       CONTINUE
            IF(KEXIS.EQ.1) GO TO 210
C...ALLOCATE VARIABLE A SPACE IN NACVA
            DO 190 I=1,MFRON
              IF(NACVA(I).NE.0) GO TO 190
              NACVA(I) = NIKNO
              KEVAB = KEVAB+1
              NDEST(KEVAB) = I
              GO TO 200
  190       CONTINUE
            GO TO 502
C...EXPAND FRONT WIDTH AS REQUIRED
  200       NF = NDEST(KEVAB)
            IF(NF.GT.NFRON) NFRON=NF
            IF(NFRON.GT.NFMAX) NFMAX = NFRON
  210     CONTINUE
C
C...CHECK MSTIF DIMENSION SUFFICIENT
          NSTIF = NFUNC(NFRON,NFRON)
          IF(NSTIF.GT.MSTIF) GO TO 503
C...WRITE ARRAYS TO FILE FOR FUTURE USE WITH MATRICES OF SAME STRUCTURE
          WRITE (20) LOCEL,NDEST,NACVA,NFRON,NEVAB
C
        ELSE
C...READ ARRAYS IF STRUCTURE ALREADY ESTABLISHED
          READ (20) LOCEL,NDEST,NACVA,NFRON,NEVAB
          IF (NFRON.GT.NFMAX) NFMAX = NFRON
        ENDIF
C
C
        READ (18) ESTIF
C
C...COMMENCE ASSEMBLY - FIRST LOADS
        DO 240 IE=1,NEVAB
          I = NDEST(IE)
          NIK = LOCEL(IE)
          NIKLE = -NIK
          IF (NIK.LT.0) GLOAD(I) = GLOAD(I) + ASLOD(NIKLE)
C...AND THEN ELEMENT STIFFNESSES - SKIP IF RESOLUTION
          DO 220 JE=1,IE
            J = NDEST(JE)
            NASH = NFUNC(I,J)
            NISH = NFUNC(J,I)
            STIF = ESTIF(IE,JE)
            IF (J.GE.I) THEN
              GSTIF(NASH) = GSTIF(NASH) + STIF
            ELSE
              GSTIF(NISH) = GSTIF(NISH) + STIF
            ENDIF
  220     CONTINUE
  240   CONTINUE
C
C...FIND OUT WHICH ELEMENT NODES CAN BE ELIMINATED
        DO 370 IE=1,NEVAB
          NIKNO = -LOCEL(IE)
          IF(NIKNO.LE.0) GO TO 370
C...VARIABLES TO BE ELIMINATED NOW IDENTIFIED - FIND THEIR POSITIONS
          DO 350 I=1,NFRON
            IF(NACVA(I).NE.NIKNO) GO TO 350
C
            NBUFA = NBUFA + 1
            IF(NBUFA.GT.MBUFA) THEN
              NBUFA = 1
              WRITE(19) EQUAT,EQRHS,NPIVO,NAMEV
            ENDIF
C
C...EXTRACT COEFS OF NEW EQN FOR ELIMINATION (FIRST SOLUTION ONLY)
            DO 250 J=1,NFMAX
              IF (I.LT.J) THEN
                NL = NFUNC(I,J)
              ELSE
                NL = NFUNC(J,I)
              ENDIF
              EQUAT(J,NBUFA) = GSTIF(NL)
              GSTIF(NL) = 0.0d0
  250       CONTINUE
C...AND THE CORRESPONDING R.H.S.
            EQRHS(NBUFA) = GLOAD(I)
            GLOAD(I) = 0.0d0
            KELVA = KELVA+1
C
            NAMEV(NBUFA) = NIKNO
            NPIVO(NBUFA) = I
            II = I
C
C...TAKE DIAGONAL COEF. OF VARIABLE TO BE ELIM. AS PIVOT. CHECK FOR SIZE
            PIVOT = EQUAT(II,NBUFA)
            IF (PIVOT.GT.PIVMAX) PIVMAX = PIVOT
            IF (PIVOT.LT.PIVMIN) PIVMIN = PIVOT
            IF(pivot.LT.100.d0*PIVAL) WRITE(6,650) real(pivot),NIKNO
            IF(ABS(PIVOT).LT.PIVAL) GO TO 501
            EQUAT(II,NBUFA) = 0.0d0
C
C...ENQUIRE IF VARIABLE NIKNO FREE OR PRESCRIBED
            DO 285 N=1,NSDIS
              IF(NFVAR(N).EQ.NIKNO) GO TO 287
  285       CONTINUE
            GO TO 300
C...NIKNO IS PRESCRIBED
  287       SPD = SPDIS(N)
            DO 290 J=1,NFRON
  290       GLOAD(J) = GLOAD(J) - SPD*EQUAT(J,NBUFA)
            GO TO 340
C
C...ELIMINATE FREE VARIABLE - FIRST DEAL WITH RHS THEN LHS COEFS.
  300       EQPIV = EQRHS(NBUFA)/PIVOT
            DO 330 J=1,NFRON
              EQJ = EQUAT(J,NBUFA)
              GLOAD(J) = GLOAD(J) - EQJ*EQPIV
              if (eqj.eq.0.0d0) goto 330
              NL = NFUNC(0,J)
              EQJP = EQJ/PIVOT
              DO 310 K=1,J
                KN = K + NL
  310         GSTIF(KN) = GSTIF(KN) - EQJP*EQUAT(K,NBUFA)
  330       CONTINUE
C
  340       EQUAT(II,NBUFA) = PIVOT
            NACVA(II) = 0
            GO TO 360
  350     CONTINUE
C...ELIMINATION COMPLETE
C
C...REDUCE FRONT WIDTH IF POSSIBLE
  360     IF(NACVA(NFRON).NE.0) GO TO 370
          NFRON = NFRON-1
          IF(NFRON.GT.0) GO TO 360
  370   CONTINUE
  380 CONTINUE
C...ELEMENT LOOP COMPLETED IN FORWARD ELIMINATION
C
      WRITE (19) EQUAT,EQRHS,NPIVO,NAMEV
      BACKSPACE (19)
C
C
C...START BACK SUBSTITUTION.
      DO 420 IE=1,KELVA
        IF(NBUFA.EQ.0) THEN
          BACKSPACE (19)
          READ (19) EQUAT,EQRHS,NPIVO,NAMEV
          BACKSPACE (19)
          NBUFA = MBUFA
        ENDIF
        I = NPIVO(NBUFA)
        NIKNO = NAMEV(NBUFA)
C
C...CHECK IF DISP. PRESCRIBED
        DO 400 N=1,NSDIS
          IF(NFVAR(N).EQ.NIKNO) GO TO 410
  400   CONTINUE
C...DISP. NOT PRESCRIBED - COMPUTE IT
        PIVOT = EQUAT(I,NBUFA)
        EQUAT(I,NBUFA) = 0.0d0
        EQR = EQRHS(NBUFA)
        DO 405 J=1,NFMAX
  405   EQR = EQR - VECRV(J)*EQUAT(J,NBUFA)
        VECRV(I) = EQR/PIVOT
        GO TO 415
C...DISP. IS PRESCRIBED - PUT REACTIONS (WHICH INCL APP. LOADS) IN ASLOD
  410   VECRV(I) = SPDIS(N)
        EQR = EQRHS(NBUFA)
        DO 412 J=1,NFMAX
  412   EQR = EQR - VECRV(J)*EQUAT(J,NBUFA)
        ASLOD(NIKNO) = -EQR
  415   ASDIS(NIKNO) = VECRV(I)
        NBUFA = NBUFA - 1
  420 CONTINUE
C
C...ADVISE WHAT MAX FRONT WIDTH WAS
      WRITE(6,600) NFMAX
      WRITE (6,610) PIVMIN,PIVMAX
      RETURN
C
C
C...THIS SECTION FOR RE-SOLUTIONS WITH PREVIOUSLY-DECOMPOSED MATRIX
  522 CONTINUE
C
C
      REWIND (21)
C
C...ENTER MAIN ASSEMBLY - REDUCTION LOOP
      NFRON = 1
      KELVA = 0
      nbufa = mbufa
      DO 580 L=1,NELEM
        READ (20) LOCEL,NDEST,NACVA,NFRON,NEVAB
        IF (NFRON.GT.NFMAX) NFMAX=NFRON
C
C...COMMENCE ASSEMBLY - FIRST LOADS
        DO 525 IE=1,NEVAB
          I = NDEST(IE)
          NIK = LOCEL(IE)
          NIKLE = -NIK
          IF (NIK.LT.0) GLOAD(I) = GLOAD(I) + ASLOD(NIKLE)
  525   CONTINUE
C
C...FIND OUT WHICH ELEMENT NODES CAN BE ELIMINATED
        DO 570 IE=1,NEVAB
          NIKNO = -LOCEL(IE)
          IF(NIKNO.LE.0) GO TO 570
C...VARIABLES TO BE ELIMINATED NOW IDENTIFIED - FIND THEIR POSITIONS
          DO 550 I=1,NFRON
            IF(NACVA(I).NE.NIKNO) GO TO 550
C
            NBUFA = NBUFA + 1
            IF(NBUFA.GT.MBUFA) THEN
              NBUFA = 1
              WRITE(21) EQRHS
              READ(19) EQUAT,EQRHS,NPIVO,NAMEV
            ENDIF
C
C...AND THE CORRESPONDING R.H.S.
            EQRHS(NBUFA) = GLOAD(I)
            GLOAD(I) = 0.0d0
            KELVA = KELVA+1
C
            NAMEV(NBUFA) = NIKNO
            NPIVO(NBUFA) = I
            II = I
C
C...TAKE DIAGONAL COEF. OF VARIABLE TO BE ELIM. AS PIVOT. CHECK FOR SIZE
            PIVOT = EQUAT(II,NBUFA)
            IF (PIVOT.GT.PIVMAX) PIVMAX = PIVOT
            IF (PIVOT.LT.PIVMIN) PIVMIN = PIVOT
            IF(pivot.LT.100.d0*PIVAL) WRITE(6,650) real(pivot),NIKNO
            IF(ABS(PIVOT).LT.PIVAL) GO TO 501
            EQUAT(II,NBUFA) = 0.0d0
C
C...ENQUIRE IF VARIABLE NIKNO FREE OR PRESCRIBED
            DO 585 N=1,NSDIS
              IF(NFVAR(N).EQ.NIKNO) GO TO 587
  585       CONTINUE
            GO TO 588
C...NIKNO IS PRESCRIBED
  587       SPD = SPDIS(N)
            DO 590 J=1,NFRON
  590       GLOAD(J) = GLOAD(J) - SPD*EQUAT(J,NBUFA)
            GO TO 540
C
C...ELIMINATE FREE VARIABLE - FIRST DEAL WITH RHS THEN LHS COEFS.
  588       EQPIV = EQRHS(NBUFA)/PIVOT
            DO 530 J=1,NFRON
              EQJ = EQUAT(J,NBUFA)
            GLOAD(J) = GLOAD(J) - EQJ*EQPIV
  530       CONTINUE
C
  540       EQUAT(II,NBUFA) = PIVOT
            NACVA(II) = 0
            GO TO 560
  550     CONTINUE
C...ELIMINATION COMPLETE
C
C...REDUCE FRONT WIDTH IF POSSIBLE
  560     IF(NACVA(NFRON).NE.0) GO TO 570
          NFRON = NFRON-1
          IF(NFRON.GT.0) GO TO 560
  570   CONTINUE
  580 CONTINUE
C...ELEMENT LOOP COMPLETED IN FORWARD ELIMINATION
C
      ENDFILE (21)
      BACKSPACE (21)
      BACKSPACE (19)
C
C
C...START BACK SUBSTITUTION.
      DO 520 IE=1,KELVA
        IF(NBUFA.EQ.0) THEN
          BACKSPACE (19)
          READ (19) EQUAT,EQRHS,NPIVO,NAMEV
          BACKSPACE (19)
          NBUFA = MBUFA
C...RESOLUTIONS ONLY
          BACKSPACE (21)
          READ (21) EQRHS
          BACKSPACE (21)
        ENDIF
        I = NPIVO(NBUFA)
        NIKNO = NAMEV(NBUFA)
C
C...CHECK IF DISP. PRESCRIBED
        DO 507 N=1,NSDIS
          IF(NFVAR(N).EQ.NIKNO) GO TO 510
  507   CONTINUE
C...DISP. NOT PRESCRIBED - COMPUTE IT
        PIVOT = EQUAT(I,NBUFA)
        EQUAT(I,NBUFA) = 0.0d0
        EQR = EQRHS(NBUFA)
        DO 505 J=1,NFMAX
  505   EQR = EQR - VECRV(J)*EQUAT(J,NBUFA)
        VECRV(I) = EQR/PIVOT
        GO TO 515
C...DISP. IS PRESCRIBED - PUT REACTIONS (WHICH INCL APP. LOADS) IN ASLOD
  510   VECRV(I) = SPDIS(N)
        EQR = EQRHS(NBUFA)
        DO 512 J=1,NFMAX
  512   EQR = EQR - VECRV(J)*EQUAT(J,NBUFA)
        ASLOD(NIKNO) = -EQR
  515   ASDIS(NIKNO) = VECRV(I)
        NBUFA = NBUFA - 1
  520 CONTINUE
C
C...ADVISE WHAT MAX FRONT WIDTH WAS
      WRITE(6,600) NFMAX
      RETURN
C
  600 FORMAT(/' FRONT COMPLETED. MAX FRONT WIDTH WAS',I4)
  610 FORMAT(1P,/' MIN. PIVOT =',G12.4,', MAX. PIVOT =',G12.4)
  650 FORMAT(1P,' WARNING IN FRONT, PIVOT =',G12.4,' FOR EQN',I4,
     1 ' IS DANGEROUS')
  700 FORMAT(/' STOPPED AT START OF FRONT.',
     1' MFRON TOO LARGE IN RELATION TO MSTIF.')
  701 FORMAT(/' STOPPED IN FRONT, PIVOT TOO SMALL FOR EQN',I4)
  702 FORMAT(/' STOPPED IN FRONT, NFRON =',I4,' EXCEEDS MFRON IN EL.'
     1,I4)
  703 FORMAT(/' STOPPED IN FRONT, MSTIF (=',I4,') EXCEEDED IN ELEMENT'
     1,I4,', NEEDS TO BE AT LEAST',I5)
C
C...ERROR MESSAGES
  500 WRITE(6,700)
      call prompt
      STOP
  501 WRITE(6,701) NIKNO
      call prompt
      STOP
  502 WRITE(6,702) NFRON,L
      call prompt
      STOP
  503 WRITE(6,703) MSTIF,L,NSTIF
      call prompt
      STOP
      END
c
c
      SUBROUTINE AFRONT(MTOTV,MPOIN,MELEM,MFIXV,MNODE,MFRON,MSTIF,
     1NPOIN,NELEM,NSDIS,LNODS,LNODN,NDOF,NVAR,NFVAR,ASLOD,ASDIS,SPDIS,
     2VECRV,GSTIF,GLOAD,EQUAT,NACVA,
     3PIVAL,IRSOL,EQRHS,NPIVO,NAMEV,MBUFA)
C************************************************************************
C...SOLVES FOR LOAD VECTOR ASLOD AND RETURNS WITH DISPLACEMENTS IN ASDIS
C...LOADS UNALTERED IN ASLOD EXCEPT FOR ADDITION OF REACTIONS.
C...ADAPTED from FRONT FOR NON-SYMMETRIC MATRICES by M Reed
C************************************************************************
      real*8 eqr,spd,pival,eqj,pivot,eqpiv,pivmin,pivmax,small
     1      ,vecrv(mfron),gstif(mstif),gload(mfron),equat(mfron,mbufa)
     2      ,eqrhs(mbufa),eqjp
      DIMENSION LNODS(MELEM,mnode),LNODN(MELEM),NdoF(MPOIN)
     1,NFVAR(MFIXV),ASLOD(MTOTV),ASDIS(MTOTV),SPDIS(MFIXV)
     3,ESTIF(16,16),NACVA(MFRON),LOCEL(16),NDEST(16)
     4,NAMEV(MBUFA),NPIVO(MBUFA),nvar(mpoin)
C
C...VECRV...NDEST INTERNAL USE ONLY
C
C...FUNCTION STATEMENT
      NFUNC(I,J) = (MAX(I,J)-1)**2 + J + DIM(J,I)
      IF(NFUNC(1,MFRON).GT.MSTIF) GO TO 500
C
C...IRSOL SHOULD BE 0 OR 1 - NO RE-SOLUTIONS WITH PREVIOUS MATRIX
      IF (IRSOL.GT.1) WRITE (6,705) IRSOL
C
      SMALL = 1.0d-5
      IF (IRSOL.EQ.0) THEN
C...CHANGE SIGN OF LAST APPEARANCE OF EACH NODE
      DO 40 NP=1,NPOIN
      KLAST = 0
      DO 30 L=1,NELEM
      NNODE = LNODN(L)
      DO 20 N=1,NNODE
      IF(LNODS(l,n).NE.NP) GO TO 20
      KLAST = L
      NLAST = N
   20 CONTINUE
   30 CONTINUE
      IF(KLAST.NE.0) LNODS(kLAST,nLAST) = -NP
   40 CONTINUE
      ENDIF
C
C...ZERO ARRAYS
C
      DO 50 I=1,MSTIF
   50 GSTIF(I) = 0.0d0
      DO 60 I=1,MFRON
      GLOAD(I) = 0.0d0
      VECRV(I) = 0.0d0
   60 NACVA(I) = 0
      PIVMAX = 0.0d0
      PIVMIN = 0.0d0
      NFMAX = 0
C
      DO 70 I=1,MBUFA
      EQRHS(I) = 0.0d0
      DO 70 J=1,MFRON
      EQUAT(J,I) = 0.0d0
   70 CONTINUE
      NBUFA = 0
C
      REWIND (18)
      REWIND (19)
      REWIND (20)
C
C...ENTER MAIN ASSEMBLY - REDUCTION LOOP
      NFRON = 1
      KELVA = 0
      DO 380 L=1,NELEM
      NNODE = LNODN(L)
      IF (IRSOL.EQ.0) THEN
C...SET UP LOCEL TO RELATE ELEMENT VARIABLE POSITIONS TO OVERALL POSNS
      NEVAB = 0
      DO 170 N=1,NNODE
      NP = LNODS(l,n)
      IS = 1
      IF(NP.GT.0) GO TO 160
      IS = -1
      NP = -NP
  160 NDOFNp = Ndof(NP)
      NIK = Nvar(NP)
      DO 165 I=1,NDOFNp
      NEVAB = NEVAB + 1
  165 LOCEL(NEVAB) = (NIK+I)*IS
      LNODS(l,n) = NP
  170 CONTINUE
C
C...LOOK FOR EXISTING DESTINATIONS
      KEVAB = 0
      DO 210 IE=1,NEVAB
      NIKNO = IABS(LOCEL(IE))
      KEXIS = 0
      DO 180 I=1,NFRON
      IF(NIKNO.NE.NACVA(I)) GO TO 180
      KEVAB = KEVAB+1
      KEXIS = 1
      NDEST(KEVAB) = I
  180 CONTINUE
      IF(KEXIS.EQ.1) GO TO 210
C...ALLOCATE VARIABLE A SPACE IN NACVA
      DO 190 I=1,MFRON
      IF(NACVA(I).NE.0) GO TO 190
      NACVA(I) = NIKNO
      KEVAB = KEVAB+1
      NDEST(KEVAB) = I
      GO TO 200
  190 CONTINUE
      GO TO 502
C...EXPAND FRONT WIDTH AS REQUIRED
  200 NF = NDEST(KEVAB)
      IF(NF.GT.NFRON) NFRON=NF
      IF(NFRON.GT.NFMAX) NFMAX = NFRON
  210 CONTINUE
C
      WRITE (20) LOCEL,NDEST,NACVA,NFRON,NEVAB
      ELSE
      READ (20) LOCEL,NDEST,NACVA,NFRON,NEVAB
      IF (NFRON.GT.NFMAX) NFMAX=NFRON
      ENDIF
C
C...CHECK MSTIF DIMENSION SUFFICIENT
C
      NSTIF = NFUNC(1,NFRON)
      IF(NSTIF.GT.MSTIF) GO TO 503
C
      READ (18) ESTIF
C
C...COMMENCE ASSEMBLY - FIRST LOADS
C
  215 DO 240 IE=1,NEVAB
      I = NDEST(IE)
      NIK = LOCEL(IE)
      NIKLE = -NIK
      IF (NIK.LT.0) GLOAD(I) = GLOAD(I) + ASLOD(NIKLE)
C...AND THEN ELEMENT STIFFNESSES - SKIP IF RESOLUTION
      DO 220 JE=1,NEVAB
      J = NDEST(JE)
      NASH = NFUNC(I,J)
      STIF = ESTIF(IE,JE)
      GSTIF(NASH) = GSTIF(NASH) + STIF
  220 CONTINUE
  240 CONTINUE
C
C...FIND OUT WHICH ELEMENT NODES CAN BE ELIMINATED
C
      DO 370 IE=1,NEVAB
      NIKNO = -LOCEL(IE)
      IF(NIKNO.LE.0) GO TO 370
C...VARIABLES TO BE ELIMINATED NOW IDENTIFIED - FIND THEIR POSITIONS
      DO 350 I=1,NFRON
      IF(NACVA(I).NE.NIKNO) GO TO 350
      NBUFA = NBUFA + 1
      IF(NBUFA.GT.MBUFA) THEN
	NBUFA = 1
	  WRITE(19) EQUAT,EQRHS,NPIVO,NAMEV
      ENDIF
C
C...EXTRACT COEFS OF NEW EQN FOR ELIMINATION (FIRST SOLUTION ONLY)
      DO 250 J=1,NFMAX
      NL = NFUNC(I,J)
      EQUAT(J,NBUFA) = GSTIF(NL)
      GSTIF(NL) = 0.0d0
  250 CONTINUE
C...AND THE CORRESPONDING R.H.S.
  260 EQRHS(NBUFA) = GLOAD(I)
      GLOAD(I) = 0.0d0
      KELVA = KELVA+1
C
      NAMEV(NBUFA) = NIKNO
      NPIVO(NBUFA) = I
      II = I
C
C...TAKE DIAGONAL COEF. OF VARIABLE TO BE ELIM. AS PIVOT.
C   CHECK FOR SIZE
C
  280 PIVOT = EQUAT(II,NBUFA)
      IF (PIVOT.GT.PIVMAX) PIVMAX = PIVOT
      IF (PIVOT.LT.PIVMIN.OR.PIVMIN.EQ.0.0d0) PIVMIN = PIVOT
      IF(PIVOT.LT.100.d0*PIVAL) WRITE(6,650) PIVOT,NIKNO
      IF(ABS(PIVOT).LT.PIVAL) GO TO 501
      EQUAT(II,NBUFA) = 0.0d0
C
C...ENQUIRE IF VARIABLE NIKNO FREE OR PRESCRIBED
C
      DO 285 N=1,NSDIS
      IF(NFVAR(N).EQ.NIKNO) GO TO 287
  285 CONTINUE
      GO TO 300
C...NIKNO IS PRESCRIBED
  287 SPD = SPDIS(N)
      DO 290 J=1,NFRON
      JI = NFUNC(J,II)
      EJI = GSTIF(JI)
      GSTIF(JI) = 0.0d0
  290 GLOAD(J) = GLOAD(J) - SPD*EJI
      GO TO 340
C
C...ELIMINATE FREE VARIABLE - FIRST DEAL WITH RHS THEN LHS COEFS.
C
  300 EQR = EQRHS(NBUFA)
      DO 330 J=1,NFRON
      JI = NFUNC(J,II)
      EQJ = GSTIF(JI)
      GSTIF(JI) = 0.0d0
      if (eqj.eq.0.d0) goto 330
      EQPIV = EQJ/PIVOT
      GLOAD(J) = GLOAD(J) - EQPIV*EQR
      DO 310 K=1,NFRON
      KN = NFUNC(J,K)
  310 GSTIF(KN) = GSTIF(KN) - EQPIV*EQUAT(K,NBUFA)
  330 CONTINUE
C
  340 EQUAT(II,NBUFA) = PIVOT
      NACVA(II) = 0
      DO 345 J=NFRON,NFMAX
      JI = NFUNC(J,II)
      GSTIF(JI) = 0.0d0
  345 CONTINUE
      GO TO 360
  350 CONTINUE
C...ELIMINATION COMPLETE
C
C...REDUCE FRONT WIDTH IF POSSIBLE
  360 IF(NACVA(NFRON).NE.0) GO TO 370
      NFRON = NFRON-1
      IF(NFRON.GT.0) GO TO 360
  370 CONTINUE
  380 CONTINUE
C
C...ELEMENT LOOP COMPLETED IN FORWARD ELIMINATION
C
      WRITE (19) EQUAT,EQRHS,NPIVO,NAMEV
      BACKSPACE (19)
C
C...START BACK SUBSTITUTION.
C
      DO 420 IE=1,KELVA
      IF(NBUFA.EQ.0) THEN
      BACKSPACE (19)
      READ (19) EQUAT,EQRHS,NPIVO,NAMEV
      BACKSPACE (19)
      NBUFA = MBUFA
      ENDIF
      I = NPIVO(NBUFA)
      NIKNO = NAMEV(NBUFA)
C
C...CHECK IF DISP. PRESCRIBED
C
      DO 400 N=1,NSDIS
      IF(NFVAR(N).EQ.NIKNO) GO TO 410
  400 CONTINUE
C...DISP. NOT PRESCRIBED - COMPUTE IT
      PIVOT = EQUAT(I,NBUFA)
      EQUAT(I,NBUFA) = 0.0d0
      EQR = EQRHS(NBUFA)
      DO 405 J=1,NFMAX
  405 EQR = EQR - VECRV(J)*EQUAT(J,NBUFA)
      VECRV(I) = EQR/PIVOT
      GO TO 415
C...DISP. IS PRESCRIBED - PUT REACTIONS (WHICH INCL APP. LOADS) IN ASLOD
  410 VECRV(I) = SPDIS(N)
      EQR = EQRHS(NBUFA)
      DO 412 J=1,NFMAX
  412 EQR = EQR - VECRV(J)*EQUAT(J,NBUFA)
      ASLOD(NIKNO) = -EQR
  415 ASDIS(NIKNO) = VECRV(I)
      NBUFA = NBUFA - 1
  420 CONTINUE
C
      WRITE (6,600) NFMAX
      WRITE (6,610) PIVMIN,PIVMAX
C
      RETURN
C
  600 FORMAT(/' AFRONT COMPLETED. MAX FRONTWIDTH WAS ',I4)
  610 FORMAT(1P,/' MIN. PIVOT =',G12.4,', MAX. PIVOT =',G12.4)
  650 FORMAT(1P, ' WARNING IN AFRONT, PIVOT =',G12.4, 8H FOR EQN,I4,
     1 ' IS DANGEROUS')
  700 FORMAT(/' STOPPED AT START OF AFRONT. MFRON TOO LARGE IN RELATION
     1 TO MSTIF. ALTER THEM IN MASTER.')
  701 FORMAT(/' STOPPED IN AFRONT, PIVOT TOO SMALL FOR EQN',I4)
  702 FORMAT(/' STOPPED IN AFRONT, NFRON =',I4,' EXCEEDS MFRON IN EL.
     1',I4)
  703 FORMAT(/' STOPPED IN AFRONT, MSTIF (=',I4,') EXCEEDED IN ELEMENT'
     1,I4, 22H, NEEDS TO BE AT LEAST,I5)
  705 FORMAT(/' WARNING IN AFRONT:IRSOL =',I3,' - NO RE-SOLUTIONS!')
C
C...ERROR MESSAGES
  500 WRITE(6,700)
              call prompt
      STOP
  501 WRITE(6,701) NIKNO
              call prompt
      STOP
  502 WRITE(6,702) NFRON,L
              call prompt
      STOP
  503 WRITE(6,703) MSTIF,L,NSTIF
              call prompt
      STOP
      END
c   
c
      subroutine result(melem,mpoin,mnode,msets,mcomp,mtotv,
     1                  nelem,npoin,lnods,lnodn,coord,nkode,
     3                  lpros,props,aslod,asdis,strsg,mohrg)
c
c...output the results to file
c
      dimension aslod(mtotv),props(msets,mcomp),
     1         lnods(melem,mnode),coord(mpoin,2),endis(16),encod(2,8),
     2        shape(8),deriv(2,8),cartd(2,8),dmatx(3,3),  
     3       bmatx(3,16),gploc(2,4),strsg(3,4,melem),mohrg(4,melem),
     4      gpwts(4),gpsig(3),gpeps(3),gpcod(2),
     5      nkode(mpoin),asdis(mtotv),lnodn(melem),lpros(melem)
c...output displacements and reactions
c...output to screen (unit 6):
      write(6,*)
      write(6,*)'RESULTS'
      write(6,*)'======='
      write(6,*)
      write(6,610)
  610 format(//' nodal reactions and displacements'
     1/28('-')//3x,'node',
     212x,'x-reaction',5x,'y-reaction',9x,'x-displt',7x,'y-displt')
c...output to results file (unit 16):
      write(16,*)
      write(16,*)'RESULTS'
      write(16,*)'======='
      write(16,*)
      write(16,610)
c...and output to end of data file, for postprocessor:
      write(17,*)'  node  x-displt  y-displt'
      do 20 n=1,npoin  
        i = 2*(n - 1)
        dispx = asdis(i+1)
        dispy = asdis(i+2)
        write(17,'(i5,2e10.3)') n,dispx,dispy
        if (mod(n,25).eq.0) call prompt
        write(6,615) n,aslod(i+1),aslod(i+2),dispx,dispy 
        write(16,615) n,aslod(i+1),aslod(i+2),dispx,dispy 
  615   format(3x,i3,11x,e10.3,5x,e10.3,10x,e10.3,5x,e10.3)    
   20 continue
c
      call prompt
c...element Gauss pt stressess
  620 format(//'Element Gauss pt. stresses:'/27('-')/)
      write(6,620)
      write(16,620)
      write(17,*)
     1' elt gp    x-coord   y-coord   sigma-x   sigma-y   tau-xy'
      do 100 l=1,nelem
        if (mod(l,3).eq.0) call prompt
        write(6,630) l
        write(16,630) l
  630   format(//' element',i5/1x,12('-')/)
        nnode = lnodn(l)
c...store nodal coords for elt in encod
        do 25 n=1,nnode
          nn = lnods(l,n)
          encod(1,n) = coord(nn,1)
          encod(2,n) = coord(nn,2)
   25   continue
c...find Gauss-pt stresses
        write(6,640) (i,i=1,3)
        write(16,640) (i,i=1,3)
  640   format(/' gauss-pt stresses and strains'/6x,'x-coord',8x,
     1  'y-coord',10x,3('stress (',i1,')',5x))
        call gauss(gploc,gpwts,ngaus)
c...loop over each Gauss point
        do 80 ij=1,ngaus
          s = gploc(1,ij)
          t = gploc(2,ij)
          call sfr(s,t,shape,deriv)
c...get Gauss point coords in gpcod
          do 27 m=1,2
            sum = 0.0
            do 26 n=1,8
   26       sum = sum + encod(m,n)*shape(n)
            gpcod(m) = sum
   27     continue
          do 40 i=1,3
            gpsig(i) = strsg(i,ij,l)
   40     continue
cc..print the plasticity code for this Gauss point
          mkode = mohrg(ij,l)
c...only stresses printed; modify this to print strains too
          write(17,'(i5,i2,i3,5e10.3)') 
     1       l,ij,mkode,(gpcod(i),i=1,2),(gpsig(i),i=1,3)
          write(6,650) (gpcod(i),i=1,2),(gpsig(i),i=1,3)
          write(16,650) (gpcod(i),i=1,2),(gpsig(i),i=1,3)
          if (mkode.ne.0) then
            write(6,*)'Gauss-point yielded: kode =',mkode
            write(16,*)'Gauss-point yielded: kode =',mkode
          endif
  650     format(3x,2(e10.3,5x),5x,3(e10.3,5x))
   80   continue
c...end of Gauss loop
  100 continue
c...end of element loop
c
      write(6,*)
      write(6,*)'END OF RESULTS'
      write(6,*)
      write(16,*)
      write(16,*)'END OF RESULTS'
      write(16,*)
      return
      end
c
c
c...subroutines for STIFF     
c
c
      subroutine gauss(gploc,gpwts,ngaus)
      dimension gploc(2,4),gpwts(4)
c...set up Gauss pts & weights for 2x2 integration
      ngaus = 4
      do 5 i=1,4
        gpwts(i) = 1.0
    5 continue
      r13 = 1.0/sqrt(3.0)
      gploc(1,1) = -r13
      gploc(1,2) = -r13
      gploc(1,3) = r13
      gploc(1,4) = r13
      gploc(2,1) = -r13
      gploc(2,2) = r13
      gploc(2,3) = -r13
      gploc(2,4) = r13
      return
      end
c
c
      subroutine dmat(ym,pr,dmatx)
      dimension dmatx(3,3)
c...ym,pr are Young's mod., Poisson's ratio
c...form the D matrix for plane strain
      g = 0.5*ym/(1.0 + pr)
      yl = g*pr/(0.5 - pr)
      g2yl = 2.*g + yl
      dmatx(1,1) = g2yl
      dmatx(1,2) = yl
      dmatx(1,3) = 0.0
      dmatx(2,1) = yl
      dmatx(2,2) = g2yl
      dmatx(2,3) = 0.0
      dmatx(3,1) = 0.0
      dmatx(3,2) = 0.0
      dmatx(3,3) = g
      return
      end
c
c
      subroutine bmat(cartd,bmatx)
      dimension cartd(2,8),bmatx(3,16)
c...form the B matrix
      do 5 j=1,16
      do 5 i=1,3
        bmatx(i,j) = 0.0
    5 continue
      j = 0
      do 10 n=1,8
        i = j + 1
        j = i + 1
        bmatx(1,i) = -cartd(1,n)
        bmatx(2,j) = -cartd(2,n)
        bmatx(3,i) = -cartd(2,n)
        bmatx(3,j) = -cartd(1,n)
c...negative sign for 'compression +ve' sign convention
   10 continue
      return
      end
c
c
      subroutine sfr(s,t,shape,deriv)
      dimension shape(8),deriv(2,8)
c...evaluate the shape functions and 1st derivs at local coords s,t
      s2 = 2.0*s
      t2 = 2.0*t
      ss = s*s
      tt = t*t
      st = s*t
      sst = ss*t
      stt = s*tt
      st2 = 2.0*st
      shape(1) = (-1.0 + st + ss + tt - sst - stt)*0.25
      shape(2) = (1.0 - t - ss + sst)*0.5
      shape(3) = (-1.0 - st + ss + tt - sst + stt)*0.25
      shape(4) = (1.0 + s - tt - stt)*0.5
      shape(5) = (-1.0 + st + ss + tt + sst + stt)*0.25
      shape(6) = (1.0 + t - ss - sst)*0.5
      shape(7) = (-1.0 - st + ss + tt + sst - stt)*0.25
      shape(8) = (1.0 - s - tt + stt)*0.5
c     
      deriv(1,1) = (t + s2 - st2 - tt)*0.25
      deriv(1,2) = -s + st
      deriv(1,3) = (-t + s2 - st2 + tt)*0.25
      deriv(1,4) = (1.0 - tt)*0.5
      deriv(1,5) = (t + s2 + st2 + tt)*0.25
      deriv(1,6) = -s - st
      deriv(1,7) = (-t + s2 + st2 - tt)*0.25
      deriv(1,8) = (-1.0 + tt)*0.5
      deriv(2,1) = (s + t2 - ss - st2)*0.25
      deriv(2,2) = (-1.0 + ss)*0.5
      deriv(2,3) = (-s + t2 - ss + st2)*0.25
      deriv(2,4) = -t - st             
      deriv(2,5) = (s + t2 + ss + st2)*0.25
      deriv(2,6) = (1.0 - ss)*0.5
      deriv(2,7) = (-s + t2 + ss - st2)*0.25
      deriv(2,8) = -t + st
      return
      end
c
c     
      subroutine jacob(encod,shape,deriv,cartd,djacb)
      dimension shape(8),deriv(2,8),cartd(2,8),encod(2,8),xjacm(2,2),
     1         xjaci(2,2)
c...evaluate the Jacobian matrix, and its determinant
      do 5 j=1,2
      do 5 i=1,2
        sum = 0.0
        do 4 n=1,8
    4   sum = sum + deriv(i,n)*encod(j,n)
        xjacm(i,j) = sum
    5 continue
c...xjacm is the Jacobian matrix. find det & inverse by Cramer        
      djacb = xjacm(1,1)*xjacm(2,2) - xjacm(1,2)*xjacm(2,1)
      if (djacb.lt.1.0e-12) then
        write(6,*)'ERROR: determinant of Jacobian is -ve or zero'
        write(6,*)'Element is inside out, or nodes defined clockwise'
              call prompt
        stop
      endif
      xjaci(1,1) = xjacm(2,2)/djacb
      xjaci(2,2) = xjacm(1,1)/djacb
      xjaci(1,2) = -xjacm(1,2)/djacb
      xjaci(2,1) = -xjacm(2,1)/djacb
c...get Cartesian derivs of shape functions using J inverse
      do 50 i=1,2
      do 50 n=1,8
        sum = 0.0
        do 49 j=1,2
   49   sum = sum + xjaci(i,j)*deriv(j,n)
        cartd(i,n) = sum
   50 continue
      return
      end
c
c
c
c
cc..SUBROUTINES FOR PLASTICITY
c
c
      subroutine stress(melem,mpoin,mnode,msets,mcomp,mtotv,
     1                 nelem,npoin,lnods,lnodn,coord,lpros,props,
     2              asdis,strsg,strsg1,mohrg,stoler,vivel,dtime,theta)
cc..calculates stress increments from asdis, updates strsg into strsg1
cc..and updates yield code in mohrg
c
      dimension props(msets,mcomp),lnods(melem,mnode),encod(2,8),
     1         bmatx(3,16),dmatx(3,3),asdis(mtotv),
     2        gpwts(4),lpros(melem),lnodn(melem),coord(mpoin,2),
     3       shape(8),gploc(2,4),gpcod(2),deriv(2,8),cartd(2,8)
cc..extra arrays in plasticity
     4  ,strsg1(3,4,melem),gpsig(3),gpeps(3),bvect(3),endis(16)
     5  ,demat(3,3),mohrg(4,melem),strsg(3,4,melem),gpsig0(3)
cv...extra arrays in viscoplasticity
     7  ,dsigma(3),dsig1(3)
     8  ,gpviv(3),gpviv0(3),vivel(3,4,melem),evec1(3),hmatx(3,3)
c
c...start loop over all elements
      do 80 l=1,nelem
c
c...extract Young's modulus & Poisson's ratio from props      
      lset = lpros(l)
      ym = props(lset,1)
      pr = props(lset,2)
cc..and the Mohr-Coulomb plasticity parameters k and sigma_c
      triax = props(lset,3)
      sigc = props(lset,4)
cv...and gamma
      gamma = props(lset,5)
      dilat = props(lset,6)
c
c...put nodal coords in encod
cc..and nodal displacements in endis
      nnode = lnodn(l)
      do 15 n=1,nnode
      k = 2*(n-1)
      nn = lnods(l,n)
      kk = 2*(nn-1)
      endis(k+1) = asdis(kk+1)
      endis(k+2) = asdis(kk+2)
      encod(1,n) = coord(nn,1)
      encod(2,n) = coord(nn,2)
   15 continue
c
c...get elastic D matrix
      call dmat(ym,pr,demat)
c
c...put local coords of 2x2 guass-pts in gploc
      call gauss(gploc,gpwts,ngaus)
c
c...start loop through gauss-pts
      do 75 ij=1,ngaus
c...put local coords of gauss-pt in s,t
      s = gploc(1,ij)
      t = gploc(2,ij)
c...get shape fns (shape) & derivs (deriv) in local coords
      call sfr(s,t,shape,deriv)
c...get xy-coords of gauss-pt into gpcod, using shape fn interpolation
      do 25 m=1,2
      sum = 0.0
      do 24 n=1,8
   24 sum = sum + encod(m,n)*shape(n)
   25 gpcod(m) = sum
c...get xy-derivs (cartd) & derterminant of Jacobian (djacb)
      call jacob(encod,shape,deriv,cartd,djacb)
c...get B matrix 
      call bmat(cartd,bmatx)
c
cc..form (B*displt_inc) in gpeps
      do 33 i=1,3
        gi = 0.0
        do 32 j=1,16
          gi = gi + bmatx(i,j)*endis(j)
   32   continue
        gpeps(i) = gi
   33 continue
c
cv...if Gauss point is elastic, no iteration needed
cv...stress inc. = D * strain inc.
      if (mohrg(ij,l).eq.0) then
        do 37 i=1,3
          si = 0.0
          do 36 j=1,3
            si = si + demat(i,j)*gpeps(j)
   36     continue
cv...store updated stress at G.Pt. in gpsig
          gpsig(i) = strsg(i,ij,l) + si
   37   continue
        go to 70
      endif
c
cv...store previous stress level \sigma_n in gpsig0
cv...store old v.p. strain inc. \epsilon^{dot}_n in gpviv0
cv...put other constant term with gpeps, for convenience
      do 38 i=1,3
        gpsig0(i) = strsg(i,ij,l)
        gpviv0(i) = vivel(i,ij,l)
        gpeps(i) = gpeps(i) - dtime*gpviv0(i)
   38 continue
c
cv...get the d^hat matrix at start of increment
      call dmatvp(triax,dilat,sigc,gamma,theta,dtime,
     1            gpsig0,demat,dmatx)
c        
cc..first guess for new stress
cv...note this is different in viscoplasticity:
cv...dmatx used instead of demat, and gpeps has extra term (above)
      do 35 i=1,3
        si = 0.0
        do 34 j=1,3
          si = si + dmatx(i,j)*gpeps(j)
   34   continue
cv...in iteration, dsigma will hold the old value of stress inc,
cv...              dsig1   "     "   "  new  "    "   "      ",
cv...              gpsig   "     "   "  old  "    "  stress, and
cv...              gpsig0 holds stress at start of inc, sigma_n 
        dsigma(i) = si
        gpsig(i) = gpsig0(i) + si
   35 continue
c
cc..start local stress iteration
c
      kount = 0
   40 continue
      kount = kount + 1
cv..evaluate F, b at current stress, and use flow rule 
cv...to get pl. strain vel. in gpviv
      call ycrit(triax,sigc,gpsig,fvalu)
      call flowv(dilat,gpsig,bvect)
cv...assume Phi(F) = F
      phifn = fvalu
      do 41 i=1,3
        gpviv(i) = gamma*phifn*bvect(i)
   41 continue
c
cv...now get D and H matrices for this stress level
      call dmatvp(triax,dilat,sigc,gamma,theta,dtime,
     1            gpsig,demat,dmatx)
      call hmat(triax,dilat,sigc,gamma,gpsig,hmatx)
c
cv...now construct l.h.s. of iteration, into dsig1
cv...evec1 is a temporary storage vector
      do 43 i=1,3
        si = 0.0
        do 42 j=1,3
          si = si + hmatx(i,j)*dsigma(j)
   42   continue
        evec1(i) = si + gpviv0(i) -gpviv(i)
   43 continue
c
      do 44 i=1,3
        evec1(i) = gpeps(i) + theta*dtime*evec1(i)
   44 continue
c
      do 47 i=1,3
        si = 0.0
        do 45 j=1,3
          si = si + dmatx(i,j)*evec1(j)
   45   continue
        dsig1(i) = si
   47 continue
c
cv...test for convergence by comparing dsig1 with dsigma
cv...also put new stress values into dsigma and gpsig 
      do 49 i=1,3
        evec1(i) = dsig1(i) - dsigma(i)
        dsigma(i) = dsig1(i)
        gpsig(i) = gpsig0(i) + dsigma(i)
   49 continue
      call norm1(3,3,evec1,dsnorm)
c
      if (dsnorm.gt.stoler) goto 40
c
cc..update stress
   70   do 72 i=1,3
          strsg1(i,ij,l) = gpsig(i)
   72   continue
c
cc..end of stress iteration for this Gauss point
   75 continue
c
cc..end of element loop
   80 continue
c
      return
      end
c
c
c
        subroutine residu(melem,mpoin,mnode,mtotv,nelem,npoin
     1       ,ntotv,lnods,lnodn,coord,nkode,accld,strsg,resld)
c
      dimension lnods(melem,mnode),encod(2,8),nkode(mpoin),
     1         bmatx(3,16),gpwts(4),lnodn(melem),
     2         coord(mpoin,2),strsg(3,4,melem),resld(mtotv),
     3       shape(8),gploc(2,4),deriv(2,8),cartd(2,8)
     4  ,gpsig(3),accld(mtotv),enres(16)
c
cc..calculate residual (out-of-balance) force vector in resld
c
      do 5 i=1,ntotv
        resld(i) = 0.0
    5 continue
c
c...start loop over all elements
      do 80 l=1,nelem
c
c...put nodal coords in encod
      nnode = lnodn(l)
      do 15 n=1,nnode
      nn = lnods(l,n)
      encod(1,n) = coord(nn,1)
      encod(2,n) = coord(nn,2)
   15 continue
      thick = 1.0
c
c...put local coords of 2x2 gauss-pts in gploc
      call gauss(gploc,gpwts,ngaus)
c
c...collect element contribution to residual in enres
      do 20 i=1,16
        enres(i) = 0.0
   20 continue
c
c...start loop through gauss-pts
      do 75 ij=1,ngaus
c...put local coords of gauss-pt in s,t
      s = gploc(1,ij)
      t = gploc(2,ij)
c...get shape fns (shape) & derivs (deriv) in local coords
      call sfr(s,t,shape,deriv)
c...get xy-derivs (cartd) & derterminant of Jacobian (djacb)
      call jacob(encod,shape,deriv,cartd,djacb)
c...get B matrix 
      call bmat(cartd,bmatx)
c...store Gauss point stresses in gpsig
      do 25 i=1,3
        gpsig(i) = strsg(i,ij,l)
   25 continue
c
cc..add GP contribution to integral(B^T * sigma) to enres
      dvolu = thick*djacb*gpwts(ij)
      do 33 i=1,16
        gi = 0.0
        do 32 j=1,3
          gi = gi + bmatx(j,i)*gpsig(j)
   32   continue
        enres(i) = enres(i) + gi*dvolu
   33 continue
c
cc..end of Gauss point loop
   75 continue
c
cc..now subtract enres into the global residual vector
      do 77 n=1,nnode
        nn = lnods(l,n)
        ni = 2*(n-1)
        nni = 2*(nn-1)
        resld(nni+1) = resld(nni+1) - enres(ni+1)
        resld(nni+2) = resld(nni+2) - enres(ni+2)
   77 continue
c
cc..end of element loop
   80 continue
c
c...finally, add the current external applied loads
      do 85 i=1,ntotv
        resld(i) = resld(i) + accld(i)
   85 continue
c...and zero the componets corresponding to fixed d.o.f.s
      do 90 np=1,npoin
        kode = nkode(np)
        npi = 2*(np-1)
c...separate kode out into two digits
        kx = kode/10
        ky = kode - 10*kx
        if (kx.eq.1) resld(npi+1) = 0.0
        if (ky.eq.1) resld(npi+2) = 0.0
   90 continue
c
      return
      end
c
c
c
      subroutine ycrit(triax,sigc,gpsig,fvalu)
      dimension gpsig(3)
cc..test stresses for plastic yield
c
cc..get principal stresses
      call princ(gpsig,sig1,sig3)
c
cc..Mohr-Coulomb yield function
      fvalu = sig1 - triax*sig3 - sigc
c
      return
      end
c
c
      subroutine princ(gpsig,sig1,sig3)
      dimension gpsig(3)
cc..form principal stresses from gpsig
c
      sx = gpsig(1)
      sy = gpsig(2)
      tau = gpsig(3)
      smean = 0.5*(sx + sy)
      sdev = 0.5*(sx - sy)
      sroot = sqrt(sdev*sdev + tau*tau)
      sig1 = smean + sroot
      sig3 = smean - sroot
c
      return
      end
c
c
c
      subroutine flowv(triax,gpsig,avect)
      dimension gpsig(3),avect(3),s1grad(3),s3grad(3)
cc..form flow vector (Mohr-Coulomb)
c
      sx = gpsig(1)
      sy = gpsig(2)
      tau = gpsig(3)
      call princ(gpsig,s1,s3)
      p = 0.5*(1 + (sx-sy)/(s1-s3))
      r = 2.0*tau/(s1-s3)
      s1grad(1) = p
      s1grad(2) = 1. - p
      s1grad(3) = r
      s3grad(1) = 1. - p
      s3grad(2) = p
      s3grad(3) = -r
      do 5 i=1,3
        avect(i) = s1grad(i) - triax*s3grad(i)
    5 continue
c
      return
      end
c
c
      subroutine flowm(dilat,gpsig,dbds)
      dimension gpsig(3),dbds(3,3),vec(3)
c
cc..forms Jacobian of flow vector in dbds
      sx = gpsig(1)
      sy = gpsig(2)
      tau = gpsig(3)
      sd = 0.5*(sx - sy)
      sdt2 = sd*sd + tau*tau
      sroot = sqrt(sdt2)
      cc = (1.+dilat)*0.25/(sdt2*sroot)
      vec(1) = -tau
      vec(2) = tau
      vec(3) = 2.*sd
      do 5 i=1,3
        do 5 j=1,3
          dbds(i,j) = cc*vec(i)*vec(j)
    5 continue
c
      return
      end
c
c
      subroutine invert(n,zmat,zinv)
      dimension zmat(n,n),zinv(n,n),zz(6,6)
c
cc..inverts zmat to get zinv
      do 10 i=1,3
        do 5 j=1,3
          zz(i,j) = zmat(i,j)
          zinv(i,j) = 0.0
    5   continue
        zinv(i,i) = 1.0
   10 continue
c
      do 35 i=1,n
        d = zz(i,i)
        k = i + 1
        if (k.gt.n) goto 40
        do 20 j=k,n
          c = zz(j,i)/d
          do 20 l=1,n
            zz(j,l) = zz(j,l) - zz(i,l)*c
            zinv(j,l) = zinv(j,l) - zinv(i,l)*c
   20   continue
   35 continue
   40 continue
c
      d = zz(n,n)
      do 42 j=1,n
        zinv(n,j) = zinv(n,j)/d
   42 continue
      do 50 k=2,n
        i = n - k + 1
        d = zz(i,i)
        l = i + 1
        do 48 jj=1,n
          t = zinv(i,jj)
          do 45 j=l,n
            t = t - zz(i,j)*zinv(j,jj)
   45     continue
          zinv(i,jj) = t/d
   48   continue
   50 continue
c
      return
      end
c
c
      subroutine check(mtotv,mhbw,ntotv,nhbw,npoin,nkode,
     1gstif,asdis,aslod,resld,rnorm)
      dimension gstif(mtotv,mhbw),asdis(mtotv),aslod(mtotv)
     1            ,resld(mtotv),nkode(npoin)
c
c...find residual as gstif*asdis - aslod
      do 50 ii=1,ntotv
        sum = 0.0
        do 10 j=1,nhbw
          k = ii + j - 1
          if (k.gt.ntotv) goto 10
          sum = sum + gstif(ii,j)*asdis(k)
   10   continue
        if (ii.eq.1) goto 30
        do 20 i=1,ii-1
          k = ii - i + 1
          if (k.gt.nhbw) goto 20
          sum = sum + gstif(i,k)*asdis(i)
   20   continue
   30   continue
        sum = sum - aslod(ii)
        resld(ii) = sum
   50 continue
c
c...zero the fixed d.o.f.s
      do 70 np=1,npoin
        kode = nkode(np)
        k1 = kode/10
        k2 = kode - k1*10
        nn = 2*(np-1)
        if (k1.eq.1) resld(nn+1) = 0
        if (k2.eq.1) resld(nn+2) = 0
   70 continue
c
c...now find the max. norm of resld
      call norm1(mtotv,ntotv,resld,rnorm)
c
      return
      end
c
c
c
      subroutine norm1(m,n,v,vnorm)
      dimension v(m)
c
c...finds the max. norm of vector v
      vnorm = 0.0
      do 5 i=1,n
        vi = abs(v(i))
        if (vi.gt.vnorm) vnorm = vi
    5 continue
c
      return
      end
c
c
c
cv...SUBROUTINES FOR VISCOPLASTICITY
c
c
      subroutine norm2(m,n,v,vnorm)
      dimension v(m)
c
c...finds the Euclidean norm of vector v
      vnorm = 0.0
      do 5 i=1,n
        vi = v(i)
        vnorm = vnorm + vi*vi
    5 continue
      vnorm = sqrt(vnorm)
c
      return
      end
c
c
      subroutine hmat(triax,dilat,sigc,gamma,gpsig,hmatx)
      dimension hmatx(3,3),gpsig(3),avect(3),bvect(3),dbds(3,3)
c
cv...forms the H matrix at current stress state gpsig
cv...  H = gamma*(Phi(F)*db/d(sigma) + Phi'*b*a^T)
      call ycrit(triax,sigc,gpsig,fvalu)
      call flowv(triax,gpsig,avect)
      call flowv(dilat,gpsig,bvect)
cv...assumes associated flow: avect=bvect
      call flowm(dilat,gpsig,dbds)
      phifn = fvalu
      dphidf = 1.0
cv...assumes Phi(F) = F
      do 5 i=1,3
        do 5 j=1,3
          hij = dphidf*bvect(i)*avect(j)
          hij = hij + phifn*dbds(i,j)
          hij = gamma*hij
          hmatx(i,j) = hij
    5 continue
      return
      end
c
c
      subroutine dmatvp(triax,dilat,sigc,gamma,theta,dtime,
     1           gpsig,demat,dmatx)
      dimension gpsig(3),demat(3,3),dmatx(3,3),hmatx(3,3),
     1           dhmatx(3,3),dhinv(3,3)
c
cv...forms the viscoplastic D matrix (D^HAT) at current stress state
cv...         D^HAT = (I + theta*dtime*D*H)^{-1} * D
cv...result returned in DMATX. Elastic D matrix is in DEMAT
      call hmat(triax,dilat,sigc,gamma,gpsig,hmatx)
      do 10 i=1,3
        do 10 j=1,3
          dhij = 0.0
          do 5 k=1,3
            dhij = dhij + demat(i,k)*hmatx(k,j)
    5     continue
          dhmatx(i,j) = theta*dtime*dhij
   10 continue
c
cv...add I by adding 1 to diagonal elements
      do 20 i=1,3
        dhmatx(i,i) = 1.0 + dhmatx(i,i)
   20 continue
c
      call invert(3,dhmatx,dhinv)
c
      do 30 i=1,3
        do 30 j=1,3
          dij = 0.0
          do 25 k=1,3
            dij = dij + dhinv(i,k)*demat(k,j)
   25     continue
          dmatx(i,j) = dij
   30 continue
      return
      end
c
c
      subroutine strain(melem,msets,mcomp,nelem,lpros,props,
     1                    strsg,vivel,mohrg,dtime,theta,newfl)
      dimension lpros(melem),props(msets,mcomp),bvect(3),gpwts(4),
     1gploc(2,4),strsg(3,4,melem),vivel(3,4,melem),mohrg(4,melem),
     2gpsig(3)
cv...tests new g.p. stresses for yield, updates mohrg, and
cv...uses flow rule to find new vpl strain velocities in vivel
c
cv...newfl will indicate if any g.p. has newly yielded
      newfl = 0
cv...start of element loop
      do 80 l=1,nelem
        lset = lpros(l)
        triax = props(lset,3)
        sigc = props(lset,4)
        gamma = props(lset,5)
        dilat = props(lset,6)
        call gauss(gploc,gpwts,ngaus)
cv...note we don't use gploc or gpwts
c
cv...start of gauss point loop 
        do 75 ij=1,ngaus
c
cv...store stress at g.p. in gpsig
          do 10 i=1,3
            gpsig(i) = strsg(i,ij,l)
   10     continue
c
cv...test for yield
          mohr0 = mohrg(ij,l)
          call ycrit(triax,sigc,gpsig,fvalu)
          if (mohr0.eq.0) then
            if (fvalu.le.0.0) then
cv...g.pt. stays elastic - nothing to do
              go to 75
            else
cv...pt. has yielded at this time - update mohrg
              mohrg(ij,l) = 1
              newfl = 1
              write(6,*)l,ij,': fvalu = ',fvalu
            endif
          endif
cv...calculate new strain vel. in vivel, by flow rule
          if (fvalu.gt.0.0) then
cv...assume Phi(F) = F
            phifn = fvalu
            call flowv(dilat,gpsig,bvect)
            do 20 i=1,3
              vivel(i,ij,l) = gamma*phifn*bvect(i)
   20       continue
          else
cv...g.p. has stopped straining (but is still plastic)
            do 30 i=1,3
              vivel(i,ij,l) = 0.0
   30       continue
          endif
c
   75   continue
c
   80 continue
c
      return
      end
c
c
        subroutine newrhs(melem,mpoin,mnode,mtotv,msets,mcomp,
     1           nelem,npoin,ntotv,lpros,props,lnods,lnodn,coord,
     2           theta,dtime,nkode,strsg,aslod,vivel,mohrg)
c
      dimension lnods(melem,mnode),encod(2,8),nkode(mpoin),
     1         bmatx(3,16),gpwts(4),lnodn(melem),
     2         coord(mpoin,2),strsg(3,4,melem),aslod(mtotv),
     3       shape(8),gploc(2,4),deriv(2,8),cartd(2,8)
     4  ,gpsig(3),enres(16)
     5  ,lpros(melem),props(msets,mcomp),vivel(3,4,melem)
     6  ,demat(3,3),dmatx(3,3),mohrg(4,melem),gpviv(3),evec1(3)
c
cc..calculate new rh.s. vector in aslod
c
      do 5 i=1,ntotv
        aslod(i) = 0.0
    5 continue
c
c...start loop over all elements
      do 80 l=1,nelem
c...extract Young's modulus & Poisson's ratio from props      
      lset = lpros(l)
      ym = props(lset,1)
      pr = props(lset,2)
cc..and the Mohr-Coulomb plasticity parameters k and sigma_c
      triax = props(lset,3)
      sigc = props(lset,4)
cv...and gamma
      gamma = props(lset,5)
      dilat = props(lset,6)
      thick = 1.0
c
c...get elastic D matrix
      call dmat(ym,pr,demat)
c
c...put nodal coords in encod
      nnode = lnodn(l)
      do 15 n=1,nnode
      nn = lnods(l,n)
      encod(1,n) = coord(nn,1)
      encod(2,n) = coord(nn,2)
   15 continue
c
c...put local coords of 2x2 gauss-pts in gploc
      call gauss(gploc,gpwts,ngaus)
c
c...collect element contribution in enres
      do 20 i=1,16
        enres(i) = 0.0
   20 continue
c
c...start loop through gauss-pts
      do 75 ij=1,ngaus
c
cv...if g.pt. is elastic, no contribution
      if (mohrg(ij,l).eq.0) goto 75
c
c...put local coords of gauss-pt in s,t
      s = gploc(1,ij)
      t = gploc(2,ij)
c...get shape fns (shape) & derivs (deriv) in local coords
      call sfr(s,t,shape,deriv)
c...get xy-derivs (cartd) & derterminant of Jacobian (djacb)
      call jacob(encod,shape,deriv,cartd,djacb)
c...get B matrix 
      call bmat(cartd,bmatx)
c...store Gauss point stresses in gpsig
      do 25 i=1,3
        gpsig(i) = strsg(i,ij,l)
        gpviv(i) = vivel(i,ij,l)
   25 continue
c
cv...get the d^hat matrix 
      call dmatvp(triax,dilat,sigc,gamma,theta,dtime,
     1            gpsig,demat,dmatx)
c        
      do 30 i=1,3
        si = 0.0
        do 28 j=1,3
          si = si + dmatx(i,j)*gpviv(j)
   28   continue
        evec1(i) = si
   30 continue
c
cc..add GP contribution to integral(B^T * D * vivel) to enres
      dvolu = thick*djacb*gpwts(ij)
      do 33 i=1,16
        gi = 0.0
        do 32 j=1,3
          gi = gi + bmatx(j,i)*evec1(j)
   32   continue
        enres(i) = enres(i) + gi*dvolu
   33 continue
c
cc..end of Gauss point loop
   75 continue
c
cc..now add enres into the global r.h.s. vector
      do 77 n=1,nnode
        nn = lnods(l,n)
        ni = 2*(n-1)
        nni = 2*(nn-1)
        aslod(nni+1) = aslod(nni+1) + enres(ni+1)
        aslod(nni+2) = aslod(nni+2) + enres(ni+2)
   77 continue
c
cc..end of element loop
   80 continue
c
cv...multiply aslod by dtime
      do 85 i=1,ntotv
        aslod(i) = dtime*aslod(i)
   85 continue
c
c...and zero the components corresponding to fixed d.o.f.s
      do 90 np=1,npoin
        kode = nkode(np)
        npi = 2*(np-1)
c...separate kode out into two digits
        kx = kode/10
        ky = kode - 10*kx
        if (kx.eq.1) aslod(npi+1) = 0.0
        if (ky.eq.1) aslod(npi+2) = 0.0
   90 continue
c
      return
      end
c
c
      subroutine steady(melem,nelem,vivel,mohrg,kount,first,ratio)
      dimension vivel(3,4,melem),mohrg(4,melem),gpviv(3)
c
cv...forms ratio of current to initial plastic strain velocities, to
cv...use in test for equilibrium
c
cv...assumes 4 Gauss points per element
      total = 0.0
      iplas = 0
cv...iplas becomes 1 when a plastic g.pt. is reached
      do 80 l=1,nelem
        do 75 ij=1,4
          mohr = mohrg(ij,l)
          if (mohr.eq.0) goto 75
          iplas = 1
          do 10 i=1,3
            gpviv(i) = vivel(i,ij,l)
   10     continue
          call norm2(3,3,gpviv,gpn)
          total = total + gpn
   75   continue
   80 continue
c
      if (iplas.eq.0) then
        ratio = 0.0
      else
        if (kount.eq.1) first = total
        ratio = 100.0*total/first
      endif
c
      return
      end